/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 *  Revision: $Revision: 1.6 $ 
 */
package org.skunk.config;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import org.skunk.trace.Debug;

public class Configurator
{
    private ConfigStore store, defaultStore, reversionStore;
    private static Configurator configurator=null;

    public static Configurator getConfigurator()
    {
	return configurator;
    }

    public static void setConfigurator(Configurator configurator)
    {
	Configurator.configurator=configurator;
    }

    public Configurator(ConfigStore store)
    {
	this.store=store;
	this.defaultStore=new InMemoryConfigStore();
	this.reversionStore=new InMemoryConfigStore();
    }

    public ConfigStore getStore()
    {
	return store;
    }

    public ConfigStore getReversionStore()
    {
	return reversionStore;
    }

    public ConfigStore getDefaultStore()
    {
	return defaultStore;
    }

    /**
     * configure the object's properties.  
     * Will take an array, too.
     * 
     */
    public void configure(Object configurable)
    {
	if (configurable instanceof Class) 
	    throw new IllegalArgumentException("configurable object may not be a Class object");
	Object[] configurables=(configurable instanceof Object[])
	    ? (Object[]) configurable
	    : new Object[] {configurable};
	for (int i=0;i<configurables.length;i++)
	{
	    _configure(configurable, store, false);
	}
    }

    /**
     * starts a try transaction
     * @param configurable object to be configured -- may not be null
     */
    public void tryOn(Object configurable)
    {
	if (Debug.DEBUG)
	    Debug.trace(this, Debug.DP4, "object to try on: {0}", configurable);
	if (configurable instanceof Class) 
	    throw new IllegalArgumentException("configurable object may not be a Class object");
	Object[] configurables=(configurable instanceof Object[])
	    ? (Object[]) configurable
	    : new Object[] {configurable};
	for (int i=0;i<configurables.length;i++)
	{
	    //we only want to store the original state of the data, before the
	    //try tranaction started.
	    boolean storeReversionData=!reversionStore.isRegistered(configurable.getClass());
	    _configure(configurable, store, storeReversionData);		      
	}
    }

    /**
     * ends a try transaction
     * @param configurable Object to be configured, or the class of such -- may not be null
     */
    public void tryOff(Object configurableOrClass)
    {
	if (configurableOrClass instanceof Class)
	{
	    reversionStore.unregister((Class)configurableOrClass);
	}
	else reversionStore.unregister(configurableOrClass.getClass());
    }

    public void revert(Object configurable)
    {
	if (configurable instanceof Class) 
	    throw new IllegalArgumentException("configurable object may not be a Class object");
	if (configurable instanceof Object[])
	{
	    revertArray((Object[])configurable);
	}
	else
	{
	    _configure(configurable, reversionStore, false);	    
	    reversionStore.unregister(configurable.getClass());
	}
    }

    public void revertToDefault(Object configurable)
    {
	if (configurable instanceof Class)
	    throw new IllegalArgumentException("configurable object may not be a Class object");
	_configure(configurable, defaultStore, false);
    }

    private void revertArray(Object[] configurables)
    {
	HashMap classSet=new HashMap();
	//to facilitate use of a HashMap as a set (HashSet simply uses a HashMap instance internally)
	byte[] nothing=new byte[0];
	for (int i=0;i<configurables.length;i++)
	{
	    Object configurable=configurables[i];
	    if (configurable instanceof Class) 
		throw new IllegalArgumentException("configurable object may not be a Class object");
	    _configure(configurable, reversionStore, false);
	    classSet.put(configurable.getClass(), nothing);	    
	}
	for (Iterator it=classSet.keySet().iterator();it.hasNext();)
	    reversionStore.unregister((Class)it.next());
    }

    private void _configure(Object configurable, ConfigStore configStore, boolean enableReversion)
    {
	Class configClass=configurable.getClass();
	boolean needDefaultValue=!defaultStore.isRegistered(configClass);
	try
	{
	    BeanInfo info=Introspector.getBeanInfo(configClass);
	    PropertyDescriptor[] propDescriptors=info.getPropertyDescriptors();
	    for (int i=0;i<propDescriptors.length;i++)
	    {
		PropertyDescriptor pd=propDescriptors[i];
		String propName=pd.getName();
		//check store for property
		Object value=configStore.getConfigValue(configClass, 
						    propName);
		if (value!=null)
		{
		    if (Debug.DEBUG) 
			Debug.trace(this, Debug.DP3, 
				    "property found to configure: {0}", 
				    propName);
	
		    //find getter method and store the current values in the reversion store
		    if (enableReversion || needDefaultValue)
		    {
			Method getter=pd.getReadMethod();
			if (getter!=null)
			{
			    Object currentValue=getter.invoke(configurable, null);
			    if (Debug.DEBUG)
				Debug.trace(this, 
					    Debug.DP3, 
					    "current config value is {0}", 
					    currentValue.toString());
			    if (enableReversion)
			    {
				if (Debug.DEBUG)
				    Debug.trace(this, 
						Debug.DP3, 
						"in reversion mode");
				reversionStore.setConfigValue(configClass, 
							      propName, 
							      currentValue);
			    }
			    if (needDefaultValue)
			    {
				if (Debug.DEBUG)
				    Debug.trace(this, 
						Debug.DP3,
						"getting default value");
				defaultStore.setConfigValue(configClass, 
							    propName, 
							    currentValue);
			    }
			}
		    }
		    if (Debug.DEBUG)
			Debug.trace(this, 
				    Debug.DP3, 
				    "getting write method: will invoke with argument {0}",
				    new Object[] {value});
		    Method setter=pd.getWriteMethod();
		    if (setter!=null)
			setter.invoke(configurable, new Object[] { value });
		}
	    }
	}
	catch (IntrospectionException intro)
	{
	    if (Debug.DEBUG) Debug.trace(this, Debug.DP2, intro);
	}
	catch (IllegalAccessException illAcc)
	{
	    if (Debug.DEBUG) Debug.trace(this, Debug.DP2, illAcc);
	}
	catch (IllegalArgumentException illArg)
	{
	    if (Debug.DEBUG) Debug.trace(this, Debug.DP2, illArg);
	}
	catch (InvocationTargetException invoTarg)
	{
	    if (Debug.DEBUG) Debug.trace(this, Debug.DP2, invoTarg);
	}
    }
}

/*
$Log: Configurator.java,v $
Revision 1.6  2001/02/16 18:15:10  smulloni
many fixes to TextEditorCustomizer.  FileMode and SyntaxStyle now have a
configData property (they will probably be made into sibling classes).

Revision 1.5  2001/02/06 00:11:18  smulloni
struggle, perhaps futile, with the TextEditorCustomizer and other implicated
classes

Revision 1.4  2000/11/06 23:35:39  smullyan
numerous changes to the configuration system, which is veering dangerously
towards correct function.  LockRenderer is a table cell renderer that displays
a lock icon.

Revision 1.3  2000/11/03 22:51:52  smullyan
various changes to makefile to get it to work properly -- had to eliminate
wildcard imports to get jikes to behave.  Continued hacking at configuration
system, which now stores default settings.  TableViewCustomizer currently
quite broken as I integrate into it my new DualListPanel component.

Revision 1.2  2000/11/03 00:16:33  smullyan
massive work on the configuration system -- still buggy and rough.  One
customizer sketched (inadequately, but for proof of concept):
TableViewCustomizer.

Revision 1.1  2000/11/01 15:30:14  smullyan
adding config classes (moved from package org.skunk.dav.client.gui.config)

Revision 1.2  2000/11/01 04:18:40  smullyan
changes to configuration system: support of reversible configuration (in
preparation for Preferences)

*/

