/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client;

import java.io.IOException;
import java.util.*;
import org.skunk.minixml.*;
import org.skunk.trace.Debug;

/**
 *  base class for DAV methods.
 */
public abstract class AbstractDAVMethod implements DAVMethod
{
    private Map reqHeaders=new HashMap();
    private Map resHeaders=new HashMap();
    private byte[] reqBody, resBody;
    private String path;
    private int status;
    private DAVFile daFile;
    
    /**
     * @param path the resource on which the method is performed
     */
    public AbstractDAVMethod(String path)
    {
        this.path=path;
	daFile=new DAVFile(path);
    }

    /**
     * sets the protocol
     * @param protocol the protocol
     */
    public void setProtocol(String protocol)
    {
	daFile.setProtocol(protocol);
    }

    /**
     * returns the protocol
     * @return the protocol 
     */
    public String getProtocol()
    {
	return daFile.getProtocol();
    }

    /**
     * sets the hostname
     * @param host the hostname
     */
    public void setHost(String host)
    {
	daFile.setHost(host);
    }

    /**
     * sets the port
     * @param port the port
     */
    public void setPort(int port)
    {
	daFile.setPort(port);
    }

    /**
     * gets the hostname of the server.
     * @return the hostname
     */
    public String getHost()
    {
	return daFile.getHost();
    }

    /**
     * gets the port where the file in question is served.
     * @return the port
     */
    public int getPort()
    {
	return daFile.getPort();
    }

    /**
     * gets the headers of the request
     * @return the request headers
     */
    public Map getRequestHeaders()
    {
        return reqHeaders;
    }

    /**
     * sets the headers of the request 
     * @param m a map of header names and values
     */
    public void setRequestHeaders(Map m)
    {
        reqHeaders=m;
    }

    /**
     * sets the body of the request.
     * @param b the new request body
     */
    public void setRequestBody(byte[] b)
    {
        reqBody=b;
    }
    
    /**
     * returns the path requested
     * @return the path requested.
     */
    public String getRequestURL()
    {
        return path;
    }
    
    /**
     * returns the body of the request
     * @return the body of the request, if any, or null
     */
    public byte[] getRequestBody()
    {
        return reqBody;
    }
    
    /**
     * used to populate the headers of the response
     * @param headers a mapping of headers names to values
     */
    public void setResponseHeaders(Map headers)
    {
	this.resHeaders=headers;
    }
    
    /**
     * used to populate the body of the response
     * 
     */
    public void setResponseBody(byte[] body)
    {
	this.resBody=body;
    }

    /**
     * @return the headers of the response
     */
    public Map getResponseHeaders()
    {
        return resHeaders;
    }

    /**
     * @return the body of the response
     */
    public byte[] getResponseBody()
    {
        return resBody;
    }

    /**
     * @return the status of the response
     */
    public int getStatus()
    {
        return status;
    }

    /**
     * used to the populate the status of the response
     * @param status the status 
     */
    public void setStatus(int status)
    {
        this.status=status;
    }

    /**
     * @return the DAVFile corresponding to the resourceURL.
     */
    public DAVFile getDAVFile()
    {
	return daFile;
    }

    /**
     * a hook into processing the body of the response.  By default will 
     * parse a multistatus response; otherwise, does nothing.
     */
    public void processResponseBody() throws MalformedXMLException
    {
	if (getStatus()==207)
	    parseMultistatus(getResponseBody());
    }

    /**
     * a hook for processing the response headers.  By default does nothing.
     */
    public void processResponseHeaders() 
    {
	//by default, do nothing
    }

    /**
     * a hook for processing the request headers.  By default does nothing.
     */
    public void processRequestHeaders()
    {
	//by default, do nothing
    }

    /**
     * a hook for processing the request body.  By default does nothing.
     */
    public void processRequestBody()
    {
	//by default, do nothing
    }

    private void parseMultistatus(byte[] responseBody) throws MalformedXMLException
    {
	XMLParser parser=new XMLParser();
	try
	{
	    XMLDocument doc=parser.parse(new String(responseBody));
	    //for debugging only!
	    if (Debug.isDebug(this, Debug.DP8))
	    {
		XMLViewer.viewDocumentDialog(doc);
	    }
	    //populate the DAVFile
	    populateFile(doc);
	}
	catch (IOException oyVeh)
	{
	    Debug.trace(this, Debug.DP1, oyVeh);
	}
    }

    private void populateFile(XMLDocument doc)
    {
	//there is no guarantee of order in the response listing of the multistatus
	XMLElement multistatus=doc.getRootElement();
	if (multistatus==null ||(! multistatus.getElementName().equals(DAVConstants.MULTISTATUS_ELEM)))
	{
	    Debug.trace(this, Debug.DP2, "multistatus element not found in response body");
	    return;
	}
	for (ListIterator lit=multistatus.children();lit.hasNext();)
	{
	    //we should have response elements here
	    Object o=lit.next();
	    if (!(o instanceof XMLElement))
	    {
		Debug.trace(this, Debug.DP3, "weird object inside multistatus response: "+ o);
		continue;
	    }
	    XMLElement elem=(XMLElement) o;
	    if (elem.getElementName().equals(DAVConstants.RESPONSEDESCRIPTION_ELEM))
	    {
		Object desc=elem.getChild(0);
		if (desc!=null)
		    daFile.setResponseDescription(desc.toString());
	    }
	    else if (elem.getElementName().equals(DAVConstants.RESPONSE_ELEM))
	    {
		XMLElement href=elem.getChild(DAVConstants.HREF_ELEM);
		
		if (href==null || ((o=href.getChild(0))==null))
		{
		    Debug.trace(this, Debug.DP2, "multistatus response object in illegal format: no href");
		    continue;
		}
		    
		String filename=o.toString().trim();
		//if the resource is a collection, its name may be returned ending with a slash.
		if (stripLastSlash(daFile.getName()).equals(stripLastSlash(filename)))
		{
		    populate(daFile, elem);
		}
		else 
		{
		    //file should be a collection: add new DAVFile as child
		    DAVFile child=new DAVFile(daFile.getProtocol(),
					      daFile.getHost(), 
					      daFile.getPort(), 
					      filename);
		    populate(child, elem);
		    daFile.addChild(child);		
		}
		
	    }
	}	
    }
    /**
     * chomps the last slash off a path, which may not be null.
     * @param path the path to chomp
     */
    protected static final String stripLastSlash(String path)
    {
	if (path.endsWith("/"))
	{
	    return path.substring(0, path.length()-1);
	}
	else return path;
    }

    private void populate(DAVFile file, XMLElement responseElem)
    {
	/*response has three possible child nodes: href, status and propstat.
	  href has already been read, in order to create the DAVFile.
	  status is what it sounds like.
	  propstat has two child nodes: prop and status. 
	  prop may return a number of properties, depending on the propfind
	  query type.
	 */
	XMLElement statusElem=responseElem.getChild(DAVConstants.STATUS_ELEM);
	if (statusElem!=null)
	{
	    Object o=statusElem.getChild(0);
	    if (o==null)
	    {
		Debug.trace(this, 
			    Debug.DP2, 
			    "null status elem found for " + file.getName());
		return;
	    }
	    file.setStatus(extractStatus(o.toString()));
	}
	//By adding the namespace specification, achieves Zope interoperability, which sends two propstats
	//with different namespaces!  Oy veh. -- js Sat Sep  1 12:59:02 2001
	XMLElement propstatElem=responseElem.getChild(DAVConstants.PROPSTAT_ELEM, DAVConstants.DAV_NAMESPACE);
	if (propstatElem==null)
	{
	    Debug.trace(this, 
			Debug.DP2, 
			"no propstat element found in  response for "+file.getName());
	    return;
	}

	/********first, the status element ************/

	XMLElement elem=propstatElem.getChild(DAVConstants.STATUS_ELEM);
	if (elem==null)
	{
	    Debug.trace(this, 
			Debug.DP2, 
			"no status element found in propstat for "+file.getName());
	    return;
	}
	Object o=elem.getChild(0);
	if (o==null)
	{
	    Debug.trace(this, 
			Debug.DP2, 
			"no status string found in status element for "+file.getName());
	    return;
	}
	file.setStatus(extractStatus(o.toString()));
	/********then, the responsedescription, if any *************/
	XMLElement responsedescriptionElem=propstatElem.getChild(DAVConstants.RESPONSEDESCRIPTION_ELEM);
	if (responsedescriptionElem!=null)
	{
	    Object desc=responsedescriptionElem.getChild(0);
	    if (desc!=null)
	    {
		//file may have a responsedescription set a the top level, too.
		Debug.trace(this, Debug.DP3, "responseDescription: "+desc);
		String currentDescription=file.getResponseDescription();
		Debug.trace(this, Debug.DP3, "current description: "+currentDescription);
		StringBuffer sb=new StringBuffer();
		if (currentDescription!=null)  
		    sb.append(currentDescription).append(": ");
		sb.append(desc);
		    file.setResponseDescription(sb.toString());
	    }
	}

	/********now, the props ************/
	
	elem=propstatElem.getChild(DAVConstants.PROP_ELEM);
	processPropElement(file, elem);


    }

    /**
     * by default, passes the properties off to PropertyHandler
     * @param file the DAVFile in which the results are stored
     * @param elem the prop XMLElement being processed
     */
    protected void processPropElement(DAVFile file, XMLElement elem)
    {
	for (ListIterator lit=elem.children();lit.hasNext();)
	{
	    Object aProperty=lit.next();
	    if (! (aProperty instanceof XMLElement))
		continue;
	    XMLElement propElem=(XMLElement) aProperty;
	    PropertyHandler.handleProperty(file, propElem);
	}	    	    
    }	

    /**
     * convenience function for parsing the status header.
     * @param statusLine the HTTP status line
     * @return the status as an Integer
     */
    protected static final Integer extractStatus(String statusLine)
    {
	StringTokenizer st=new StringTokenizer(statusLine);
	Integer status=null;
	if (st.countTokens()>=2)
	{
	    st.nextToken(); //discard "HTTP/1.1"	    
	    String statusStr=st.nextToken();
	    try
	    {				
		status= Integer.decode(statusStr);
	    }
	    catch (NumberFormatException nafta)
	    {
		Debug.trace(AbstractDAVMethod.class, Debug.DP3, nafta);
	    }
	}
	return status;
    }
}

/* $Log: AbstractDAVMethod.java,v $
/* Revision 1.16  2001/09/01 18:25:56  smulloni
/* fixes for Zope interoperability; version 1.0.2.4 dev
/*
/* Revision 1.15  2001/07/20 18:45:21  smulloni
/* fixed handling of <response/> elements, which was skipping status elements
/* not contained in a propstat.
/*
/* Revision 1.14  2001/01/03 22:51:53  smulloni
/* added documentation
/*
/* Revision 1.13  2001/01/03 20:11:30  smulloni
/* the DAVFileChooser now replaces JFileChooser for remote file access.
/* DAVMethod now has a protocol property.
/*
/* Revision 1.12  2001/01/02 17:43:38  smulloni
/* changed debug level for XMLViewer popups to Debug.DP8.
/*
/* Revision 1.11  2000/12/19 22:06:15  smulloni
/* adding documentation.
/*
/* Revision 1.10  2000/12/03 23:53:25  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.9  2000/11/14 23:18:31  smullyan
/* fix for responsedescription property in multistatus
/*
/* Revision 1.8  2000/11/09 23:34:47  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
