/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/


package org.skunk.dav.client;

import HTTPClient.AuthorizationHandler;
import HTTPClient.AuthorizationInfo;
import HTTPClient.AuthSchemeNotImplException;
import HTTPClient.Codecs;
import HTTPClient.CookieModule;
import HTTPClient.HTTPConnection;
import HTTPClient.HTTPResponse;
import HTTPClient.ModuleException;
import HTTPClient.NVPair;
import HTTPClient.ProtocolNotSuppException;
import HTTPClient.Response;
import HTTPClient.RoRequest;
import HTTPClient.RoResponse;
import java.net.PasswordAuthentication;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import org.skunk.minixml.MalformedXMLException;
import org.skunk.trace.Debug;

/**
 * An http connection that can execute <code>DAVMethod</code> objects.
 *
 * @author Jacob Smullyan
 */
public class DAVConnection
{
    /*
      rather than subclassing HTTPConnection directly,
      I'm using it as a delegate to free myself from
      its peculiarities
    */
    private HTTPConnection connectionDelegate;
    private String host;
    private int port;
    private String username="anonymous";
    private String authorization;
    private boolean closed=true;
    private String protocol;
    private int socketTimeout=5000;

    public static final String SOCKET_TIMEOUT_PROPERTY="socketTimeout";
    public static final int DEFAULT_SOCKET_TIMEOUT=5000;

//     static
//     {
// 	//this permits all cookies -- let the floodgates open!
// 	CookieModule.setCookiePolicyHandler(null);
//     }
      
    /**
     * creates an DAVConnection over the http protocol
     * @param host the hostname 
     * @param port the port
     */
    public DAVConnection(String host, int port)
    {
	this(host, port, false);
    }
    /**
     * creates a DAVConnection
     * @param host the host to which to connect
     * @param port the port on which to connect
     * @param https whether to use https or http (https if true, http otherwise)
     */
    public DAVConnection(String host, int port, boolean https)
    {
	this.host=host;
	this.port=port;
	this.protocol=(https) ? DAVConstants.HTTPS : DAVConstants.HTTP;
	openConnection();
    }

    /**
     *  returns the protocol, either "http" or "https"
     * @return the protocol
     */
    public String getProtocol()
    {
	return this.protocol;
    }

    /**
     *  set a DAVAuthenticator for this connection.
     *
     *  the DAVAuthenticator is only being used for its requestPA() method (which provides
     *  access to the output of Authenticator's protected requestPasswordAuthentication() method)
     *  
     *  @param authenticator a DAVAuthenticator
     */
    public void setAuthenticator(DAVAuthenticator authenticator)
    {
	AuthorizationInfo.setAuthHandler(new HTTPAuthHandler(authenticator, this));
    }

    /**
     * sets an encoded authorization string
     * @param authorization the authorization string
     */
    public void setAuthorization(String authorization)
    {
	this.authorization=authorization;
    }

    /**
     * indicates whether the connection is closed or not
     * @return whether the connection is closed
     */
    public boolean isClosed()
    {
	return this.closed;
    }

    public int getSocketTimeout()
    {
	return socketTimeout;
    }

    public void setSocketTimeout(int socketTimeout)
    {
	this.socketTimeout=socketTimeout;
	connectionDelegate.setTimeout(socketTimeout);
    }

    /**
     * opens the connection
     */
    public void openConnection()
    {
	if (!closed) closeConnection();
	if (protocol.equals(DAVConstants.HTTP))
	    connectionDelegate=new HTTPConnection(host, port);
	else
	{
	    try
	    {
		connectionDelegate=new HTTPConnection(protocol, host, port);
	    }
	    catch (ProtocolNotSuppException pencil)
	    {
		Debug.trace(this, Debug.DP2, pencil);
		return;
	    }
	}
	NVPair[] def_hdrs = { 
	    new NVPair("Connection", "close"),
	    new NVPair("User-Agent", "Skunkzilla/1.0")
	};
	connectionDelegate.setDefaultHeaders(def_hdrs);
	connectionDelegate.setTimeout(this.socketTimeout);

	this.closed=false;
    }
    
    /**
     * closes the connection
     */
    public void closeConnection()
    {
	connectionDelegate.stop();
	connectionDelegate=null;
	this.closed=true;
    }
 
    /**
     * executes a DAVMethod
     * 
     * @param method the DAVMethod instance
     */
    public synchronized void execute(DAVMethod method)
	throws IOException, DAVException
    {
	if (authorization!=null)
	{
	    Debug.trace(this, Debug.DP4, "adding authorization header: "+authorization);
	    method.getRequestHeaders().put("Authorization", authorization);
	}
	method.setProtocol(protocol);
	method.setHost(host);
	method.setPort(port);
	method.processRequestHeaders();
	method.processRequestBody();
	Debug.trace(this, Debug.DP4, "request headers: "+method.getRequestHeaders());
	if (Debug.isDebug(this, Debug.DP4))
	{
	    byte[] reqbod=method.getRequestBody();
	    if (reqbod!=null)
		Debug.trace(this, Debug.DP4, "request body: {0} ", 
			    new String(method.getRequestBody()));
	    else Debug.trace(this, Debug.DP4, "request body is null");
	}
	try
	{
	    HTTPResponse res=
		connectionDelegate.ExtensionMethod(
						   method.getRequestMethodName().toString(),
						   method.getRequestURL(),
						   method.getRequestBody(),
						   convertToNameValuePairs(method.getRequestHeaders()));
	    processResponse(method, res);
	}
	catch (ModuleException modex)
	{
	    Debug.trace(this, Debug.DP3, modex); 
	    throw new DAVException(modex.getMessage());
	}
	catch (IOException oyVeh)
	{
	    Debug.trace(this, Debug.DP3,oyVeh);
	    throw new DAVException(oyVeh.getMessage());
	}
    }
      

    protected synchronized void processResponse(DAVMethod method, HTTPResponse res)
        throws IOException, ModuleException, DAVException
    {
        method.setResponseHeaders(extractHeaders(res));
        Debug.trace(this, Debug.DP4, "headers of response: "+method.getResponseHeaders());
        method.setResponseBody(res.getData());
        method.setStatus(res.getStatusCode());
	method.processResponseHeaders();
	try
	{
	    method.processResponseBody();
	}
	catch (MalformedXMLException malex)
	{
	    Debug.trace(this, Debug.DP3, malex);
	    throw new DAVException(malex.getMessage());
	}
        Debug.trace(this, Debug.DP4, "status of response: "+method.getStatus());
        byte[] body=method.getResponseBody();
        Debug.trace(this, Debug.DP4, "body of response: " 
		    + ((body!=null) ? new String(body) : "null"));
    }
    
    private Map extractHeaders(HTTPResponse res) throws IOException, DAVException
    {
	try
	{
	    Map m=new HeaderMap();
	    Enumeration headerList=res.listHeaders();
	    while (headerList.hasMoreElements())
	    {
		String headerKey=headerList.nextElement().toString();
		m.put(headerKey, res.getHeader(headerKey));
	    }
	    return m;
	}
	catch (ModuleException modex)
	{
	    Debug.trace(this, Debug.DP3, modex);
	    throw new DAVException(modex.getMessage());
	}
    }
    
    /**
     *  convert Maps to arrays of HTTPClient package's NVPair
     */
    static NVPair[] convertToNameValuePairs(Map m)
    {
        if ((m==null) || (m.keySet()==null)) return null;
        Set keys=m.keySet();
	
        NVPair[] pairs=new NVPair[keys.size()];
        int i=0;
        for (Iterator it=keys.iterator();it.hasNext();i++)
        {
            Object key=it.next();
            Object val=m.get(key);
            pairs[i]=new NVPair(key.toString(), val.toString());
        }
        return pairs;
    }
    
    /**
     * returns the host
     * @return the host
     */
    public String getHost() { return this.host; }

    /**
     * returns the port
     * @return the port
     */    
    public int getPort() { return this.port; } 

    /**
     * returns the username
     * @return the username
     */
    public String getUsername() { return this.username; }

    /**
     * sets the username
     * @param username the username
     */
    public void setUsername(String username) { this.username=username; }
}
       
	

/**
 *  for use with the HTTPClient connection delegate.  This is largely cribbed
 *  from DAVExplorer, which in turn cribbed it from HTTPClient's more fully
 *  functional (but ugly) DefaultAuthHandler.  Should the HTTPClient delegate
 *  be removed, a standard java.net.Authenticator can replace this.
 *
 *  @author Jacob Smullyan
 */
class HTTPAuthHandler implements AuthorizationHandler
{
    private DAVAuthenticator authenticator;
    private DAVConnection connection;

    /**
     * creates an HTTPAuthHandler
     * @param authenticator the authenticator
     * @param connection the DAVConnection
     */
    HTTPAuthHandler(DAVAuthenticator authenticator, DAVConnection connection)
    {
	this.authenticator=authenticator;
	this.connection=connection;
    }
    /**
     * populates the connection with the necessary authorization information
     */
    public AuthorizationInfo getAuthorization(AuthorizationInfo info,
				       RoRequest req, 
				       RoResponse resp)
	throws AuthSchemeNotImplException
    {
        if( !info.getScheme().equalsIgnoreCase("Basic") )
            throw new AuthSchemeNotImplException( info.getScheme() );
	authenticator.reset();
	PasswordAuthentication pa=authenticator.getPA();
	if (pa==null)
	    return null;
	String username=pa.getUserName();
	String password=new String(pa.getPassword());
	Debug.trace(this, Debug.DP6, "username is "+username+", password is "+password);
	if (username==null || username.equals(""))
	    return null;
	//pass along the username to the connection class, so it can be used elsewhere
	// (e.g., for lockowner)
	connection.setUsername(username);
        AuthorizationInfo ai= new AuthorizationInfo(info.getHost(), 
						    info.getPort(), 
						    info.getScheme(),
						    info.getRealm(),
						    Codecs.base64Encode(username + ":" + password));
	connection.setAuthorization(ai.toString());
	return ai;
    }

    /**
     * enforces that the authorization scheme is Basic
     */
    public AuthorizationInfo fixupAuthInfo(AuthorizationInfo info, 
				    RoRequest req,
				    AuthorizationInfo challenge, 
				    RoResponse resp)
	throws AuthSchemeNotImplException
    {
        if( info.getScheme().equalsIgnoreCase("Basic") )
            return challenge;
        else throw new AuthSchemeNotImplException( info.getScheme() );

    }
    public void handleAuthHeaders(Response resp, 
			   RoRequest req,
			   AuthorizationInfo prev, 
			   AuthorizationInfo prxy)
	throws IOException
    {
	//evidently okay to do nothing here
    }
    
    public void handleAuthTrailers(Response resp,  
			    RoRequest req,
			    AuthorizationInfo prev, 
			    AuthorizationInfo prxy)
	throws IOException
    {
	//evidently okay to do nothing here, too
    }
}

/* $Log: DAVConnection.java,v $
/* Revision 1.23  2001/05/30 03:49:35  smulloni
/* Some bug fixes and improvements to authentication handling.  Users are now
/* prompted for a password the first time they connect to a given server
/* in a session; if the password is rejected, they are asked again.
/* Users also now have the option of whether a given password will be remembered
/* or not.
/*
/* Revision 1.22  2001/01/23 20:50:09  smulloni
/* added configurable socket timeout
/*
/* Revision 1.21  2001/01/03 22:51:53  smulloni
/* added documentation
/*
/* Revision 1.20  2001/01/03 20:11:30  smulloni
/* the DAVFileChooser now replaces JFileChooser for remote file access.
/* DAVMethod now has a protocol property.
/*
/* Revision 1.19  2000/12/19 22:06:15  smulloni
/* adding documentation.
/*
/* Revision 1.18  2000/12/18 21:04:56  smulloni
/* changes to support SSL.
/*
/* Revision 1.17  2000/12/03 23:53:25  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.16  2000/11/15 19:45:35  smullyan
/* beginning of revamp of application to include multiple buffers.
/*
/* Revision 1.15  2000/11/15 05:20:04  smullyan
/* rolled back, for the time being, sending the Connection: close header and
/* nullifying the cookie policy handler.  mydocsonline has stopped working,
/* although probably not for this reason; I need to eliminate potentially
/* relevant changes to find the problem.  (Perhaps they have stopped accepted
/* allprop?)
/*
/* Revision 1.14  2000/11/14 23:18:31  smullyan
/* fix for responsedescription property in multistatus
/*
/* Revision 1.13  2000/11/09 23:34:49  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
