/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client;

import HTTPClient.SSLSupport;
import java.util.Enumeration;
import java.util.Hashtable;
import org.skunk.trace.Debug;

/**
 * a pool for DAVConnection objects.
 *
 * @author Jacob Smullyan
 */
public abstract class DAVConnectionPool
{
    private static Hashtable connectionHash=new Hashtable();

    /**
     * indicates at runtime whether the current build can support SSL
     * @return whether SSL is supported
     */
    public static final boolean canSupportSSL()
    {
	return SSLSupport.canSupport();
    }

    /**
     * returns a pooled DAVConnection instance
     * @param host the host
     * @param port the port
     * @param secure whether the protocol of the connection is https (otherwise, http)
     * @return the pooled DAVConnection instance
     */
    public static DAVConnection getDAVConnection(String host, int port, boolean secure)
    {
	String key=getKey(host, port);
	Debug.trace(DAVConnectionPool.class, Debug.DP4, "hashKey: "+key);
	DAVConnection conn;
	if (connectionHash.containsKey(key))
	{
	    conn= (DAVConnection) connectionHash.get(key);
	}
	else 
	{
	    Debug.trace(DAVConnection.class, Debug.DP4, 
			"new connection created for host "
			+ host + " and port " + port);	    
	    conn=new DAVConnection(host, port, secure);
	    connectionHash.put(key, conn);
	}
	return conn;
    }

    /**
     * returns a pooled DAVConnection instance with the http protocol
     * @param host the host
     * @param port the port
     * @return the pooled DAVConnection instance
     */
    public static DAVConnection getDAVConnection(String host, int port)
    {
	return getDAVConnection(host, port, false);
    }

    /**
     * removes the DAVConnection with the given host and port from the pool
     * @param host the host
     * @param port the port
     */
    public static void removeDAVConnection(String host, int port)
    {
	String key=getKey(host, port);
	if (connectionHash.containsKey(key))
	{
	    DAVConnection dc=(DAVConnection)connectionHash.remove(key);
	    if (!dc.isClosed())
	    {
		dc.closeConnection();
	    }
	}
    }
    /**
     * returns the key used in the DAVConnectionPool hashtable for a DAVConnection with the given host and port.
     * @param host the host
     * @param port the port
     * @return the key used
     */
    protected static String getKey(String host, int port)
    {
	return new StringBuffer()
	    .append(host)
	    .append(":")
	    .append(port)
	    .toString();
    }

    public static Enumeration getDAVConnections()
    {
	return connectionHash.elements();
    }
}

/* $Log: DAVConnectionPool.java,v $
/* Revision 1.7  2001/01/23 20:50:09  smulloni
/* added configurable socket timeout
/*
/* Revision 1.6  2001/01/03 22:51:53  smulloni
/* added documentation
/*
/* Revision 1.5  2000/12/19 22:06:15  smulloni
/* adding documentation.
/*
/* Revision 1.4  2000/12/18 23:22:47  smulloni
/* added optional SSL capability to the gui application.  Separate make targets
/* exist for building it will SSL support.
/*
/* Revision 1.3  2000/12/03 23:53:25  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.2  2000/11/09 23:34:49  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
