/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ListIterator;
import org.skunk.minixml.XMLCData;
import org.skunk.minixml.XMLElement;
import org.skunk.trace.Debug;

public class PropertyHandler
{
    private static Class[] paramTypeArray=new Class[] 
    {
	DAVFile.class, 
	XMLElement.class
    };
    //prefix for handler methods invoked through reflection
    public static final String HANDLER_METHOD_PREFIX="handle";

    public static void handleProperty(DAVFile file, XMLElement elem)
    {
	try
	{
	    Method handler=getHandlerMethod(elem.getElementName());
	    handler.invoke(null, new Object[] { file, elem } );
	}
	catch (NoSuchMethodException noSoap)
	{
	    handleCustomProperty(file, elem);
	}
	catch (SecurityException sucky)
	{
	    Debug.trace(PropertyHandler.class, 
			Debug.DP2,
			sucky);
	}
	catch (IllegalAccessException illack)
	{
	    Debug.trace(PropertyHandler.class,
			Debug.DP2,
			illack);
	}
	catch (IllegalArgumentException illarg)
	{
	    Debug.trace(PropertyHandler.class,
			Debug.DP4, //should only occur during development
			illarg);
	}
	catch (InvocationTargetException invook)
	{
	    Debug.trace(PropertyHandler.class,
			Debug.DP2,
			invook);
	}
    }

    private static Method getHandlerMethod(String propertyName) 
	throws NoSuchMethodException, SecurityException
    {
	if (propertyName==null || propertyName.length()==0) 
	    return null;
	String methodName=HANDLER_METHOD_PREFIX
	    + Character.toUpperCase(propertyName.charAt(0))
	    + propertyName.substring(1);
	Method handlerMethod=PropertyHandler.class.getDeclaredMethod(methodName, 
							     paramTypeArray);
	return handlerMethod;
    }

    private static void handleCustomProperty(DAVFile file, XMLElement elem)
    {
	// handling of dead and other special properties
	file.setCustomProperty((new DAVProperty(elem)), getChildAsString(elem));
    }

    private static void handleCreationdate(DAVFile file, XMLElement elem)
    {
	file.setCreationDate(getChildAsString(elem));
    }

    private static void handleDisplayname(DAVFile file, XMLElement elem)
    {
	file.setDisplayName(getChildAsString(elem));
    }

    private static void handleGetetag(DAVFile file, XMLElement elem)
    {
	file.setEtag(getChildAsString(elem));;
    }

    private static void handleGetlastmodified(DAVFile file, XMLElement elem)
    {
	file.setLastModified(getChildAsString(elem));
    }

    private static void handleResourcetype(DAVFile file, XMLElement elem)
    {
	if (elem.getChild(DAVConstants.COLLECTION_ELEM)!=null)
	    file.setResourceType(DAVConstants.COLLECTION_ELEM);
	else file.setResourceType(null);
    }

    private static void handleSupportedlock(DAVFile file, XMLElement elem)
    {
	for (ListIterator lit=elem.children();lit.hasNext();)
	{
	    Object o=lit.next();
	    if (!(o instanceof XMLElement))
		continue;
	    XMLElement lockEntry=(XMLElement)o;
	    if (!lockEntry.getElementName().equals(DAVConstants.LOCKENTRY_ELEM))
	    {
		Debug.trace(PropertyHandler.class, Debug.DP3, 
			    "bizarre child encountered in supported lock element: " 
			    + lockEntry);
		continue;
	    }
	    XMLElement lockScope=lockEntry.getChild(DAVConstants.LOCKSCOPE_ELEM);
	    XMLElement scopeCharacterization=lockScope.getChildElement(0);
	    XMLElement lockType=lockEntry.getChild(DAVConstants.LOCKTYPE_ELEM);
	    XMLElement typeCharacterization=lockType.getChildElement(0);

	    LockScope scopeArg=null;
	    if (scopeCharacterization!=null)
	    {
		String name=scopeCharacterization.getElementName();
		if (name.equals(DAVConstants.SHARED_ELEM))
		    scopeArg=LockScope.SHARED;
		else if (name.equals(DAVConstants.EXCLUSIVE_ELEM))
		    scopeArg=LockScope.EXCLUSIVE;
	    }
	    LockType typeArg=null;
	    if (typeCharacterization!=null)
	    {
		String name=typeCharacterization.getElementName();
		if (name.equals(DAVConstants.WRITE_ELEM))
		    typeArg=LockType.WRITE;
	    }
	    if (scopeArg!=null || typeArg!=null) //tolerant to a fault
		file.addSupportedLock(new Lock(scopeArg, typeArg));
	}	    
    }

    private static void handleGetcontentlanguage(DAVFile file, XMLElement elem)
    {
	file.setContentLanguage(getChildAsString(elem));
    }

    private static void handleGetcontenttype(DAVFile file, XMLElement elem)
    {
	file.setContentType(getChildAsString(elem));
    }

    private static void handleGetcontentlength(DAVFile file, XMLElement elem)
    {
	String lengthStr=getChildAsString(elem);
	if (lengthStr!=null)
	{
	    try
	    {
		file.setContentLength(Long.decode(lengthStr));
	    }
	    catch (NumberFormatException nafta)
	    {
		Debug.trace(PropertyHandler.class, Debug.DP2, nafta);
	    }
	}
    }

    private static void handleLockdiscovery(DAVFile file, XMLElement elem)
    {
	for (ListIterator lit=elem.children();lit.hasNext();)
	{
	    Object o=lit.next();
	    if (!(o instanceof XMLElement))
		continue;
	    XMLElement activelock=(XMLElement) o;
	    if (!activelock.getElementName().equals(DAVConstants.ACTIVELOCK_ELEM))
	    {
		Debug.trace(PropertyHandler.class, Debug.DP3, "weird child encountered in activelock: "+ activelock);
		continue;
	    }
	    LockType lt=null;
	    LockScope ls=null;
	    Depth depth=null;
	    String owner=null;
	    Timeout timeout=null;
	    String lockToken=null;
	    
	    for (ListIterator lit2=activelock.children();lit2.hasNext();)
	    {
		Object o2=lit2.next();
		if (!(o2 instanceof XMLElement))
		    continue;
		XMLElement subElem=(XMLElement) o2;
		String name=subElem.getElementName();
		if (name.equals(DAVConstants.LOCKTYPE_ELEM))
		{
		    XMLElement typeElem=subElem.getChildElement(0);
		    if (typeElem!=null)
		    {
			String typeStr=typeElem.getElementName();
			if (DAVConstants.WRITE_ELEM.equals(typeStr))
			{
			    lt=LockType.WRITE;
			}
			//the only recognized value at this time, others will be ignored
		    }
		}
		else if (name.equals(DAVConstants.LOCKSCOPE_ELEM))
		{
		    XMLElement scopeElem=subElem.getChildElement(0);
		    if (scopeElem!=null)
		    {
			String scopeStr=scopeElem.getElementName();
			if (DAVConstants.SHARED_ELEM.equals(scopeStr))
			{
			    ls=LockScope.SHARED;
			}
			else if (DAVConstants.EXCLUSIVE_ELEM.equals(scopeStr))
			{
			    ls=LockScope.EXCLUSIVE;
			}
		    }		    
		}
		else if (name.equals(DAVConstants.DEPTH_ELEM))
		{
		    String depthStr=getChildAsString(subElem);
		    depth=Depth.getDepth(depthStr);
		}
		else if (name.equals(DAVConstants.OWNER_ELEM))
		{
		    XMLElement hrefElem=subElem.getChild(DAVConstants.HREF_ELEM);
		    /**
		     * mod_dav puts a href element inside the owner tag.  RFC2518
		     * does not require this (the DTD allows any element inside an 
		     * owner tag), but all its examples include it.
		     */
		    owner=(hrefElem==null) 
			? getChildAsString(subElem)
			: getChildAsString(hrefElem);
		}
		else if (name.equals(DAVConstants.TIMEOUT_ELEM))
		{
		    timeout=Timeout.getTimeout(getChildAsString(subElem));
		}
		else if (name.equals(DAVConstants.LOCKTOKEN_ELEM))
		{
		    XMLElement hrefElem=subElem.getChild(DAVConstants.HREF_ELEM);
		    if (hrefElem!=null)
			lockToken=getChildAsString(hrefElem);
		}
	    }
	    Lock daLock=new Lock(ls, lt, depth, timeout, lockToken, owner);
	    file.addLock(daLock);
	}
    }

    private static void handleSource(DAVFile file, XMLElement elem)
    {
	for (ListIterator lit=elem.children();lit.hasNext();)
	{
	    Object subObj=lit.next();
	    if (!(subObj instanceof XMLElement))
		continue;
	    XMLElement linkElem=(XMLElement) subObj;
	    if (!linkElem.getElementName().equals(DAVConstants.LINK_ELEM))
	    {
		Debug.trace(PropertyHandler.class, Debug.DP3, "unrecognized child of source tag: "+linkElem);
		continue;
	    }
	    XMLElement srcElem=linkElem.getChild(DAVConstants.SRC_ELEM);
	    XMLElement dstElem=linkElem.getChild(DAVConstants.DST_ELEM);
	    if (srcElem!=null && dstElem!=null)
	    {
		Link daLink=new Link(getChildAsString(srcElem), getChildAsString(dstElem));
		file.addLink(daLink);
	    }	    
	}	
    }

    private static String getChildAsString(XMLElement elem)
    {
	Object o=elem.getChild(0);
	return (o==null) 
	    ? null 
	    : ((o instanceof XMLCData)
	       ? ((XMLCData)o).getData()
	       : o.toString());
    }	
}

/* $Log: PropertyHandler.java,v $
/* Revision 1.9  2001/07/17 03:01:22  smulloni
/* Handle CDATA in properties in a more appropriate manner (thanks to Jim
/* Whitehead for nudging me to do this).
/*
/* Revision 1.8  2001/06/13 01:09:03  smulloni
/* Some improvements to the network customizer (which is still buggy).
/* Also improvements for using the doAllprop switch, and for the layout of
/* the PropertiesDialog.
/*
/* Revision 1.7  2000/12/19 22:06:15  smulloni
/* adding documentation.
/*
/* Revision 1.6  2000/12/03 23:53:25  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.5  2000/11/09 23:34:53  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
