/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.border.TitledBorder;
import org.skunk.assertF.Assertion;
import org.skunk.dav.client.DAVConstants;
import org.skunk.trace.Debug;

/**
 * a dialog for opening a connection to a dav server.
 */
public class ConnectionDialog 
{
    private JComboBox dropdown;
    private JButton connectButton, cancelButton;
    private JDialog dialog;
    private Explorer ex;
    private boolean shouldConnect=false;  
    /**
     * constructs a ConnectionDialog for the given explorer
     * @param ex the explorer
     */
    public ConnectionDialog(Explorer ex)
    {
	this.ex=ex;
	initComponents();
    }

    private Frame getFrame()
    {
	Window w=SwingUtilities.windowForComponent(ex);
	return (w instanceof Frame)
	    ? (Frame) w
	    : null;
    }

    private void initComponents()
    {
	dialog=new JDialog(getFrame(), 
			   ResourceManager.getMessage(ResourceManager.CONNECTION_DIALOG_TITLE),
			   true);
	dropdown=new JComboBox();
	dropdown.setEditable(true);
	dropdown.setMinimumSize(new Dimension(660, dropdown.getPreferredSize().height));

	connectButton=new JButton(ResourceManager.getMessage(ResourceManager.CONNECT_KEY));
	connectButton.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    ConnectionDialog.this.shouldConnect=true;
		    dialog.setVisible(false);
		}
	    });

	cancelButton=new JButton(ResourceManager.getMessage(ResourceManager.CANCEL));
	cancelButton.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    ConnectionDialog.this.shouldConnect=false;
		    dialog.setVisible(false);
		}
	    });

	JPanel panel=new JPanel(new GridBagLayout());
	GridBagConstraints gbc=new GridBagConstraints();
	gbc.gridx=0;
	gbc.gridy=0;
	gbc.gridwidth=1;
	gbc.gridheight=1;
	gbc.insets=new Insets(4, 4, 4, 4);
	gbc.fill=GridBagConstraints.HORIZONTAL;
	gbc.anchor=GridBagConstraints.CENTER;
	gbc.ipadx=2;
	gbc.ipady=2;

	panel.add(dropdown, gbc);
	gbc.gridy++;
	
	Box buttonBox=Box.createHorizontalBox();
	buttonBox.add(Box.createHorizontalGlue());
	buttonBox.add(connectButton);
	buttonBox.add(Box.createHorizontalStrut(10));
	buttonBox.add(cancelButton);
	buttonBox.add(Box.createHorizontalGlue());
	panel.add(buttonBox, gbc);
	String borderTitle=ResourceManager.getMessage(ResourceManager.CONNECTION_DIALOG_PROMPT);
	panel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
							 borderTitle,
							 TitledBorder.LEFT, 
							 TitledBorder.TOP));

	dialog.setContentPane(panel);
	dialog.getRootPane().setDefaultButton(connectButton);
	dialog.pack();
    }

    /**
     * show the dialog.
     * @return whether or not the user wants to connect
     */
    public boolean show()
    {
	DefaultComboBoxModel model=new DefaultComboBoxModel((Vector)ServerData.getServers().clone());
	dropdown.setModel(model);
	int size=model.getSize();
	if (size>0) 
	    dropdown.setSelectedIndex(size-1);
	dialog.pack();
	dialog.setLocationRelativeTo(ex);
	shouldConnect=false;
	dialog.show();
	return shouldConnect;
    }

    /**
     * returns the ServerData chosen
     * @return the ServerData
     */
    public ServerData getServerData()
    {
	Object val=dropdown.getSelectedItem();
	if (val==null) 
	    return null;
	else if (val instanceof ServerData) 
	{
	    Debug.trace(this, Debug.DP5, "we have a ServerData: {0}", val);
	    return (ServerData) val;
	}
	else
	{
	    ServerData retSd=ServerData.getServer(val.toString().trim());
	    if (retSd==null)
		retSd=parseNewServer(val);
	    if (retSd!=null) 
		ExplorerMenuBar.refreshMenus();
	    return retSd;
	}
    }

    private ServerData parseNewServer(Object value)
    {
	Debug.trace(this, Debug.DP5, "in parseNewServer with value {0}", value);
	Debug.trace(this, Debug.DP5, "value class: {0}", value.getClass());
	String s=value.toString();
	boolean https=s.startsWith(DAVConstants.HTTPS);
	int port=80;
	String host=null;
	String initialPath=DAVConstants.DAV_FILE_SEPARATOR;
	int protocolPadding="://".length();
	if (https) 
	    s=s.substring(DAVConstants.HTTPS.length()+protocolPadding);
	else if (s.startsWith(DAVConstants.HTTP))
	    s=s.substring(DAVConstants.HTTP.length()+protocolPadding);
	int firstColon=s.indexOf(":");
	int firstSlash=s.indexOf(DAVConstants.DAV_FILE_SEPARATOR);
	/* 
	 *  possible valid states:
	 *  colon and slash:         www.webdav.org:80/
	 *  a colon, but no slash:   www.webdav.org:80
	 *  a slash, but no colon:   www.webdav.org/
	 *  no colon and no slash:   www.webdav.org
	 */
	int stateBit=((firstColon>0) ? 2 : 0)
	    + ((firstSlash>firstColon) ? 1 : 0);
	switch (stateBit)
	{	    
	case 3: //there is a port and a path. 
	    initialPath=s.substring(firstSlash);
	    //FALL THROUGH!
	case 2: //there is a port, no path
	    host=s.substring(0, firstColon);
	    try
	    {
		int portEnd=(stateBit==3) ? firstSlash : s.length();	
		port =Integer.parseInt(s.substring(firstColon+1, portEnd));	
	    }
	    catch (NumberFormatException nafta)
	    {
		doParseError();
		return null;
	    }
	    break;
	case 1: //there is a path, no port
	    host=s.substring(0, firstSlash);
	    initialPath=s.substring(firstSlash);
	    break;	    
	case 0: //there is neither path nor port
	    host=s;
	    break;
	}
	if (host==null)
	{
	    Assertion.assertF(false, "host is not null");
	    return null;
	}
	else
	    return new ServerData(https, 
				  host, 
				  port, 
				  initialPath, 
				  "", 
				  "");
    }

    private void doParseError()
    {
	String title=ResourceManager.getMessage(ResourceManager.ERROR_DIALOG_TITLE);
	String message=ResourceManager.getMessage(ResourceManager.CONNECTION_PARSE_ERROR_MESSAGE);
	JOptionPane.showMessageDialog(ex, message, title, JOptionPane.ERROR_MESSAGE);
    }
}

/* $Log: ConnectionDialog.java,v $
/* Revision 1.3  2001/02/12 05:01:12  smulloni
/* made the ConnectionDialog a tad more spacious
/*
/* Revision 1.2  2001/01/09 19:08:24  smulloni
/* adjusted the preferred size of the combo box.
/*
/* Revision 1.1  2001/01/05 08:01:12  smulloni
/* changes to the connection gui for the Explorer; added depth configurability to
/* propfind in the jpython test script; added an experimental "allprop" system
/* property which affects the propfind query type
/* */
