/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.io.IOException;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import org.skunk.assertF.Assertion;
import org.skunk.dav.client.DAVAuthenticator;
import org.skunk.dav.client.DAVConnection;
import org.skunk.dav.client.DAVConnectionPool;
import org.skunk.dav.client.DAVConstants;
import org.skunk.dav.client.DAVException;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.DAVProperty;
import org.skunk.dav.client.Depth;
import org.skunk.dav.client.method.CopyMethod;
import org.skunk.dav.client.method.DeleteMethod;
import org.skunk.dav.client.method.GetMethod;
import org.skunk.dav.client.method.LockMethod;
import org.skunk.dav.client.method.MkcolMethod;
import org.skunk.dav.client.method.MoveMethod;
import org.skunk.dav.client.method.PropFindMethod;
import org.skunk.dav.client.method.PropFindQueryType;
import org.skunk.dav.client.method.PropPatchMethod;
import org.skunk.dav.client.method.PutMethod;
import org.skunk.dav.client.method.UnlockMethod;
import org.skunk.trace.Debug;

public class DAVTreeModel extends DefaultTreeModel 
{ 

    private static boolean DO_ALLPROP;  
    static 
    {
	/*
	 * allProp is an experimental property, for testing.
	 * setting it to false may cause tree model queries to be more efficient,
	 * and enables the app to work with some recalcitrant dav server implementations
	 * that don't like allprop.  However, at present it breaks some other features.
	 */
	String doAllprop=System.getProperty("doAllprop");
	if (doAllprop!=null && doAllprop.equalsIgnoreCase("false"))
	    DO_ALLPROP=false;
	else DO_ALLPROP=true;
    }

    private boolean showNonCollections;

    public static final boolean usesAllprop()
    {
	return DAVTreeModel.DO_ALLPROP;
    }


    public DAVTreeModel(DefaultMutableTreeNode rootNode)
    {
	this(rootNode, false);
    }

    public DAVTreeModel(DefaultMutableTreeNode rootNode, boolean showNonCollections)
    {
	super(rootNode);
	this.showNonCollections=showNonCollections;

	Debug.trace(this, Debug.DP3, "in constructor: rootNode {0}, show non-collections: {1}",
		    new Object[] { rootNode, new Boolean(showNonCollections)});
    }

    public boolean getShowNonCollections()
    {
	return this.showNonCollections;
    }

    public boolean isLeaf(Object node)
    {
	if (node instanceof DAVTreeNode)
	{
	    return ((DAVTreeNode)node).isLeaf();
	}
	else if (node.equals(getRoot()))
	    return false;
	else 
	{
	    return super.isLeaf(node);
	}
    }

    public Object getChild(Object parent, int index)
    {
	testNode(parent);
	return super.getChild(parent, index);
    }

    public int getChildCount(Object parent)
    {
	testNode(parent);
	return super.getChildCount(parent);
    }

    public int getIndexOfChild(Object parent, Object child)
    {
	Debug.trace(this, Debug.DP6, 
		    "getting index of child for parent {0} and child {1}",
		    new Object[] {parent, child} );
	testNode(parent);
	return super.getIndexOfChild(parent, child);
    }

    /**
     *  if the node has not been refreshed and is a collection, refresh it
     */
    public void testNode(Object node)
    {
	if (node instanceof DAVTreeNode)
	{
	    DAVTreeNode davnode=(DAVTreeNode) node;
	    if (!(davnode.isRefreshed() || davnode.isLeaf()))
	    {
		refreshNode(davnode);
	    }
	}
    }

    private void reloadNode(final DAVTreeNode node)
    {
	BusinessTracker.addBusy("reload");
	Runnable r=new Runnable() {
		public void run()
		{
		    Debug.trace(this, Debug.DP3, "refreshing {0}", 
				new Object[] {node});
		    refreshNode(node); //why am I doing this in the event dispatch thread?
		    reload(node);
		    BusinessTracker.removeBusy("reload");
		}
	    };
	if (SwingUtilities.isEventDispatchThread())
	    r.run();
	else SwingUtilities.invokeLater(r);
    }

    public void refreshNode(DAVTreeNode node)
    {
	refreshNode(node, DO_ALLPROP);
    }

    public void refreshNode(DAVTreeNode node, boolean doAllprop)
    {
	BusinessTracker.addBusy("refresh");
	synchronized (node)
	{
	    String resource=node.getDAVFile().getName();
	    PropFindMethod method=(doAllprop) 
		? new PropFindMethod(resource)
		: new PropFindMethod(resource, Depth.ONE, DAVProperty.BASIC_PROPERTIES);

	    DAVFile newFile=null;
	    try
	    {
		DAVConnection conn=node.getDAVFile().getDAVConnection();
		conn.execute(method);
		newFile=method.getDAVFile();
	    }			
	    catch (DAVException davie)
	    {
		Debug.trace(this, Debug.DP2, davie);
	    }
	    catch (IOException oyVeh)
	    {
		Debug.trace(this, Debug.DP2, oyVeh);
	    }
	    if (newFile!=null) 
	    {
		Debug.trace(this, Debug.DP3, "refreshed file:\n"+newFile.toVerboseString());
		node.setDAVFile(newFile);
		node.setRefreshed(true);
		StateMonitor.setProperty(StateProperties.RELOAD, 
					 this,                   //reference to DAVTreeModel is application value
					 newFile.getFullName()); //full path of resource is domain key
	    }
	    else 
	    {
		Debug.trace(this, Debug.DP2, "refresh on {0} failed", new Object[] {node});
	    }
	    BusinessTracker.removeBusy("refresh");
	}
    }

    private void addConnection(DAVTreeNode node)
    {
	BusinessTracker.addBusy("addConn");
	Object root=getRoot();
	if (root instanceof DefaultMutableTreeNode)
	{
	    ((DefaultMutableTreeNode) root).add(node);
	    reload();
	}
	else 
	{
	    Debug.trace(this, Debug.DP2, 
			"Programming Error -- root node not a DefaultMutableTreeNode");
	}
	BusinessTracker.removeBusy("addConn");
    }

    public DAVTreeNode addConnectionNode(ServerData sd, boolean waitForIt, Runnable postRunner)
    {
	BusinessTracker.addBusy("addNode");
	String host=sd.getHost();
	int port=sd.getPort();
	String initialPath=sd.getInitialPath();
	Object nodeObj=getConnectionNode(host, port);	
	if (nodeObj!=null && nodeObj instanceof DAVTreeNode)
	{	    
	    DAVTreeNode possibleNode=(DAVTreeNode) nodeObj;
	    String nodePath=possibleNode.getDAVFile().getName();
	    Debug.trace(this, Debug.DP3, "possible node found, path is {0}",
			new Object[] {nodePath});
	    /*
	      four possibilities for the new initial path: 
	      1. the same initial path.  Do not connect.
	      2. a subpath.  Do not connect.
	      3. a superpath.  remove possibleNode and connect.
	      4. a parallel path.  Connect, leave possibleNode alone.
	    */
	    if (initialPath.startsWith(nodePath)) //cases 1 and 2
	    {
		if (!initialPath.equals(nodePath))
		{
		    String fullPath=new StringBuffer(host)
			.append(':')
			.append(port)
			.append(initialPath)
			.toString();
		    DAVTreeNode desiredNode=getNodeMatchingPath(fullPath);
		    BusinessTracker.removeBusy("addNode");
		    return desiredNode;
		    
		}
		BusinessTracker.removeBusy("addNode");
		return possibleNode;
	    }
	    if (nodePath.startsWith(initialPath)) //case 3
	    {
		removeNodeFromParent(possibleNode);	
	    }	      
	}	
	DAVAuthenticator auth=new DAVAuthenticatorImpl(sd);
	DAVTreeNode node=_addConnectionNode(sd.usesSSL(), 
					    host, 
					    port, 
					    initialPath, 
					    auth, 
					    waitForIt,
					    postRunner);
	BusinessTracker.removeBusy("addNode");
	return node;
    }

    public DAVTreeNode addConnectionNode(ServerData sd, Runnable postRunner)
    {
	return addConnectionNode(sd, false, postRunner);
    }

    private synchronized DAVTreeNode  _addConnectionNode(final boolean usingSSL,
						 final String host, 
						 final int port, 
						 final String initialPath, 
						 final DAVAuthenticator auth,
						 final boolean waitForIt,
						 final Runnable postRunner)
    {
	final DAVTreeNode retNode=null;
//	final DAVTreeNode retNode=null;   //     modified by sachiho
	Thread t=new Thread() {
		public void run() {

		    DAVConnection dc=DAVConnectionPool.getDAVConnection(host, port, usingSSL);
		    ExplorerApp.getAppContext().getConfigurator().configure(dc);
		    Debug.trace(this, Debug.DP4, "configured new connection: its timeout is {0}",
				new Integer(dc.getSocketTimeout()));
		    dc.setAuthenticator(auth);
		    PropFindMethod method;
		    if (DO_ALLPROP) 
			method=new PropFindMethod(initialPath);
		    else
			method=new PropFindMethod(initialPath, Depth.ONE, DAVProperty.BASIC_PROPERTIES);
		    try
		    {
			dc.execute(method);
			if (method.getStatus()>=400) 
			    return;
			DAVFile file=method.getDAVFile();

			final DAVTreeNode newNode=new DAVTreeNode(file, getShowNonCollections());
			newNode.setShowingHost(true); //display host, port and path
			final DAVTreeNode temp_retNode=newNode;
				    
			Runnable runner=new Runnable() 
			    {
				public void run() 
				{
				    addConnection(newNode);
				    if (postRunner!=null)
					postRunner.run();
				}
			    };

			if (!waitForIt)
			{
			    SwingUtilities.invokeLater(runner);
			}
			else 
			{
			    runner.run();
			}
		    }
		    catch (Exception ugh)
		    {
			final Exception ugh2=ugh;
			Debug.trace(this, Debug.DP2, ugh);
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
				    String message=ResourceManager
					.getMessage(ResourceManager.CONNECTION_ERROR_MESSAGE, 
						    new Object[] { new StringBuffer(host)
								   .append(':')
								   .append(port)
								   .toString(), 
								   ugh2.getMessage() } );
				    String title=ResourceManager
					.getMessage(ResourceManager.ERROR_DIALOG_TITLE);
				    JOptionPane.showMessageDialog(Explorer.guessActiveExplorer(),
								  message, 
								  title,
								  JOptionPane.ERROR_MESSAGE);
				}
			    });
		    }
		}
	    };
	if (waitForIt)
	    t.run();
	else
	    t.start();
	return retNode;
    }

    public void removeConnectionNode(String host, int port, String initialPath)
    {
	Object child=getConnectionNode(host, port, initialPath);
	if (child instanceof MutableTreeNode)
	    removeNodeFromParent((MutableTreeNode)child);
	DAVConnectionPool.removeDAVConnection(host, port);
    }

    public void removeConnectionNode(ServerData sd)
    {
	Assertion.assertF((sd!=null), "serverData is not null");
	removeConnectionNode(sd.getHost(), sd.getPort(), sd.getInitialPath());
    }

    private Object getConnectionNode(String host, int port, String initialPath)
    {
	Object root=getRoot();
	String connName=new StringBuffer(host)
	    .append(':')
	    .append(port)
	    .append(initialPath)
	    .toString();
	for (int i=0;i<getChildCount(root);i++)
	{
	    Object child=getChild(root, i);
	    if (child.toString().equals(connName))
	    {
		    return child;
	    }
	}
	return null;
    }

    private Object getConnectionNode(String host, int port)
    {
	Object root=getRoot();
	String connName=new StringBuffer(host)
	    .append(':')
	    .append(port)
	    .append('/')
	    .toString();
	for (int i=0;i<getChildCount(root);i++)
	{
	    Object child=getChild(root, i);
	    if (child.toString().startsWith(connName))
	    {
		    return child;
	    }
	}
	return null;
    }	

    public byte[] get(DAVFile file)
    {
	BusinessTracker.addBusy(file);
	DAVConnection dc=DAVConnectionPool.getDAVConnection(file.getHost(), file.getPort());
	GetMethod method=new GetMethod(file.getName());
	byte[] body=null;
	try
	{
	    dc.execute(method);
	    int status=method.getStatus();
	    Debug.trace(this, Debug.DP3, "status of response: {0}", 
			new Object[] {new Integer(status) });
	    if (status>=400)
	    {
		handleStatusError(file.getName(), status, method.getResponseBody());
	    }
	    else body=method.getResponseBody();
	}
	catch (Exception davie)
	{
	    Debug.trace(this, Debug.DP2, davie);
	    handleConnectionError(davie, file.getHost(), file.getPort());
	}
	BusinessTracker.removeBusy(file);
	return body;	
    }

    public void put(final DAVTreeNode parentNode,
		    final String fileName,
		    final byte[] bodyBytes)
    {
	put(parentNode, fileName, bodyBytes, null);
    }

    public void put(final DAVTreeNode parentNode, 
		    final String fileName, 
		    final byte[] bodyBytes,
		    final Runnable successRunner)
    {
	Assertion.assertF((parentNode!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {parentNode});
	BusinessTracker.addBusy("put");
	Thread t=new Thread() {
		public void run()
		{
		    ConnectStruct cs=getConnectStruct(parentNode, fileName);
		    String filePath=cs.getFilePath();
		    PutMethod method=new PutMethod(filePath, bodyBytes);
		    /* 
		     * look for a locktoken on the target resource, if it exists
		     * and we already know about it
		     */
		    DAVFile possibleTarget=null;
		    for (Iterator it=parentNode.getDAVFile().children();it.hasNext();)
		    {
			DAVFile kidFile=(DAVFile) it.next();
			if (kidFile.getFileName().equals(fileName))
			{
			    possibleTarget=kidFile;
			    break;
			}
		    }
		    if (possibleTarget!=null)
		    {
			if (possibleTarget.isCollection())
			{
			    //not allowed
			    String title=ResourceManager.getMessage(ResourceManager.ERROR_DIALOG_TITLE);
			    String message=ResourceManager.getMessage(ResourceManager.PUT_NOT_ALLOWED_MESSAGE);
			    JOptionPane.showMessageDialog(Explorer.guessActiveExplorer(), 
							  message,
							  title,
							  JOptionPane.ERROR_MESSAGE);
			    BusinessTracker.removeBusy("put");
			    return;
			}
			else if (possibleTarget.isLocked())
			{
			    method.setLockToken(possibleTarget.getExclusiveLockToken());
			}
		    }
		    try
		    {
			cs.getDAVConnection().execute(method);
			//check status
			int status=method.getStatus();
			Debug.trace(this, Debug.DP3, "status of response: "+status);
			if (status>=400)
			    handleStatusError(filePath, status, method.getResponseBody());
			else
			{
			    reloadNode(parentNode);
			    if (successRunner!=null)
				SwingUtilities.invokeLater(successRunner);
			    BusinessTracker.removeBusy("put");
			}
		    }
		    catch (Exception davie)
		    {
			Debug.trace(this, Debug.DP2, davie);
			handleConnectionError(davie, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    }

    public void lock(final DAVTreeNode parentNode, final DAVFile file, final String userInfo, final Runnable runnable)
    {
	Assertion.assertF((parentNode!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {parentNode});
	BusinessTracker.addBusy("lock");
	Thread t=new Thread() {
		public void run() {
		    ConnectStruct cs=getConnectStruct(parentNode, file.getFileName());
		    String filePath=cs.getFilePath();
		    LockMethod method=new LockMethod(filePath, userInfo);
		    try
		    {
			cs.getDAVConnection().execute(method);
			int status=method.getStatus();
			if (status>=400)
			    handleStatusError(filePath, status, method.getResponseBody());
			else 
			{
			    reloadNode(parentNode);
			    if (runnable!=null)
			    {
				SwingUtilities.invokeLater(runnable);
			    }
			    BusinessTracker.removeBusy("lock");
			}
		    }
		    catch (Exception e)
		    {
			handleConnectionError(e, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    }

    public void proppatch(final DAVTreeNode parentNode, final DAVFile file, final Map propertyValueMap)
    {
	Assertion.assertF((parentNode!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {parentNode});
	BusinessTracker.addBusy("proppatch");
	Thread t=new Thread() {
		public void run() {
		    ConnectStruct cs=getConnectStruct(parentNode, file.getFileName());
		    String filePath=cs.getFilePath();
		    PropPatchMethod method=new PropPatchMethod(filePath);
		    if (file.isExclusiveLocked())
		    {
			method.setLockToken(file.getExclusiveLockToken());
		    }
		    
		    for (Iterator it=propertyValueMap.keySet().iterator();it.hasNext();)
		    {
			Object o=it.next();
			Object val=propertyValueMap.get(o);
			if (!(o instanceof DAVProperty))
			{
			    BusinessTracker.removeAllBusy();
			    throw new IllegalArgumentException("propertyValueMap must map DAVProperty objects "
							       + "to Object values.  A null value indicates that "
							       + "the property should be removed.");
			}
			if (val!=null)
			{
			    method.putProperty((DAVProperty) o, val);
			}
			else method.removeProperty((DAVProperty) o);
		    }
		    try
		    {
			cs.getDAVConnection().execute(method);
			int status=method.getStatus();
			if (status>=400)
			    handleStatusError(filePath, status, method.getResponseBody());
			else if (status==207) //multistatus
			{
			    //check response body of children
			    DAVFile retFile = method.getDAVFile();
			    Integer statusObj=retFile.getStatus();
			    if (statusObj==null) 
			    {
				Debug.trace(this, Debug.DP2, "null status for "+ retFile);
			    }
			    else
			    {
				status=statusObj.intValue();
				if (status>=400)
				{
				    String s=retFile.getResponseDescription();
				    if (s==null) s="";
				    handleStatusError(retFile.getFullName(), 
						      status, 
						      s.getBytes());
				    return;
				}
			    }
			}
			reloadNode(parentNode);
			BusinessTracker.removeBusy("proppatch");
		    }
		    catch (Exception e)
		    {
			handleConnectionError(e, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    }

		    

    public void stealLock(final DAVTreeNode parentNode, final DAVFile file, final String userInfo)
    {
	Assertion.assertF((parentNode!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {parentNode});
	BusinessTracker.addBusy("steal");
	Thread t=new Thread() {
		public void run() {
		    ConnectStruct cs=getConnectStruct(parentNode, file.getFileName());
		    String filePath=cs.getFilePath();
		    String lockToken=file.getExclusiveLockToken();
		    Assertion.assertF((lockToken!=null),
				     "lockToken for {0} is not null",
				     new Object[] {file});		    
		    UnlockMethod unlockMethod=new UnlockMethod(filePath, lockToken);
		    LockMethod lockMethod=new LockMethod(filePath, userInfo);
		    try
		    {
			cs.getDAVConnection().execute(unlockMethod);
			int status=unlockMethod.getStatus();
			if (status>=400)
			{
			    if (status==403)
			    {
				//obtain server data for node; set its permitsLockStealing property
				ServerData.getServer(((DAVTreeNode)getPathToRoot(parentNode)[1]).toString())
				    .setPermitsLockStealing(false);	  
			    }
			    handleStatusError(filePath, status, unlockMethod.getResponseBody());
			}
			else 
			{
			    cs.getDAVConnection().execute(lockMethod);
			    status=lockMethod.getStatus();
			    if (status>=400)
				handleStatusError(filePath, status, lockMethod.getResponseBody());
			    else 
			    {
				reloadNode(parentNode);
				BusinessTracker.removeBusy("steal");
			    }
			}
		    }
		    catch (Exception e)
		    {
			handleConnectionError(e, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    }

    public void unlock(final DAVTreeNode parentNode, final DAVFile file)
    {
	unlock(parentNode, file, null);
    }

    public void unlock(final DAVTreeNode parentNode, final DAVFile file, final Runnable successRunner)
    {
	Assertion.assertF((parentNode!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {parentNode});
		    BusinessTracker.addBusy("unlock");
	Thread t=new Thread() {
		public void run() {
		    ConnectStruct cs=getConnectStruct(parentNode, file.getFileName());
		    String filePath=cs.getFilePath();
		    String lockToken=file.getExclusiveLockToken();
		    Assertion.assertF((lockToken!=null),
				     "lockToken for {0} is not null",
				     new Object[] {file});		    
		    UnlockMethod method=new UnlockMethod(filePath, lockToken);
		    try
		    {
			cs.getDAVConnection().execute(method);
			int status=method.getStatus();
			if (status>=400)
			    handleStatusError(filePath, status, method.getResponseBody());
			else 
			{
			    reloadNode(parentNode);
			    if (successRunner!=null) SwingUtilities.invokeLater(successRunner);
			    BusinessTracker.removeBusy("unlock");
			}
		    }
		    catch (Exception e)
		    {
			handleConnectionError(e, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    }

    public void delete(final DAVTreeNode parentNode, final String fileName)
    {
	Assertion.assertF((parentNode!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {parentNode});
	BusinessTracker.addBusy("delete");
	Thread t=new Thread() {
		public void run() {
		    ConnectStruct cs=getConnectStruct(parentNode, fileName);
		    String filePath=cs.getFilePath();
		    DeleteMethod method=new DeleteMethod(filePath);
		    try
		    {
			cs.getDAVConnection().execute(method);
			int status=method.getStatus();
			if (status>=400)
			    handleStatusError(filePath, status, method.getResponseBody());
			else 
			{
			    reloadNode(parentNode);
			    BusinessTracker.removeBusy("delete");
			}
		    }
		    catch (Exception e)
		    {
			handleConnectionError(e, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    }
 
    public void mkcol(final DAVTreeNode parentNode, final String collectionName)
    {
	Assertion.assertF((parentNode!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {parentNode});
	BusinessTracker.addBusy("mkcol");
	Thread t=new Thread() {
		public void run()
		{
		    ConnectStruct cs=getConnectStruct(parentNode, collectionName);
		    String filePath=cs.getFilePath();
		    MkcolMethod method=new MkcolMethod(filePath);
		    try
		    {
			cs.getDAVConnection().execute(method);
			//check status
			int status=method.getStatus();
			Debug.trace(this, Debug.DP3, "status of response: "+status);
			if (status>=400)
			    handleStatusError(filePath, status, method.getResponseBody());
			else 
			{
			    reloadNode(parentNode);
			    BusinessTracker.removeBusy("mkcol");
			}
		    }
		    catch (Exception davie)
		    {
			Debug.trace(this, Debug.DP2, davie);
			handleConnectionError(davie, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    } 

    protected String getURLForPath(TreePath teepee)
    {
	Object[] path=teepee.getPath();
	return _getURLForPath(path);
    }

    private String _getURLForPath(Object[] path)
    {
	if (path==null || path.length<2)
	{
	    return null;
	}
	String prefix=getProtocolForNode((DAVTreeNode)path[path.length-1]);
	StringBuffer sb=new StringBuffer(prefix);
	//discard root, which we don't show
	for (int i=1;i<path.length;i++)
	{
	    if (i>1) sb.append(DAVConstants.DAV_FILE_SEPARATOR);
	    sb.append(path[i]);
	}
	return sb.toString();
    }

    public DAVTreeNode getNodeMatchingPath(String path)
    {
 	if (path.startsWith(DAVConstants.DAV_FILE_SEPARATOR))
 	    path=path.substring(1);
	if (!path.endsWith(DAVConstants.DAV_FILE_SEPARATOR))
	    path=path+DAVConstants.DAV_FILE_SEPARATOR;
	return getNodeMatchingURL(path, getRoot(), false);
    }

    public DAVTreeNode getNodeMatchingURL(String url)
    {
	return getNodeMatchingURL(url, getRoot(), true);
    }

    protected DAVTreeNode getNodeMatchingURL(String url, Object nodeObj)
    {
	return getNodeMatchingURL(url, nodeObj, true);
    }

    private String getProtocolForNode(DAVTreeNode node)
    {
	if (node!=null)
	    return node.getDAVFile().getDAVConnection().getProtocol();
	return null;
    }
	

    protected DAVTreeNode getNodeMatchingURL(String url, 
					     Object nodeObj, 
					     boolean useHttpPrefix)
    {
	String fullname=null;
	if (nodeObj instanceof DAVTreeNode)
	{
	    DAVTreeNode dtn=(DAVTreeNode) nodeObj;
	    StringBuffer nameBuffer=new StringBuffer();
	    if (useHttpPrefix)
	    {
		nameBuffer.append(getProtocolForNode(dtn))
		    .append("://");
	    }		
	    fullname=nameBuffer.append(((DAVTreeNode)nodeObj).getDAVFile().getFullName()).toString();
	    Debug.trace(this, Debug.DP5, "testing node with fullname {0} against url {1}",
			new Object[] { fullname, url });
	    if (url.equals(fullname))
		return (DAVTreeNode)nodeObj;
	}
	if (nodeObj.equals(getRoot()) 
	    || (fullname!=null && url.startsWith(fullname))) //recurse into directory
	{
	    int cnt=getChildCount(nodeObj);
	    for (int i=0;i<cnt;i++)
	    {
		DAVTreeNode dtn=getNodeMatchingURL(url, getChild(nodeObj, i), useHttpPrefix);
		if (dtn!=null)
		    return dtn;
	    }
	}
	return null;
    }

    private void copyOrMove(final DAVTreeNode node, 
			    final DAVFile file, 
			    final String destinationURL, 
			    final boolean move)
    {
	Assertion.assertF((node!=null), 
			 "DAVTreeNode {0} is not null", 
			 new Object[] {node});
	BusinessTracker.addBusy("copyOrMove");
	int fileDivider=destinationURL.lastIndexOf(DAVConstants.DAV_FILE_SEPARATOR)+1;
	String destDir=destinationURL.substring(0, fileDivider);
	String destFileName=destinationURL.substring(fileDivider);
	Debug.trace(this, Debug.DP3, 
		    "destination directory is {0}, destination filename is {1}", 
		    new Object[] {destDir, destFileName});
	DAVTreeNode destNode=getNodeMatchingURL(destDir, getRoot());
	Debug.trace(this, Debug.DP3, "destNode: {0}",
		    new Object[] {destNode});
	if (destNode==null)
	{
	    Debug.trace(this, Debug.DP2, "destNode is null");
	    //should throw or handle an error -- TO BE DONE
	    BusinessTracker.removeBusy(this);
	    return;
	}
	DAVFile possiblyLocked=null;
	//see if the destination file already exists and has a lockToken
	for (Iterator it=destNode.getDAVFile().children();it.hasNext();)
	{
	    DAVFile kidFile=(DAVFile) it.next();
	    if (destFileName.equals(kidFile.getFileName()))
	    {
		possiblyLocked=kidFile;
		Debug.trace(this, Debug.DP3, "possiblyLocked file: {0}",
			    new Object[] {kidFile});
		break;
	    }
	}

	/*
	 * at this point, check whether lock owner is same as current user;
	 * if not, ??? TO BE DONE
	 */
	
	final String destinationLockToken=(possiblyLocked!=null) 
	    ? possiblyLocked.getExclusiveLockToken()
	    : null;

	Thread t=new Thread() {
		public void run()
		{
		    ConnectStruct cs=getConnectStruct(node, file.getFileName());
		    String filePath=cs.getFilePath();
		    CopyMethod method=(move) 
			? (new MoveMethod(filePath, destinationURL))
			: (new CopyMethod(filePath, destinationURL));
		    method.setSourceLockToken(file.getExclusiveLockToken());
		    method.setDestLockToken(destinationLockToken);
		    try
		    {
			cs.getDAVConnection().execute(method);
			//check status
			int status=method.getStatus();
			Debug.trace(this, Debug.DP3, "status of response: "+status);
			if (status>=400)
			    handleStatusError(filePath, status, method.getResponseBody());
			else 
			{
			    if (move) reloadNode(node);
			    int lastDex=destinationURL.lastIndexOf(DAVConstants.DAV_FILE_SEPARATOR)+1;
			    DAVTreeNode newNode=getNodeMatchingURL(destinationURL.substring(0, lastDex));
			    reloadNode(newNode);
			    BusinessTracker.removeBusy("copyOrMove");
			}
		    }
		    catch (Exception davie)
		    {
			Debug.trace(this, Debug.DP2, davie);
			handleConnectionError(davie, cs.getHost(), cs.getPort());
		    }
		}
	    };
	t.start();
    }

    public void copy(DAVTreeNode node, DAVFile file, String destinationURL)
    {
	copyOrMove(node, file, destinationURL, false);
    }
    

    public void move(DAVTreeNode node, DAVFile file, String destinationURL)
    {
	copyOrMove(node, file, destinationURL, true);
    }


    protected void handleConnectionError(final Exception ugh, final String host, final int port)
    {
	Debug.trace(this, Debug.DP2, ugh);
	Runnable r = new Runnable() {
		public void run() {
		    BusinessTracker.removeAllBusy();
		    String message=ResourceManager
			.getMessage(ResourceManager.CONNECTION_ERROR_MESSAGE, 
				    new Object[] { host+DAVConstants.DAV_FILE_SEPARATOR+port, 
						   ugh.getMessage() } );
		    String title=ResourceManager.getMessage(ResourceManager.ERROR_DIALOG_TITLE);
		    JOptionPane.showMessageDialog(Explorer.guessActiveExplorer(),
						  message, 
						  title,
						  JOptionPane.ERROR_MESSAGE);
		}
	    };
	if (SwingUtilities.isEventDispatchThread())
	    r.run();
	else SwingUtilities.invokeLater(r);
    }

    protected void handleStatusError(final String filePath, final int status, final byte[] bodyBytes)
    {
	Runnable r=new Runnable() {
		public void run() {
		    BusinessTracker.removeAllBusy();
		    String title=ResourceManager.getMessage(ResourceManager.ERROR_DIALOG_TITLE);
		    String message=ResourceManager.getMessage(ResourceManager.STATUS_ERROR_MESSAGE, 
							      new Object[] {filePath, 
									    new Integer(status), 
									    new String(bodyBytes)});
		    JOptionPane.showMessageDialog(Explorer.guessActiveExplorer(),
						  message, 
						  title,
						  JOptionPane.ERROR_MESSAGE);
		}
	    };
	if (SwingUtilities.isEventDispatchThread())
	    r.run();
	else SwingUtilities.invokeLater(r);
    }

    private void handleInvalidDestinationError(final String destination)
    {
	Runnable r=new Runnable() {
		public void run() {
		    BusinessTracker.removeAllBusy();
		    String title=ResourceManager.getMessage(ResourceManager.ERROR_DIALOG_TITLE);
		    String message=ResourceManager.getMessage(ResourceManager.INVALID_DESTINATION_ERROR_MESSAGE,
							      new Object[] {destination});
		    JOptionPane.showMessageDialog(Explorer.guessActiveExplorer(),
						  message,
						  title,
						  JOptionPane.ERROR_MESSAGE);
		}
	    };
	if (SwingUtilities.isEventDispatchThread())
	    r.run();
	else SwingUtilities.invokeLater(r);
    }	

    private ConnectStruct getConnectStruct(DAVTreeNode parentNode, String fileName)
    {
	DAVFile file=parentNode.getDAVFile();
	String host=file.getHost();
	int port=file.getPort();
	String parentName=file.getName();
	StringBuffer sb=new StringBuffer(parentName);
	if (!parentName.endsWith(DAVConstants.DAV_FILE_SEPARATOR)) 
	    sb.append(DAVConstants.DAV_FILE_SEPARATOR);
	sb.append(fileName);
	DAVConnection dc=DAVConnectionPool.getDAVConnection(host, port);
	return new ConnectStruct(dc, sb.toString(), host, port);
    }

    private class ConnectStruct
    {
	private DAVConnection dc;
	private String filePath;
	private String host;
	private int port;

	public DAVConnection getDAVConnection() { return dc; }
	public String getFilePath() { return filePath; }
	public String getHost() { return host; }
	public int getPort() { return port; }

	ConnectStruct(DAVConnection dc, String filePath, String host, int port)
	{
	    this.dc=dc;
	    this.filePath=filePath;
	    this.host=host;
	    this.port=port;
	}
    }
}

/* $Log: DAVTreeModel.java,v $
/* Revision 1.34  2001/07/25 21:17:39  smulloni
/* fixed typo in error message.
/*
/* Revision 1.33  2001/06/13 01:09:03  smulloni
/* Some improvements to the network customizer (which is still buggy).
/* Also improvements for using the doAllprop switch, and for the layout of
/* the PropertiesDialog.
/*
/* Revision 1.32  2001/06/05 01:09:53  smulloni
/* fixed bug wtih failed connection attempts, which used to set the
/* connection flag as if the connection had succeeded.
/*
/* Revision 1.31  2001/01/23 20:50:09  smulloni
/* added configurable socket timeout
/*
/* Revision 1.30  2001/01/05 08:01:12  smulloni
/* changes to the connection gui for the Explorer; added depth configurability to
/* propfind in the jpython test script; added an experimental "allprop" system
/* property which affects the propfind query type
/*
/* Revision 1.29  2001/01/03 00:30:35  smulloni
/* a number of modifications along the way to replacing JFileChooser with
/* something more suitable for remote (virtual) files.
/*
/* Revision 1.28  2001/01/02 17:34:21  smulloni
/* added the ability to create DAVTreeModels that show non-directories.  This
/* permits me to use one with an org.skunk.swing.TreeNodeChooser.
/*
/* Revision 1.27  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.26  2000/12/18 23:22:47  smulloni
/* added optional SSL capability to the gui application.  Separate make targets
/* exist for building it will SSL support.
/*
/* Revision 1.25  2000/12/04 23:51:16  smulloni
/* added ImageViewer; fixed word in SimpleTextEditor
/*
/* Revision 1.24  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.23  2000/11/28 19:36:19  smullyan
/* some bug fixes; notes.
/*
/* Revision 1.22  2000/11/23 04:37:31  smullyan
/* editor and explorer now synch their files using the StateMonitor, which has
/* been improved significantly
/*
/* Revision 1.21  2000/11/22 00:11:26  smullyan
/* editor now locks and unlocks, more or less appropriately.
/*
/* Revision 1.20  2000/11/20 23:30:19  smullyan
/* more editor integration work.
/*
/* Revision 1.19  2000/11/17 20:25:04  smullyan
/* new SaveAction; a StateMonitor being added to handle application state.
/*
/* Revision 1.18  2000/11/14 23:17:08  smullyan
/* more improvements to the PropertiesDialog.
/*
/* Revision 1.17  2000/11/13 23:28:30  smullyan
/* first pass at a gui for proppatch, added to the propertiese dialog.  Highly
/* problematic still, needs substantial work.
/*
/* Revision 1.16  2000/11/10 22:40:06  smullyan
/* added icon to table for resource type; fixes to copy and move; disabling of
/* menu items that are inappropriate.
/*
/* Revision 1.15  2000/11/09 23:34:55  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
