/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Component;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Map;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JTree;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.event.TableModelListener;
import javax.swing.event.TreeModelListener;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import org.skunk.assertF.Assertion;
import org.skunk.config.Configurator;
import org.skunk.dav.client.DAVConstants;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.Lock;
import org.skunk.dav.client.gui.editor.DAVEditor;
import org.skunk.dav.client.gui.editor.EditEvent;
import org.skunk.dav.client.gui.editor.EditListener;
import org.skunk.swing.TreeNodeChooser;
import org.skunk.trace.Debug;

/**
 * An embeddable widget that presents conjoined 
 * tree and table views of a DAV file hierarchy.
 *
 * @author Jacob Smullyan
 * @version $Version$
 */
public class Explorer extends JPanel implements Buffer, EditListener
{
    static final long serialVersionUID = 3129234418500414985L;

    private static ArrayList allExplorers;
    private DefaultMutableTreeNode rootNode;
    private DAVTreeModel treeModel;
    private DAVTableModel tableModel;
    private JTable table;
    private JTree tree;

    /*
     *
     *configurable properties --default values
     *
     */

    private DAVTableHeader[] tableHeaders=new DAVTableHeader[]
    { 
	DAVTableHeader.RESOURCETYPE,
	DAVTableHeader.LOCK,
	DAVTableHeader.FILE_NAME, 
	DAVTableHeader.CONTENT_TYPE,
	DAVTableHeader.CONTENT_LENGTH,
	DAVTableHeader.LAST_MODIFIED,
	DAVTableHeader.OWNER
    };

    private boolean showTableVerticalLines=true;
    private boolean showTableHorizontalLines=true;

    public static final String TABLE_HEADERS_PROPERTY="tableHeaders";

    static
    {
	allExplorers=new ArrayList();
    }

    public static synchronized Iterator explorers()
    {
	return ((ArrayList)allExplorers.clone()).iterator();
    }

    public static int getExplorerCount()
    {
	return allExplorers.size();
    }

    public static Explorer guessActiveExplorer()
    {
	for (ListIterator lit=allExplorers.listIterator();lit.hasNext();)
	{
	    Explorer ex=(Explorer) lit.next();
	    //if the explorer is contained by something 
	    //which in turn contains something that has focus, 
	    //return it
	    if (SwingUtilities.findFocusOwner(SwingUtilities.getRoot(ex))!=null)
		return ex;
	}
	return null;
    }

    public Explorer()
    {
	super();
	initComponents();
	allExplorers.add(this);
    }

    /* IMPLEMENTATION OF Buffer INTERFACE */

    public String getName()
    {
	return ResourceManager.getMessage(ResourceManager.EXPLORER_NAME);
    }

    public JComponent getComponent()
    {
	return this;
    }

    public void docking() 
    {
    }

    public void undocking()
    {
    }

    /* END OF Buffer IMPLEMENTATION */

    /* IMPLEMENTATION OF EditListener INTERFACE */

    public void editPerformed(EditEvent event)
    {
	int code=event.getEditCode();
	final DAVEditor editor=(DAVEditor)event.getSource();
	Runnable successRunner;
	switch (code)
	{
	case EditEvent.CLOSE:
	    Debug.trace(this, Debug.DP3, "received EditEvent.CLOSE");
	    //will probably want to unlock, but that should be configurable
	    break;
	case EditEvent.SAVE:	    
	    Debug.trace(this, Debug.DP3, "received EditEvent.SAVE");

	    successRunner=new Runnable()
		{
		    public void run()
		    {
			editor.setDirty(false);
		    }
		};
	    createFile(editor.getResourceName(), editor.getResourceBody(), successRunner);
	    editor.setDirty(false);
	    break;
	case EditEvent.SAVE_AS:
	    Debug.trace(this, Debug.DP3, "received EditEvent.SAVE_AS");
	    //get old file info, so we can perform an unlock
	    DAVFile oldFile=editor.getDAVFile();
	    DAVTreeNode oldParentNode=getNodeForFile(oldFile);
	    //get save target
	    String title=ResourceManager.getMessage(ResourceManager.SAVE_AS_KEY);
	    DAVFileChooser.ChoiceStruct choice=chooseFile(oldParentNode, 
							  title,
							  true,
							  oldFile.getFileName());
	    if (choice==null)
	    {
		Debug.trace(this, Debug.DP3, "user cancelled save as");
		return;
	    }
	    final DAVTreeNode parentNode=choice.getParentNode();
	    Assertion.assertF((parentNode!=null), "parent node is not null");
	    final String newFileName=choice.getFilename();	    
	    if (newFileName==null)
	    {
		Debug.trace(this, Debug.DP3, "user didn't select anything");
		return;
	    }	 
	    final String newPath=choice.getFilePath();
	    successRunner=new Runnable()
		{
		    public void run()
		    {
			Debug.trace(this, Debug.DP3, "in runnable");
			refreshNode(parentNode);
			DAVFile parentFile=parentNode.getDAVFile();
			//DAVFile's getChildNamed() method doesn't like a host name in the path
			String pathWithoutHost=parentFile.getName()+newFileName;
			final DAVFile newDAVFile=parentFile.getChildNamed(pathWithoutHost);
			if (newDAVFile==null) 
			{
			    Debug.trace(this, Debug.DP1, "*** ERROR: DAVFile not found with path "+newPath);
			    return;
			}
			editor.setDAVFile(newDAVFile);
			editor.setDirty(false);
			editor.setResourceName(newDAVFile.getFullName());
			editor.setName(newFileName);
			Debug.trace(this, Debug.DP3, 
				    "editor is {0}, its file is {1}, its dirtiness is {2}, its name is {3}",
				    new Object[] { editor, editor.getDAVFile(), new Boolean(editor.isDirty()), 
						       editor.getName() });
			//lock the new file for editing
			Explorer.this.lock(parentNode, newDAVFile);
			ExplorerApp.getViewForBuffer(editor).dock(editor);
		    }
		};
	    //save new file
	    createFile(newPath, editor.getResourceBody(), successRunner);
	    //unlock old file
	    unlock(oldParentNode, oldFile);	    
	    break;
	default:
	    Assertion.assertF(false, 
			     "this default case of the switch on editCode should not be reached");
	}
    }
    
    /* END OF EditListener IMPLEMENTATION */


    private void initComponents()
    {
	//root node is not displayed.
	rootNode=new DefaultMutableTreeNode("root");
	treeModel=new DAVTreeModel(rootNode);
	treeModel.addTreeModelListener(new DAVTreeModelListener(this));
	tree=new JTree(treeModel);

	//setting largeModel to be true produces a noticeable performance enhancement (Metal L&F, IBM Linux 1.3)
	tree.setLargeModel(true);

	//in order to enable tooltips, use custom renderer
	tree.setCellRenderer(new DAVTreeCellRenderer());
	//tree must be registered with ToolTipManager for this to work
	ToolTipManager.sharedInstance().registerComponent(tree);
	tree.setRootVisible(false);
	tree.setShowsRootHandles(true);

	tableModel=new DAVTableModel(tableHeaders);

	/*
	  DAVTreeSelectionListener needs a references to the table model in order in for the
	  table to respond to tree events
	*/
	tree.addTreeSelectionListener(new DAVTreeSelectionListener(treeModel, tableModel));

	table=new JTable(tableModel);
	addTableColumns();

	table.setShowVerticalLines(getShowHorizontalLines());
	table.setShowHorizontalLines(getShowHorizontalLines());
	TableMouser mouseCatcher=new TableMouser(this);
	table.addMouseListener(mouseCatcher);
	table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	table.setDefaultRenderer(Lock.class, new LockRenderer());
	table.getTableHeader().setReorderingAllowed(false);
	
	JScrollPane treePane=new JScrollPane(tree, 
					     JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
					     JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
	//ensure that the tree gets enough room initially in the splitpane
	Dimension treeD=tree.getPreferredSize();
	treeD.width=Math.max(treeD.width, 200);
	treePane.setPreferredSize(treeD);

	JScrollPane tablePane=new JScrollPane(table);
	tablePane.addMouseListener(mouseCatcher);
	JSplitPane splitPane=new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, 
					    treePane, 
					    tablePane);
	splitPane.setPreferredSize(new Dimension(800, 600));
	setLayout(new BorderLayout());
	add(splitPane, BorderLayout.CENTER);
    }

    private void addTableColumns()
    {
	table.createDefaultColumnsFromModel();
	//cheesy, despicable, doomed.
	int doneCanAbort=0;
	for (Enumeration enum=table.getColumnModel().getColumns();enum.hasMoreElements();)
	{
	    TableColumn tc=(TableColumn) enum.nextElement();
	    Object identifier = tc.getIdentifier();
	    if (identifier.equals(DAVTableHeader.LOCK.toString()))
	    {
		setColumnForIcon(tc, ResourceManager.getImageIcon(ResourceManager.LOCK_ICON, null));
		doneCanAbort++;
	    }
	    else if (identifier.equals(DAVTableHeader.CONTENT_LENGTH.toString()))
	    {
		tc.setPreferredWidth(40);
		doneCanAbort++;
	    }
	    else if (identifier.equals(DAVTableHeader.RESOURCETYPE.toString()))
	    {
		setColumnForIcon(tc, ResourceManager.getImageIcon(ResourceManager.FOLDER_ICON, null));
		tc.setCellRenderer(new ResourceTypeRenderer());
		doneCanAbort++;
	    }
	    if (doneCanAbort==3) return;
	}
    }

    private void setColumnForIcon(TableColumn tc, ImageIcon ii)
    {
	int width=ii.getIconWidth();
	tc.setMaxWidth(width);
	tc.setPreferredWidth(width);
	tc.setMinWidth(width);
	tc.setResizable(false);
	tc.setHeaderValue("");
    }	

    /**
     * @return array of table headers
     */
    public DAVTableHeader[] getTableHeaders()
    {
	return this.tableHeaders;
    }  

    public void setTableHeaders(DAVTableHeader[] tableHeaders)
    {
	this.tableHeaders=tableHeaders;
	/*
	 *  it is simpler to use side-effects in the mutators of 
	 *  configurable properties rather than PropertyChangeEvents.
	 *  These properties only exist so they can be configured
	 */
	tableModel.setTableHeaders(tableHeaders);
	addTableColumns();
    }

    public boolean getShowVerticalLines()
    {
	return this.showTableVerticalLines;
    }

    public void setShowVerticalLines(boolean showVerticalLines)
    {
	this.showTableVerticalLines=showVerticalLines;
	table.setShowVerticalLines(showVerticalLines);
    }

    public boolean getShowHorizontalLines()
    {
	return this.showTableHorizontalLines;
    }

    public void setShowHorizontalLines(boolean showHorizontalLines)
    {
	this.showTableHorizontalLines=showHorizontalLines;
	table.setShowHorizontalLines(showHorizontalLines);
    }

    /**
     * tries to connect to the server <code>sd</code> and 
     * displays a corresponding tree node in the explorer.  
     * If the attempt succeeded, the <code>Runnable</code> <code>postRunner</code>
     * will then be executed.
     * @param sd the <code>ServerData</code> representing the target server
     * @param postRunner the <code>Runnable</a> to be executed upon success
     */
    public void addConnectionNode(ServerData sd, Runnable postRunner)
    {
	selectNode(treeModel.addConnectionNode(sd, postRunner));

    }

    public void removeConnectionNode(String host, int port, String initialPath)
    {
	treeModel.removeConnectionNode(host, port, initialPath);
	tree.clearSelection();
    }

    public void removeConnectionNode(ServerData sd)
    {
	treeModel.removeConnectionNode(sd);
	tree.clearSelection();
    }

    public int getConnectedServerCount()
    {
	return treeModel.getChildCount(rootNode);
    }

    public ServerData[] getConnectedServers()
    {
	int numServers=getConnectedServerCount();
	ServerData[] servers=new ServerData[numServers];
	for (int i=0;i< getConnectedServerCount();i++)
	{
	    Object nodeObj=treeModel.getChild(rootNode, i);
	    if (nodeObj instanceof DAVTreeNode)
		servers[i]=ServerData.getServer(nodeObj.toString());
	}
	return servers;
    }

    public boolean isConnected(ServerData sd)
    {
	ServerData[] servers=getConnectedServers();
	for (int i=0;i<servers.length;i++)
	{
	    if (servers[i].equals(sd))
		return true;
	}
	return false;
    }

    public byte[] get(DAVFile file)
    {
	return treeModel.get(file);	
    }

    public void delete(DAVTreeNode parentNode, String fileName)
    {
	treeModel.delete(parentNode, fileName);
	selectNode(parentNode);
    }

    public void lock(DAVTreeNode parentNode, DAVFile file)
    {
	lock(parentNode, file, null);
    }

    public void lock(DAVTreeNode parentNode, DAVFile file, Runnable runnable)
    {
	ServerData sd=getConnection(parentNode);
	Assertion.assertF((sd!=null), "node {0} has non-null ServerData", 
			 new Object[] {parentNode});
	String owner=sd.getOwner();
	if (owner==null) 
	{
	    owner=promptForLockOwner(sd);
	    if (owner==null)
		return;
	}
	treeModel.lock(parentNode, file, owner, runnable);
	selectNode(parentNode);
    }

    private String promptForLockOwner(ServerData sd)
    {
	String owner;
	String message=ResourceManager.getMessage(ResourceManager.LOCK_OWNER_PROMPT, 
						  new Object[] {sd.getHost()});
	String title=ResourceManager.getMessage(ResourceManager.LOCK_OWNER_DIALOG_TITLE);
	while (true)
	{
	    Object result=JOptionPane.showInputDialog(this, 
						      message, 
						      title, 
						      JOptionPane.QUESTION_MESSAGE,
						      null,
						      null,
						      sd.getUsername());
	    if (result==null)
		return null;
	    else 
	    {
		owner=result.toString().trim();
		if (owner.equals(""))
		    continue;
		else break;
	    }
	}
	sd.setOwner(owner);
	try
	{
	    ServerData.saveServers();
	}
	catch (IOException oyVeh)
	{
	    Debug.trace(this, Debug.DP1, oyVeh);
	}
	return owner;
    }

    public void unlock(DAVTreeNode parentNode, DAVFile file)
    {
	treeModel.unlock(parentNode, file);
	selectNode(parentNode);
    }

    public void stealLock(DAVTreeNode parentNode, DAVFile file)
    {
	ServerData sd=getConnection(parentNode);
	Assertion.assertF((sd!=null), "node {0} has non-null ServerData", 
			 new Object[] {parentNode});
	String owner=sd.getOwner();
	if (owner==null) 
	    owner=promptForLockOwner(sd);
	if (owner==null)
	    return;
	treeModel.stealLock(parentNode, file, owner);
	selectNode(parentNode);
    }	

    public void proppatch(DAVTreeNode parentNode, DAVFile file, Map propertyValueMap)
    {
	treeModel.proppatch(parentNode, file, propertyValueMap);
    }

    public void copy(DAVTreeNode parentNode, DAVFile file, String destinationURL)
    {
	treeModel.copy(parentNode, file, destinationURL);
	int lastDex=destinationURL.lastIndexOf(DAVConstants.DAV_FILE_SEPARATOR)+1;
	selectNode(treeModel.getNodeMatchingURL(destinationURL.substring(0, lastDex)));
    }

    public void move(DAVTreeNode parentNode, DAVFile file, String destinationURL)
    {
	treeModel.move(parentNode, file, destinationURL);
	int lastDex=destinationURL.lastIndexOf(DAVConstants.DAV_FILE_SEPARATOR)+1;
	selectNode(treeModel.getNodeMatchingURL(destinationURL.substring(0, lastDex)));
    }

    public ServerData getConnection(DAVTreeNode node)
    {
	Debug.trace(this, Debug.DP3, "node for getConnection() is {0}",
		    new Object[] {node} );
	Object[] path=treeModel.getPathToRoot(node);
	return _getConnection(path);
    }

    private ServerData _getConnection(Object[] path)
    {
	if (path==null) return null;
	if (path.length>1) //first element in path is root
	{
	    Object serverNodeObj=path[1];
	    if (serverNodeObj instanceof DAVTreeNode)
	    {
		DAVTreeNode serverNode=(DAVTreeNode) serverNodeObj;
		ServerData foundSD=ServerData.getServer(serverNode.toString());
		Debug.trace(this, Debug.DP3, "ServerData found is "+foundSD);
		return foundSD;
	    }		
	}
	Debug.trace(this, Debug.DP3, "returning null from _getConnection");
	return null;
    }

    public ServerData getSelectedConnection()
    {
	TreePath teepee=tree.getSelectionPath();
	Debug.trace(this, Debug.DP3, "selected TreePath is "+teepee);
	if (teepee!=null) 
	{
	    Object[] path=teepee.getPath();
	    if (Debug.isDebug(this, Debug.DP4))
	    {
		StringBuffer sb=new StringBuffer("path returned by treepath:\n");
		for (int i=0;i<path.length;i++)
		    sb.append("\t").append(path[i]).append("\n");
		Debug.trace(this, Debug.DP3, sb.toString());			
	    }
	    return _getConnection(path);
	}
	Debug.trace(this, Debug.DP3, "returning null from getSelectedConnection()");
	return null;
    }

    public void selectLastNode()
    {
	TreePath path=null;
	if (rootNode.getChildCount()>0)
	{
	    path=new TreePath(((DefaultMutableTreeNode)rootNode.getLastChild()).getPath());
	}
	else 
	{
	    path=new TreePath(rootNode.getPath());
	}
	Debug.trace(this, Debug.DP3, "selecting path "+path);
	tree.setSelectionPath(path);
    }

    public void selectNode(DAVTreeNode node)
    {
	Debug.trace(this, Debug.DP3, "in selectNode with node " + node);
	if (node!=null)
	    tree.setSelectionPath(new TreePath(treeModel.getPathToRoot(node)));
    }

    public void displayCollection(String filename)
    {
	DAVTreeNode parentNode=getSelectedNode();
	if (parentNode!=null)
	
	    displayCollection(parentNode, filename);
    }



    /**
     * displays a given collection in the explorer.
     * @param parentNode the node of the containing collection
     * @param filename the name of the collection resource to display
     */
    public void displayCollection(DAVTreeNode parentNode, String filename)
    {
	DAVTreeNode childNode=getChildNode(parentNode, filename);
	if (childNode==null)
	{
	    Debug.trace(this, Debug.DP2, 
			"filename {0} not found under node {1}", 
			new Object[] {filename, parentNode});
	    return;
	}
	TreeNode[] nodeArray=treeModel.getPathToRoot(childNode);
	if (Debug.isDebug(this, Debug.DP5))
	{
	    StringBuffer sb=new StringBuffer("path returned by model for node ")
		.append(childNode)
		.append(":\n");
	    for (int i=0;i<nodeArray.length;i++)
		sb.append("\t")
		    .append(nodeArray[i])
		    .append("\n");
	    Debug.trace(this, Debug.DP5, sb.toString());			
	}
	TreePath newPath=new TreePath(nodeArray);
	tree.makeVisible(newPath);
	tree.scrollPathToVisible(newPath);
	tree.setSelectionPath(newPath);
    }

    protected DAVTreeNode getChildNode(DAVTreeNode parentNode, String filename)
    {
	for (Enumeration enum=parentNode.children();enum.hasMoreElements();)
	{
	    Object nextObj=enum.nextElement();
	    Assertion.assertF((nextObj instanceof DAVTreeNode), 
			  nextObj + " is a DAVTreeNode");
	    DAVTreeNode childNode=(DAVTreeNode) nextObj;
	    DAVFile file=childNode.getDAVFile();
	    if (file!=null 
		&& file.isCollection() 
		&& filename.equals(file.getFileName()))
	    {
		return childNode;
	    }
	}
	return null;
    }

    public DAVTreeNode getNodeForFile(DAVFile file)
    {
	String fullPath=file.getFullName();
	String directoryPath=(file.isCollection()) 
	    ? fullPath
	    : fullPath.substring(0, fullPath.lastIndexOf(DAVConstants.DAV_FILE_SEPARATOR));
	Debug.trace(this, Debug.DP3, "directory path: "+ directoryPath);
	DAVTreeNode node=treeModel.getNodeMatchingPath(directoryPath);
	Debug.trace(this, Debug.DP3, "returning node "+node);
	return node;
    }

    public void createCollection(DAVTreeNode parentNode, String collectionName)
    {
	treeModel.mkcol(parentNode, collectionName);
	selectNode(parentNode);
    }

    public void createFile(String fullPath, byte[] bodyBytes)
    {
	createFile(fullPath, bodyBytes, null);
    }

    public void createFile(String fullPath, byte[] bodyBytes, Runnable successRunner)
    {
	int lastSep=fullPath.lastIndexOf(DAVConstants.DAV_FILE_SEPARATOR);
	String directoryPath=fullPath.substring(0, lastSep);
	String filename=fullPath.substring(lastSep+1);
	DAVTreeNode node=treeModel.getNodeMatchingPath(directoryPath);
	if (node!=null)
	{
	    createFile(node, filename, bodyBytes, successRunner);
	}
	else
	{
	    Debug.trace(this, Debug.DP2, 
			"no node found for path {0}", 
			new Object[] {fullPath});
	}
    }

    public void createFile(DAVTreeNode parentNode, String fileName)
    {
	createFile(parentNode, fileName, new byte[0], null);
    }

    public void createFile(DAVTreeNode parentNode, String fileName, Runnable successRunner)
    {
	createFile(parentNode, fileName, new byte[0], successRunner);
    }

    public void createFile(DAVTreeNode parentNode, String fileName, byte[] bodyBytes)
    {
	createFile(parentNode, fileName, bodyBytes, null);
    }

    public void createFile(DAVTreeNode parentNode, String fileName, byte[] bodyBytes, Runnable successRunner)
    {
	treeModel.put(parentNode, fileName, bodyBytes, successRunner);
	selectNode(parentNode);
    }

    protected void setTableFile(DAVFile file)
    {
	tableModel.setDAVFile(file);
	table.clearSelection();
    }

    public DAVFile getCurrentFile()
    {
	if (table.getSelectedRow()!=-1)
	    return getSelectedTableFile();
	DAVTreeNode node=getSelectedNode();
	if (node==null) return null;
	else return node.getDAVFile();
    }

    public DAVFile getSelectedTableFile()
    {
	return tableModel.getDAVFile(table.getSelectedRow());
    }

    public DAVFile getFile(DAVTreeNode parentNode, String filename)
    {
	for (Iterator it=parentNode.getDAVFile().children();it.hasNext();)
	{
	    DAVFile kidFile=(DAVFile)it.next();
	    if (kidFile.getFileName().equals(filename))
		return kidFile;
	}
	return null;
    }

    public DAVTreeNode getSelectedNode()
    {
	TreePath currentPath=tree.getSelectionPath();
	if (currentPath==null) return null;
	Object lastObj=currentPath.getLastPathComponent();
	Debug.trace(this, Debug.DP3, "current node is {0}", new Object[] {lastObj});
	return (lastObj instanceof DAVTreeNode) ? (DAVTreeNode) lastObj : null;
    }

    public void refreshNode(DAVTreeNode node)
    {
	Assertion.assertF((node!=null), "node does not equal null");
	treeModel.refreshNode(node);
	if (node.equals(getSelectedNode()))
	    setTableFile(node.getDAVFile());
    }

    public void refreshNode(DAVTreeNode node, boolean useAllprop)
    {
	Assertion.assertF((node!=null), "node does not equal null");
	treeModel.refreshNode(node, useAllprop);
	if (node.equals(getSelectedNode()))
	    setTableFile(node.getDAVFile());
    }

    /**
     * utility method for showing a DAVFileChooser
     * @param node a node on the DAVServer where the user may choose a file
     * @param title the title of the choice dialog
     * @param editable whether the user may freely enter a new filename
     * @param initialFileName the initial contents of the filename entry field
     * @return a struct of information about the chosen file, or null
     *
     */
    public DAVFileChooser.ChoiceStruct chooseFile(DAVTreeNode node, 
						  String title, 
						  boolean editable, 
						  String initialFileName)
    {
	DAVFileChooser chooser=(node==null)
	    ? getDAVFileChooser()
	    : getDAVFileChooser(node);
	chooser.setSelectionMode(TreeNodeChooser.SelectionMode.LEAF_ONLY);
	if (initialFileName!=null)
	{
	    chooser.setEntryFieldText(initialFileName);
	    chooser.setEntryFieldTextSticky(true);
	}
	chooser.setEntryFieldEditable(editable);
	    
	//configure the chooser's labels -- TO BE DONE

	//I should use a JDialog and roll my own, 
	//probably in the TreeNodeChooser class (an instance showDialog() method)
	int option=JOptionPane.showConfirmDialog(this, 
						 chooser, 
						 title, 
						 JOptionPane.OK_CANCEL_OPTION,
						 JOptionPane.QUESTION_MESSAGE,
						 new ImageIcon());
	if (option!=JOptionPane.OK_OPTION)
	{
	    Debug.trace(this, Debug.DP3, "user cancelled ");
	    return null;
	}
	DAVFileChooser.ChoiceStruct struct=chooser.getChoiceStruct();
	Object[] retStruct=new Object[4];
	Object lastPathComponent=chooser.getSelectedPath().getLastPathComponent();
	DAVTreeNode parentNode=struct.getParentNode();
	//replace the parent node in the struct with the corresponding node in the explorer's tree model
	struct.setParentNode(treeModel.getNodeMatchingPath(parentNode.getDAVFile().getFullName()));
	return struct;    
    }

    public DAVFileChooser getDAVFileChooser(ServerData[] serversToConnect)
    {
	final DefaultMutableTreeNode chooserRoot=new DefaultMutableTreeNode("chooser");
	final DAVTreeModel chooserModel=new DAVTreeModel(chooserRoot, true);
	final int numServers=serversToConnect.length;
	for (int i=0;i<numServers;i++)
	{
	    chooserModel.addConnectionNode(serversToConnect[i], true, null);
	}

	DAVFileChooser chooser=new DAVFileChooser(chooserModel);
	Object firstChild=chooserModel.getChild(chooserRoot, 0);
	chooser.setCurrentPath(new TreePath(chooserModel.getPathToRoot((TreeNode)firstChild)));
	return chooser;
    }

    public DAVFileChooser getDAVFileChooser()
    {
	return getDAVFileChooser(getConnectedServers());
    }

    public DAVFileChooser getDAVFileChooser(DAVTreeNode node)
    {
	ServerData sd=getConnection(node);
	DAVFileChooser chooser=getDAVFileChooser(new ServerData[] {sd});
	// the below doesn't work, yet -- setCurrentPath won't take you in one hop
	// to any depth in the tree -- you need to walk the tree path.  
	// to fix this, I will wait for the MAJOR FIX I intend to make to the way
	// tree models are handled!!
// 	DAVFile file=node.getDAVFile();
// 	if (file!=null)
// 	{
// 	    DAVTreeModel model = (DAVTreeModel) chooser.getModel();
// 	    DAVTreeNode chooserNode=model.getNodeMatchingPath(file.getFullName());
// 	    if (chooserNode!=null)
// 	    {	
// 		TreePath newPath=new TreePath(model.getPathToRoot(chooserNode));
// 		chooser.setCurrentPath(newPath);
// 	    }
// 	}
	return chooser;
    }

    public void addTableModelListener(TableModelListener tml)
    {
	tableModel.addTableModelListener(tml);
    }

    
    public void removeTableModelListener(TableModelListener tml)
    {
	tableModel.removeTableModelListener(tml);
    }

    public void addTreeModelListener(TreeModelListener tml)
    {
	treeModel.addTreeModelListener(tml);
    }

    public void removeTreeModelListener(TreeModelListener tml)
    {
	treeModel.removeTreeModelListener(tml);
    }

    protected void finalize() throws Throwable
    {
	allExplorers.remove(this);
    }
}

/* $Log: Explorer.java,v $
/* Revision 1.41  2001/06/13 01:09:03  smulloni
/* Some improvements to the network customizer (which is still buggy).
/* Also improvements for using the doAllprop switch, and for the layout of
/* the PropertiesDialog.
/*
/* Revision 1.40  2001/06/05 01:09:53  smulloni
/* fixed bug wtih failed connection attempts, which used to set the
/* connection flag as if the connection had succeeded.
/*
/* Revision 1.39  2001/05/30 15:25:28  smulloni
/* Added prompt for lock owner the first time a resource is locked for a
/* given server -- default value is username, if any.
/*
/* Revision 1.38  2001/01/03 20:11:31  smulloni
/* the DAVFileChooser now replaces JFileChooser for remote file access.
/* DAVMethod now has a protocol property.
/*
/* Revision 1.37  2001/01/03 00:30:35  smulloni
/* a number of modifications along the way to replacing JFileChooser with
/* something more suitable for remote (virtual) files.
/*
/* Revision 1.36  2000/12/21 18:53:13  smulloni
/* cosmetic improvements.
/*
/* Revision 1.35  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.34  2000/12/07 00:01:34  smulloni
/* adding an additional editor, still hard-coded into the DAVEditorFactory.
/*
/* Revision 1.33  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.32  2000/12/01 16:25:51  smullyan
/* improvements to look and feel; fixed NPE in DAVFile; new actions for text
/* editor
/*
/* Revision 1.31  2000/11/28 19:36:19  smullyan
/* some bug fixes; notes.
/*
/* Revision 1.30  2000/11/28 00:01:37  smullyan
/* added a status bar/minibuffer, with a location field showing the current line and
/* column number (for the SimpleTextEditor and kin only).
/*
/* Revision 1.29  2000/11/22 00:11:26  smullyan
/* editor now locks and unlocks, more or less appropriately.
/*
/* Revision 1.28  2000/11/20 23:30:20  smullyan
/* more editor integration work.
/*
/* Revision 1.27  2000/11/18 04:36:03  smullyan
/* work on StateMonitor and related functionality.
/*
/* Revision 1.26  2000/11/17 20:25:05  smullyan
/* new SaveAction; a StateMonitor being added to handle application state.
/*
/* Revision 1.25  2000/11/15 20:17:02  smullyan
/* added a Buffer interface, which is a wrapper around a displayable component.
/*
/* Revision 1.24  2000/11/14 17:13:34  smullyan
/* Added TreeModelListener hooks into Explorer, and revised PropertiesDialog to
/* use it.
/*
/* Revision 1.23  2000/11/13 23:28:30  smullyan
/* first pass at a gui for proppatch, added to the propertiese dialog.  Highly
/* problematic still, needs substantial work.
/*
/* Revision 1.22  2000/11/10 22:40:06  smullyan
/* added icon to table for resource type; fixes to copy and move; disabling of
/* menu items that are inappropriate.
/*
/* Revision 1.21  2000/11/09 23:34:56  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
