/*
 *  Copyright (c) 2000 - 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.awt.Cursor;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyVetoException;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.ListIterator;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import org.skunk.config.Configurator;
import org.skunk.config.LocalConfigStore;
import org.skunk.dav.client.gui.editor.EditEvent;
import org.skunk.dav.client.gui.editor.EditListener;
import org.skunk.trace.Debug;

public class ExplorerApp implements AppContext
{
    public static final String DAV_DIR=".skunkdav";
    public static final String CONFIG_DATA="config.dat";

    private Configurator configurator;
    private static AppContext appContext;

    private ArrayList viewList;
    private View currentView;

    /**
     * @return a singleton AppContext application object
     */
    public static AppContext getAppContext()
    {
	return appContext;
    }

    /**
     * convenience method for obtaining the view object that contains a given buffer
     * Note that it would be insane to call this right now were I not planning to 
     * allow the creation of multiple views.
     */
    public static View getViewForBuffer(Buffer buffer)
    {
	JComponent jc=buffer.getComponent();
	for (Iterator it=getAppContext().views();it.hasNext();)
	{
	    View v=(View)it.next();
	    for (Iterator it2=v.getDockedBuffers();it2.hasNext();)
	    {
		Buffer b=(Buffer) it2.next();
		if (b.equals(buffer))
		    return v;
	    }
	}
	return null;
    }

    /**
     * return the explorer in the same view as the given buffer
     */
    public static Explorer getExplorerForBuffer(Buffer buffer)
    {
	if (buffer instanceof Explorer) return (Explorer)buffer;
	View v=getViewForBuffer(buffer);
	if (v==null) return null;
	for (Iterator it=v.getDockedBuffers();it.hasNext();)
	{
	    Object o= it.next();
	    if (o instanceof Explorer) return (Explorer) o;
	}
	return null;
    }

    private ExplorerApp()
    {
	StateMonitor.registerProperty(this, "busy", StateProperties.BUSY, null, null);
	initPrefs();
	initViews();
    }

    private void initPrefs()
    {
	File appdir=new File(System.getProperty("user.home"), DAV_DIR);	   
	try
	{
	    makePrefDir(appdir);
	}
	catch (Exception e)
	{
	    Debug.trace(this, Debug.DP2, e);
	}	 
	String path=new File(appdir, CONFIG_DATA).getPath();
	configurator=new Configurator(new LocalConfigStore(path));
	//I'll remove this static method later.  I must ensure that the explorer component can be used outside
	//of the app.
	Configurator.setConfigurator(configurator);
    }

    private void initViews()
    {
	viewList=new ArrayList();
	currentView=new ExplorerView(this);
	addView(currentView);
	
    }

    public void setBusy(boolean busy)
    {
	currentView.getComponent().setCursor((busy) 
					     ? Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR)
					     : Cursor.getDefaultCursor());
    }


    /**
     * adds a view
     */
    public void addView(View view)
    {
	viewList.add(view);
    }

    /**
     * removes a view
     * Precondition: appContext contains this view
     */
    public void removeView(View view)
    {
	viewList.remove(view);
    }

    /**
     * @return an iterator of all Views in the AppContext
     */
    public Iterator views()
    {
	return viewList.listIterator();
    }

   /**
     * @return the number of current views
     */
    public int getViewCount()
    {
	return viewList.size();
    }

    /**
     * @return the view at the given index
     */
    public View getView(int index)
    {
	return (View) viewList.get(index);
    }

    /**
     * @return the currently active (focussed) view
     */
    public View getCurrentView()
    {
	//TO BE DONE --add a mechanism to track the current view
	return currentView;
    }

    public static void main(String[] args)
    {
	if (System.getProperty("LandF")==null)
	    LookAndFeel.setLookAndFeel();
	appContext=new ExplorerApp();	
	// N.B. -- the SPI version is no more; now folded into the main app.
        // However, this is where plugins should be loaded -- TO BE DONE.
// 	//this is a hack for the SPI version of this app.
// 	//it saves me the trouble of writing a subclass of ExplorerApp to do this,
// 	//and using a separate jar manifest for spi.  
// 	try
// 	{
// 	    Class.forName("org.skunk.spi.SpiInitializer");
// 	}
// 	catch (ClassNotFoundException conifer)
// 	{
// 	    //OK, this should happen
// 	}
    }

    private static void makePrefDir(File appdir) throws IOException, SecurityException
    {

	if (appdir.exists())
	{
	    if (appdir.isDirectory())
		return;
	    else throw new IOException("unable to create directory: name taken by file");
	}
	else
	{
	    boolean succeeded=appdir.mkdir();
	    if (!succeeded) throw new IOException ("unable to create directory");
	}
    }

    /**
     * @return the dock mode (interface style)
     */
    public DockMode getDockMode()
    {
	//for now, hard-code this
	return DockMode.TABBED_PANE_MODE;
    }

    /**
     *  @return the configurator object for the application
     */
    public Configurator getConfigurator()
    {
	return configurator;
    }

    /**
     *  @return whether or not the application is trusted to access the filesystem,
     *          open sockets to multiple hosts, etc.
     */
    public boolean isTrusted()
    {
	return true;
    }
}

/* $Log: ExplorerApp.java,v $
/* Revision 1.20  2001/07/22 07:04:24  smulloni
/* changed the handling of release number.
/*
/* Revision 1.19  2001/07/11 02:15:01  smulloni
/* the previous spi build targets have been folded into the main build;
/* also added a javadoc target.
/*
/* Revision 1.18  2001/01/30 23:03:19  smulloni
/* beginning of integration of syntax highlighting into SimpleTextEditor.
/*
/* Revision 1.17  2000/12/21 18:53:13  smulloni
/* cosmetic improvements.
/*
/* Revision 1.16  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.15  2000/12/05 03:47:43  smulloni
/* added preamble
/*
/* Revision 1.14  2000/12/04 23:51:16  smulloni
/* added ImageViewer; fixed word in SimpleTextEditor
/*
/* Revision 1.13  2000/12/01 16:25:52  smullyan
/* improvements to look and feel; fixed NPE in DAVFile; new actions for text
/* editor
/*
/* Revision 1.12  2000/11/28 00:01:38  smullyan
/* added a status bar/minibuffer, with a location field showing the current line and
/* column number (for the SimpleTextEditor and kin only).
/*
/* Revision 1.11  2000/11/22 00:11:26  smullyan
/* editor now locks and unlocks, more or less appropriately.
/*
/* Revision 1.10  2000/11/20 23:30:20  smullyan
/* more editor integration work.
/*
/* Revision 1.9  2000/11/17 20:25:05  smullyan
/* new SaveAction; a StateMonitor being added to handle application state.
/*
/* Revision 1.8  2000/11/16 20:45:17  smullyan
/* the start of editor integration.
/*
/* Revision 1.7  2000/11/15 20:17:03  smullyan
/* added a Buffer interface, which is a wrapper around a displayable component.
/*
/* Revision 1.6  2000/11/15 19:45:36  smullyan
/* beginning of revamp of application to include multiple buffers.
/*
/* Revision 1.5  2000/11/09 23:34:57  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
