/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;
import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.skunk.config.Configurator;
import org.skunk.trace.Debug;

public abstract class PreferencesCustomizer
{
    private String title;
    private Configurator configurator;
    private volatile boolean dirty=false;
    private HashSet configTargets;

    //public static final String DIRTY_PROPERTY="dirty";

    private ChangeEvent reusableChangeEvent=new ChangeEvent(this);
    private Vector changeListeners=new Vector();

    public PreferencesCustomizer(String title, Configurator configurator)
    {
	this.title=title;
	this.configurator=configurator;
	this.configTargets=new HashSet();
    }

    public void addDirtListener(ChangeListener listener)
    {
	changeListeners.add(listener);
    }

    public void removeDirtListener(ChangeListener listener)
    {
	changeListeners.remove(listener);
    }

    public abstract JComponent getCustomizer();

    protected final boolean isDirty()
    {
	return dirty;
    }

    protected final void setDirty(boolean dirty)
    {
	if (Debug.DEBUG) 
	    Debug.trace(this, Debug.DP4, "in setDirty({0})", new Boolean(dirty));
	this.dirty=dirty;
	fireChangeEvent();
    }



    private void fireChangeEvent()
    {
	for (Enumeration enum=changeListeners.elements();enum.hasMoreElements();)
	{
	    ((ChangeListener)enum.nextElement()).stateChanged(reusableChangeEvent);
	}
    }

    public final String getTitle()
    {
	return this.title;
    }

    public void ok()
    {
	for (Iterator it=getConfigTargets();it.hasNext();)
	{
	    Object target=it.next();
	    if (isDirty()) 
		configurator.configure(target);
	    configurator.tryOff(target);
	}
	setDirty(false);
    }

    public void tryOn()
    {
	if (isDirty())
	{
	    for (Iterator it=getConfigTargets();it.hasNext();)
	    {
		Object o=it.next();
		if (Debug.DEBUG)
		    Debug.trace(this, Debug.DP4, "about to try on target {0}", o);
		configurator.tryOn(o);
	    }
	}
	setDirty(false);	
    }


    public void revert()
    {
	configurator.revert(configTargets.toArray());
	setDirty(false);
    }

    public void cancel()
    {
	configurator.revert(configTargets.toArray());
	setDirty(false);
    }

    public void revertToDefault()
    {
	for (Iterator it=getConfigTargets();it.hasNext();)
	{
	    configurator.revertToDefault(it.next());
	}
	setDirty(false);
    }
	    

    protected final Iterator getConfigTargets()
    {
	return configTargets.iterator();
    }

    protected final void removeConfigTarget(Object configTarget)
    {
	configTargets.remove(configTarget);
    }

    protected final void removeConfigTargetsOfClass(Class c)
    {
	synchronized (configTargets)
	{
	    ArrayList targetsToRemove=new ArrayList();
	    for (Iterator it=getConfigTargets();it.hasNext();)
	    {
		Object o=it.next();
		if (o==null || c.isAssignableFrom(o.getClass()))
		    targetsToRemove.add(o);
	    }
	    for (Iterator it=targetsToRemove.iterator();it.hasNext();)
		configTargets.remove(it.next());
	}
    }

    protected final void addConfigTarget(Object configTarget)
    {
	configTargets.add(configTarget);
    }

    protected final void setConfigValue(Class configClass, 
					String configProperty, 
					Object configValue)
    {
	configurator.getStore().setConfigValue(configClass, 
					       configProperty, 
					       configValue);
    }  

    protected Configurator getConfigurator()
    {
	return configurator;
    }
}

/* $Log: PreferencesCustomizer.java,v $
/* Revision 1.9  2001/07/26 18:35:33  smulloni
/* patched HTTPClient so jikes doesn't issue warnings about catching Exception;
/* replaced PropertyChangeEvents with ChangeEvents in PreferencesCustomizer and
/* related classes; incremented app version to 1.0.2.3.
/*
/* Revision 1.8  2001/02/16 18:15:10  smulloni
/* many fixes to TextEditorCustomizer.  FileMode and SyntaxStyle now have a
/* configData property (they will probably be made into sibling classes).
/*
/* Revision 1.7  2001/02/06 00:11:18  smulloni
/* struggle, perhaps futile, with the TextEditorCustomizer and other implicated
/* classes
/*
/* Revision 1.6  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.5  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.4  2000/11/09 23:34:58  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
