/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumn;
import javax.swing.tree.TreeNode;
import org.skunk.assertF.Assertion;
import org.skunk.dav.client.DAVConnection;
import org.skunk.dav.client.DAVException;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.DAVProperty;
import org.skunk.dav.client.Depth;
import org.skunk.dav.client.method.PropFindMethod;
import org.skunk.trace.Debug;

public class PropertiesDialog extends JDialog
{
    private DAVFile file;
    private DAVTreeNode node;
    private Explorer ex;
    private DAVPropertiesTableModel davModel;
    private CustomPropertiesTableModel customModel;

    private String[] headers= 
    {
	ResourceManager.getMessage(ResourceManager.NAME),
	ResourceManager.getMessage(ResourceManager.NAMESPACE),
	ResourceManager.getMessage(ResourceManager.VALUE)
    };

    public PropertiesDialog(Frame owner, Explorer ex, DAVFile file)
    {
	super(owner, false); //does not need to be modal
	if (file==null) 
	    throw new IllegalArgumentException("file must be non-null");
	this.file=file;
	this.ex=ex;
	this.node=ex.getSelectedNode();
	if (node.getDAVFile().getName().equals(file.getName()))
	{
	    TreeNode pops=node.getParent();
	    if (pops!=null && pops instanceof DAVTreeNode)
	    {
		node=(DAVTreeNode)pops;
	    }
	    else
	    {
		Debug.trace(this, Debug.DP2, "OOPS -- node not a DAVTreeNode");
	    }
	}
	refreshIfNeeded();
	ex.addTreeModelListener(new ExplorerListener());
	initComponents();
    }

    protected void setDAVFile(DAVFile file)
    {
	this.file=file;
	if (davModel!=null)
	    davModel.setDAVFile(file);
	if (customModel!=null)
	    customModel.setDAVFile(file);
    }

    private void refreshIfNeeded()
    {
	if (! DAVTreeModel.usesAllprop())
	{
	    DAVFile newFile=null;
	    String resource=file.getName();
	    PropFindMethod method=new PropFindMethod(resource, Depth.ZERO, false);
	    DAVConnection conn=file.getDAVConnection();
	    try
	    {	
		conn.execute(method);
		newFile=method.getDAVFile();
	    }
	    catch (DAVException davex)
	    {
		Debug.trace(this, Debug.DP2, davex);
	    }
	    catch (IOException oyVeh)
	    {
		Debug.trace(this, Debug.DP2, oyVeh);
	    }
	    if (newFile!=null)
	    {
		DAVProperty[] props=newFile.getAvailableProperties();
		Assertion.assertF(props!=null, 
				 "available properties are not null");
		method=new PropFindMethod(resource, Depth.ZERO, props);
		try
		{
		    conn.execute(method);
		    newFile=method.getDAVFile();
		}
		catch (DAVException davex)
		{
		    Debug.trace(this, Debug.DP2, davex);
		}
		catch (IOException oyVeh)
		{
		    Debug.trace(this, Debug.DP2, oyVeh);
		}
	    }
	    Assertion.assertF(newFile!=null, "new file is not null");
	    Debug.trace(this, Debug.DP3, newFile.toVerboseString());
	    setDAVFile(newFile);
	}
    }

    private void initComponents()
    {
	setTitle(ResourceManager.getMessage(ResourceManager.PROPERTIES_DIALOG_TITLE, 
					    new Object[] {file.getFileName()}));
	JTabbedPane tabby=new JTabbedPane(JTabbedPane.TOP);
	tabby.add(getDAVPropertiesTab(),
		  ResourceManager.getMessage(ResourceManager.DAV_PROPERTIES_TAB_TITLE));
	tabby.add(getCustomPropertiesTab(),
		  ResourceManager.getMessage(ResourceManager.CUSTOM_PROPERTIES_TAB_TITLE));
	JPanel contentPanel=new JPanel();
	contentPanel.setLayout(new BorderLayout());
	contentPanel.add(tabby, BorderLayout.CENTER);
	
	/* 
	   CONSIDER ADDING A DIALOG DISMISSAL BUTTON HERE
	   contentPanel.add(getButtonBox(), BorderLayout.SOUTH);
	*/
	setContentPane(contentPanel);
	pack();	
    }

    private void initTableColumns(JTable table)
    {
	for (Enumeration enum=table.getColumnModel().getColumns();
	     enum.hasMoreElements();)
	{
	    TableColumn tc=(TableColumn) enum.nextElement();
	    Object identifier = tc.getIdentifier();
	    if (identifier.equals(headers[0])) //name
	    {
		tc.setPreferredWidth(110);
	    }
	    else if (identifier.equals(headers[1])) //namespace
	    {
		tc.setPreferredWidth(75);
	    }
	    else if (identifier.equals(headers[2])) //value
	    {
		tc.setPreferredWidth(180);
	    }
	}
    }

    private JComponent getDAVPropertiesTab()
    {
	davModel=new DAVPropertiesTableModel(file);
	JTable table=new JTable(davModel);
	table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
	initTableColumns(table);
	JScrollPane jsp= new JScrollPane(table);
	jsp.setPreferredSize(new Dimension(350, 330));
	return jsp;
    }

    private JComponent getCustomPropertiesTab()
    {
	customModel=new CustomPropertiesTableModel(file);
	final JTable table=new JTable(customModel);
	initTableColumns(table);
	table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	table.clearSelection();
	table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
	JScrollPane jsp= new JScrollPane(table);
	jsp.setPreferredSize(new Dimension(250, 350));
	final JButton addCustomButton=new JButton(ResourceManager.getMessage(ResourceManager.ADD_KEY));
	addCustomButton.addActionListener(new AddCustomAction());
	final JButton editCustomButton=new JButton(ResourceManager.getMessage(ResourceManager.EDIT_KEY));
	editCustomButton.addActionListener(new EditCustomAction(table));
	editCustomButton.setEnabled(false);
	final JButton deleteCustomButton=new JButton(ResourceManager.getMessage(ResourceManager.DELETE_KEY));
	deleteCustomButton.addActionListener(new DeleteCustomAction(table));
	deleteCustomButton.setEnabled(false);
	table.getSelectionModel().addListSelectionListener(new ListSelectionListener() 
	    {
		public void valueChanged(ListSelectionEvent lassie) 
		{
		    boolean enableEm=table.getSelectedRowCount()==1;
		    deleteCustomButton.setEnabled(enableEm);
		    editCustomButton.setEnabled(enableEm);
		}
	    });

	Box buttonBox = Box.createHorizontalBox();
	buttonBox.add(Box.createHorizontalGlue());
	buttonBox.add(addCustomButton);
	buttonBox.add(Box.createHorizontalGlue());
	buttonBox.add(editCustomButton);
	buttonBox.add(Box.createHorizontalGlue());
	buttonBox.add(deleteCustomButton);
	buttonBox.add(Box.createHorizontalGlue());
	
	JPanel wholeTab=new JPanel(new BorderLayout());
	wholeTab.add(jsp, BorderLayout.CENTER);
	wholeTab.add(buttonBox, BorderLayout.SOUTH);		
	return wholeTab;
    }

    private class ExplorerListener implements TreeModelListener
    {
	public void treeNodesChanged(TreeModelEvent tme)
	{
	    Debug.trace(this, Debug.DP3, "in treeNodesChanged");
	    _process();
	}

	public void treeNodesInserted(TreeModelEvent tme) {}

	public void treeNodesRemoved(TreeModelEvent tme) {}
	
	public void treeStructureChanged(TreeModelEvent tme)
	{
	    Debug.trace(this, Debug.DP3, "in treeStructureChanged");
	    _process();
	}

	public void _process()
	{
	    String filename=file.getFileName();	
	    DAVFile newFile=ex.getFile(node, filename);
	    if (newFile!=null) setDAVFile(newFile);
	    refreshIfNeeded();
	}
    }


    private class AddCustomAction implements ActionListener
    {
	public void actionPerformed(ActionEvent ae)
	{
	    PropertyEditPanel  propPanel=new PropertyEditPanel();
	    int option=propPanel.showAddDialog();
	    if (option==JOptionPane.OK_OPTION)
	    {
		String name=propPanel.getPropertyName();
		String namespace=propPanel.getPropertyNamespace();
		String value=propPanel.getPropertyValue();
		if (name==null || namespace==null || value==null)
		{
		    Debug.trace(this, Debug.DP3, "incomplete submission for action");
		    //should show an error message -- TO BE DONE
		    return;
		}
		//do a proppatch
		Map m=new HashMap();
		m.put(new DAVProperty(name, 
				      "ns"+(int)Math.floor(111*Math.random()), 
				      namespace), 
		      value);
		ex.proppatch(node, file, m);
		refreshIfNeeded();
	    }	    	    
	}
    }

    private class EditCustomAction implements ActionListener
    {
	private JTable table;

	EditCustomAction(JTable table)
	{
	    this.table=table;
	}
	public void actionPerformed(ActionEvent ae)
	{
	    int selected=table.getSelectedRow();
	    if (selected==-1) 
	    {
		Debug.trace(this, Debug.DP3, "no row selected!");
		return;
	    }
     
	    String name=table.getValueAt(selected, 0).toString();
	    String namespace=table.getValueAt(selected, 1).toString();
	    String value=table.getValueAt(selected, 2).toString();
	    PropertyEditPanel  propPanel=new PropertyEditPanel(name, namespace, value);
	    int option=propPanel.showEditDialog();
	    if (option==JOptionPane.OK_OPTION)
	    {
		String newName=propPanel.getPropertyName();
		String newNamespace=propPanel.getPropertyNamespace();
		value=propPanel.getPropertyValue();
		if (newName==null || newNamespace==null || value==null)
		{
		    Debug.trace(this, Debug.DP3, "incomplete submission for action");
		    return;
		}
		Map m=new HashMap();
		if ((!name.equals(newName)) || (!namespace.equals(newNamespace)))
		{
		    //user has changed the name or namespace of the property; delete old version
		    m.put(new DAVProperty(name, "ns"+(int)Math.floor(111*Math.random()), namespace), null);
		}
		m.put(new DAVProperty(newName, "ns"+(int)Math.floor(111*Math.random()), newNamespace), value);
		ex.proppatch(node, file, m);
		refreshIfNeeded();
	    }	    
	}
    }

    private class DeleteCustomAction implements ActionListener
    {
	private JTable table;

	DeleteCustomAction(JTable table)
	{
	    this.table=table;
	}

	public void actionPerformed(ActionEvent ae)
	{
	    int selected=table.getSelectedRow();
	    if (selected==-1) 
	    {
		Debug.trace(this, Debug.DP3, "no row selected!");
		return;
	    }     
	    String name=table.getValueAt(selected, 0).toString();
	    String namespace=table.getValueAt(selected, 1).toString();

	    //do a proppatch
	    Map m=new HashMap();
	    m.put(new DAVProperty(name, "ns"+(int)Math.floor(111*Math.random()), namespace), null);
	    ex.proppatch(node, file, m);
	    refreshIfNeeded();

	}
    }

    private class PropertyEditPanel extends JPanel
    {
	JTextField nameField, namespaceField, valueField;

	int showAddDialog()
	{
	    String title=ResourceManager.getMessage(ResourceManager.PROPERTY_ADD_DIALOG_TITLE);
	    return JOptionPane.showConfirmDialog(PropertiesDialog.this, 
						 this, 
						 title, 
						 JOptionPane.OK_CANCEL_OPTION, 
						 JOptionPane.QUESTION_MESSAGE,
						 null);
	}

	int showEditDialog()
	{
	    String title=ResourceManager.getMessage(ResourceManager.PROPERTY_EDIT_DIALOG_TITLE);
	    return JOptionPane.showConfirmDialog(PropertiesDialog.this, 
						 (Object) this, 
						 title, 
						 JOptionPane.OK_CANCEL_OPTION, 
						 JOptionPane.QUESTION_MESSAGE,
						 null);
	}

	public String getPropertyName()
	{
	    return nameField.getText();
	}

	public void setPropertyName(String name)
	{
	    this.nameField.setText(name);
	}

	public String getPropertyNamespace()
	{
	    return namespaceField.getText();
	}

	public void setPropertyNamespace(String namespace)
	{
	    this.namespaceField.setText(namespace);
	}

	public String getPropertyValue()
	{
	    return valueField.getText();
	}

	public void setPropertyValue(String value)
	{
	    this.valueField.setText(value);
	}
	    
	protected PropertyEditPanel()
	{
	    this(null, null, null);
	}

	protected PropertyEditPanel(String name, String namespace, String value)
	{
	    super(new BorderLayout());
	    this.add(getPropertyEditPanel(name, namespace, value), BorderLayout.CENTER);
	}

	private JPanel getPropertyEditPanel(String propertyName, 
					    String propertyNamespace, 
					    String propertyValue)
	{
	    JPanel propPanel=new JPanel(new GridBagLayout());
	    GridBagConstraints gbc=new GridBagConstraints();
	    gbc.gridx=0;
	    gbc.gridy=0;
	    gbc.fill=GridBagConstraints.NONE;
	    gbc.anchor=GridBagConstraints.NORTHWEST;
	    gbc.insets=new Insets(2, 2, 2, 2);
	    gbc.weightx=0.0;
	    gbc.weighty=0.0;
	    gbc.gridwidth=1;
	    gbc.gridheight=1;

	    JLabel tempLabel=new JLabel(ResourceManager.getMessage(ResourceManager.PROPERTY_NAME));
	    propPanel.add(tempLabel, gbc);
	    gbc.gridx++;

	    nameField=new JTextField(22);
	    if (propertyName!=null) nameField.setText(propertyName);
	    propPanel.add(nameField, gbc);
	    gbc.gridx=0;
	    gbc.gridy++;

	    tempLabel=new JLabel(ResourceManager.getMessage(ResourceManager.PROPERTY_NAMESPACE));
	    propPanel.add(tempLabel, gbc);
	    gbc.gridx++;
	
	    namespaceField=new JTextField(22);
	    if (propertyNamespace!=null) namespaceField.setText(propertyNamespace);
	    propPanel.add(namespaceField, gbc);
	    gbc.gridx=0;
	    gbc.gridy++;

	    tempLabel=new JLabel(ResourceManager.getMessage(ResourceManager.PROPERTY_VALUE));
	    propPanel.add(tempLabel, gbc);
	    gbc.gridx++;

	    valueField=new JTextField(22);
	    if (propertyValue!=null) valueField.setText(propertyValue);
	    propPanel.add(valueField, gbc);

	    return propPanel;
	}
    }

    private class DAVPropertiesTableModel extends AbstractTableModel
    {
       /**
	*  data for the table -- two columns, property names and values 
	*/
	private Object[][] tableData=
	{
	    { ResourceManager.getMessage(ResourceManager.HOST_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.PORT_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.FILE_NAME_FILE_PROP), "" },
 	    { ResourceManager.getMessage(ResourceManager.CREATION_DATE_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.LAST_MODIFIED_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.CONTENT_TYPE_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.CONTENT_LENGTH_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.CONTENT_LANGUAGE_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.DISPLAY_NAME_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.RESOURCE_TYPE_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.ETAG_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.SUPPORTED_LOCKS_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.ACTIVE_LOCKS_FILE_PROP), "" },
	    { ResourceManager.getMessage(ResourceManager.LINKS_FILE_PROP), "" }
	};


	DAVPropertiesTableModel(DAVFile file)
	{
	    super();
	    makePropertyList(file);
	}

	void setDAVFile(DAVFile file)
	{
	    makePropertyList(file);
	}

	private void makePropertyList(DAVFile file)
	{
	    int i=0;
	    tableData[i++][1]=file.getHost();
	    tableData[i++][1]=new Integer(file.getPort());
	    tableData[i++][1]=file.getName();
	    tableData[i++][1]=file.getCreationDate();
	    tableData[i++][1]=file.getLastModified();
	    tableData[i++][1]=file.getContentType();
	    tableData[i++][1]=file.getContentLength();
	    tableData[i++][1]=file.getContentLanguage();
	    tableData[i++][1]=file.getDisplayName();
	    tableData[i++][1]=file.getResourceType();
	    tableData[i++][1]=file.getEtag();
	    tableData[i++][1]=printArray(file.getSupportedLocks());
	    tableData[i++][1]=printArray(file.getLocks());
	    tableData[i][1]=printArray(file.getLinks());
	}

	private String printArray(Object[] stuff)
	{
	    if (stuff==null) return null;
	    StringBuffer sb=new StringBuffer();
	    for (int i=0;i<stuff.length;i++)
	    {
		if (i>0) sb.append(", ");
		sb.append(stuff[i]);
	    }
	    return sb.toString();
	}

	public int getRowCount()
	{
	    return tableData.length;
	}

	public String getColumnName(int c)
	{
	    //skip the namespace header
	    return headers[(c==1) ? 2: 0];
	}

	public int getColumnCount()
	{
	    return 2;
	}

	public Object getValueAt(int row, int column)
	{
	    return tableData[row][column];
	}
    }

    private class CustomPropertiesTableModel extends AbstractTableModel
    {
	ArrayList tableData;
	CustomPropertiesTableModel(DAVFile file)
	{
	    super();
	    this.setDAVFile(file);
	}

	void setDAVFile(DAVFile file)
	{
	    tableData=extractPairs(file.getCustomProperties());
	    Debug.trace(this, Debug.DP3, "tableData for custom properties: {0}",
			new Object[] {tableData});
	    fireTableDataChanged();
	}

	private ArrayList extractPairs(Map m)
	{
	    ArrayList daList=new ArrayList();
	    if (m!=null)
	    {
		for (Iterator it=m.keySet().iterator();it.hasNext();)
		{
		    Object key=it.next();
		    Object value=m.get(key);
		    Debug.trace(this, Debug.DP3, "reading custom property {0} with value {1}",
				new Object[] {key, value});
		    daList.add(new Object[] {key, value});
		}
	    }
	    return daList;
	}
	    
	public int getRowCount()
	{
	    return tableData.size();
	}

	public int getColumnCount()
	{
	    return 3;
	}

	public String getColumnName(int c)
	{
	    return headers[c];
	}

	public Object getValueAt(int row, int column)
	{
	    Assertion.assertF((column<3), "only three columns");
	    Object o=tableData.get(row);
	    switch (column)
	    {
	    case 0:
	    case 1:
		Object propObj=((Object[]) o)[0];
		if (propObj instanceof DAVProperty)
		{
		    DAVProperty davProp=(DAVProperty) propObj;
		    if (column==0) return davProp.getElementName();
		    else return davProp.getNamespace();
		}
		else return propObj;		
	    case 2:
		return ((Object[]) o)[1];		
	    default:
		Assertion.assertF(false, "this should not be reached");
		return null;
	    }
	}
    }


}


/* $Log: PropertiesDialog.java,v $
/* Revision 1.13  2001/06/13 01:09:03  smulloni
/* Some improvements to the network customizer (which is still buggy).
/* Also improvements for using the doAllprop switch, and for the layout of
/* the PropertiesDialog.
/*
/* Revision 1.12  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.11  2000/12/14 23:09:17  smulloni
/* fixes to search and replace; partial fix to adding custom properties bug.
/*
/* Revision 1.10  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.9  2000/11/29 23:16:04  smullyan
/* adding first rough cut of search capability to the text editor.  View
/* is being updated to allow components to be docked into the status bar.
/*
/* Revision 1.8  2000/11/14 23:17:08  smullyan
/* more improvements to the PropertiesDialog.
/*
/* Revision 1.7  2000/11/14 17:13:35  smullyan
/* Added TreeModelListener hooks into Explorer, and revised PropertiesDialog to
/* use it.
/*
/* Revision 1.6  2000/11/13 23:28:31  smullyan
/* first pass at a gui for proppatch, added to the propertiese dialog.  Highly
/* problematic still, needs substantial work.
/*
/* Revision 1.5  2000/11/09 23:34:59  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
