/*  $Id: ServerData.java,v 1.19 2001/07/29 22:37:36 smulloni Exp $
 *  Time-stamp: <01/07/27 15:21:32 smulloni>
 *
 *  Copyright (c) 2000-2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.io.IOException;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;
import org.skunk.assertF.Assertion;
import org.skunk.config.LocalConfigStore;
import org.skunk.config.Configurator;
import org.skunk.dav.client.DAVConstants;
import org.skunk.trace.Debug;

/**
 *  encapsulates data about a DAV server.
 */
public class ServerData implements Cloneable, Serializable
{
    static final long serialVersionUID = -6873822752692653196L;
    protected static ServerDataContainer container=new ServerDataContainer();
 
    public static void init()
    {
	try
	{
	    loadServerList();
	}
	catch (Exception e)
	{
	    Debug.trace(ServerData.class, Debug.DP2, e);
	    container.setServerDataVector(new Vector());
	}
    }
    
    private String host, initialPath, username, password;
    private int port;
    private boolean permitsLockStealing=true;
    private boolean https=false;
    private boolean rememberPassword=false;
    private String owner=null;
    private boolean touched=false;

    public ServerData(boolean https,
		      String host,
		      int port,
		      String initialPath,
		      String username,
		      String password)
    {
	this(https, 
	     host, 
	     port, 
	     initialPath, 
	     username, 
	     password, 
	     false);
    }

    public ServerData(boolean https,
		      String host,
		      int port,
		      String initialPath,
		      String username,
		      String password,
		      boolean rememberPassword)
    {
	this(https, host, port, initialPath, username, password, rememberPassword, true);
    }
    
    public ServerData(boolean https, 
		      String host, 
		      int port, 
		      String initialPath, 
		      String username, 
		      String password,
		      boolean rememberPassword,
		      boolean autotouch)
    {
	this.https=https;
	this.host=host;
	this.port=port;
	this.initialPath=correctPath(initialPath);
	this.username=username;
	this.password=password;
	this.rememberPassword=rememberPassword;
	if (autotouch)
	try
	{
	    touch(this);	
	}
	catch (IOException oyVeh)
	{
	    Debug.trace(this, Debug.DP2, oyVeh);
	}
    }

    public ServerData(String host, 
		      int port, 
		      String initialPath, 
		      String username, 
		      String password)
    {
	this(false, 
	     host, 
	     port, 
	     initialPath, 
	     username, 
	     password,
	     false);
    }

    public boolean usesSSL() 
    { 
	return https; 
    }

    public void setUsesSSL(boolean https) 
    { 
	this.https=https; 
    }
    
    public String getHost() 
    { 
	return host; 
    }

    public void setHost(String host) 
    { 
	this.host=host; 
    }

    public int getPort() 
    { 
	return port; 
    }

    public void setPort(int port) 
    { 
	this.port=port; 
    }

    public String getInitialPath() 
    { 
	return initialPath; 
    }

    public void setInitialPath(String initialPath) 
    { 
	this.initialPath=correctPath(initialPath); 
    }

    public String getUsername() { return username; }

    public void setUsername(String username)
    {
	this.username=username;
    }

    public String getPassword() { return password; }

    public void setPassword(String password)
    {
	this.password=password;
    }

    /**
     * whether or not password information for the server should be
     * saved.
     * @return whether or not to remember the password.
     */
    public boolean getRememberPassword() 
    { 
	return rememberPassword; 
    }

    public void setRememberPassword(boolean rememberPassword) 
    { 
	this.rememberPassword=rememberPassword; 
    }

    public boolean getPermitsLockStealing() 
    { 
	return this.permitsLockStealing; 
    }
    
    public void setPermitsLockStealing(boolean permitsLockStealing) 
    { 
	this.permitsLockStealing=permitsLockStealing; 
	Debug.trace(this, Debug.DP3, 
		    "permitsLockStealing set to " 
		    + permitsLockStealing 
		    + " for {0}",
		    this);
	try
	{
	    saveServers();
	}
	catch (IOException oyVeh)
	{
	    Debug.trace(this, Debug.DP2, oyVeh);
	}
    }

    public String getOwner() { return this.owner; }

    public void setOwner(String owner)
    {
	this.owner=owner;
    }
	       
    public String toString() 
    { 
	return makeString(https, host, port, initialPath);
    }

    private static String makeString(String host, 
				     int port, 
				     String initialPath)
    {
	return new StringBuffer(host)
	    .append(':')
	    .append(port)
	    .append(initialPath)
	    .toString();
    }

    private static String makeString(boolean https, 
				     String host, 
				     int port, 
				     String initialPath)
    {
	return new StringBuffer((https)
				? DAVConstants.HTTPS
				: DAVConstants.HTTP)
	    .append("://")
	    .append(host)
	    .append(':')
	    .append(port)
	    .append(initialPath)
	    .toString();
    }

    private static String correctPath(String path)
    {
	if (path==null) 
	    return DAVConstants.DAV_FILE_SEPARATOR;
	else if (path.endsWith(DAVConstants.DAV_FILE_SEPARATOR))
	    return path;
	else return path+DAVConstants.DAV_FILE_SEPARATOR;
    }
			
    /**
     * makes this serverData the last selected one.
     */
    protected static void touch(ServerData sd) throws IOException
    {
	Vector servers=getServers();
	servers.remove(sd);
	addNewServer(sd);
	saveServers();
	sd.touched=true;
    }
    
    public static Vector getServers()
    {
	return container.getServerDataVector(); 
    }
    
    public static ServerData getLastSelectedServer()
    {
	Vector servers=getServers();
	int size=servers.size();
	return (size>0) 
	    ?  (ServerData) servers.elementAt(size-1)
	    : null;
    }
    
    private static void addNewServer(ServerData sd)
    {
	Vector servers=getServers();
	for (Enumeration enum=servers.elements();enum.hasMoreElements();)
	{
	    ServerData next=(ServerData)enum.nextElement();
	    if (next.getHost().equals(sd.getHost()) 
		&& next.getPort() ==sd.getPort()
		&& next.getInitialPath().equals(sd.getInitialPath()))
	    {
		servers.remove(next);
		break;
	    }
	}
	servers.addElement(sd);
    }

    protected static void saveServers() throws IOException
    {
	Configurator.getConfigurator().getStore()
	    .setConfigValue(ServerDataContainer.class, 
			    ServerDataContainer.SERVER_DATA_VECTOR_PROPERTY, 
			    getServers());
    }

    
    public static void removeServer(ServerData sd) throws IOException
    {
	Vector servers=getServers();
	if (servers.contains(sd)) 
	{
	    servers.remove(sd);
	    saveServers();
	}
    }
    
    private static void loadServerList()
            throws IOException, ClassNotFoundException
    {
	Configurator.getConfigurator().configure(container);
	Debug.trace(ServerData.class, 
		    Debug.DP3, 
		    "ServerDataContainer loaded; serverdata is {0}",
		    getServers());
    }

    public Object clone() 
    {
	try
	{
	    return super.clone();
	}
	catch (CloneNotSupportedException inconceivable)
	{
	    return null;
	}
    }

    public static ServerData getServer(String serverString)
    {
	if (!serverString.startsWith(DAVConstants.HTTP) 
	    && !serverString.startsWith(DAVConstants.HTTPS))
	{
	    serverString=new StringBuffer(DAVConstants.HTTP)
		.append("://")
		.append(serverString)
		.toString();
	}
	for (Enumeration enum=getServers().elements();
	     enum.hasMoreElements();)
	{
	    Object nextObj=enum.nextElement();
	    if (serverString.equals(nextObj.toString()))
		return (ServerData) nextObj;
	}
	Debug.trace(ServerData.class, Debug.DP3, 
		    "server {0} not found", 
		    new Object[] {serverString});
	return null;
    }

    public static ServerData getServer(String host, 
				       int port, 
				       String initialPath)
    {
	return getServer(makeString(false, 
				    host, 
				    port, 
				    initialPath));
    }  

    public static ServerData getServer(boolean https, 
				       String host, 
				       int port, 
				       String initialPath)
    {
	return getServer(makeString(https, 
				    host, 
				    port, 
				    initialPath));
    }  
}

/* $Log: ServerData.java,v $
/* Revision 1.19  2001/07/29 22:37:36  smulloni
/* fix to NetworkCustomizer; workaround to bad design of ServerData
/*
/* Revision 1.18  2001/06/09 03:09:09  smulloni
/* added server customization to the NetworkCustomizer.  Buggy.
/*
/* Revision 1.17  2001/05/30 03:49:35  smulloni
/* Some bug fixes and improvements to authentication handling.  Users are now
/* prompted for a password the first time they connect to a given server
/* in a session; if the password is rejected, they are asked again.
/* Users also now have the option of whether a given password will be remembered
/* or not.
/*
/* Revision 1.16  2001/01/05 08:01:12  smulloni
/* changes to the connection gui for the Explorer; added depth configurability to
/* propfind in the jpython test script; added an experimental "allprop" system
/* property which affects the propfind query type
/*
/* Revision 1.15  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.14  2000/12/18 23:22:47  smulloni
/* added optional SSL capability to the gui application.  Separate make targets
/* exist for building it will SSL support.
/*
/* Revision 1.13  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.12  2000/11/09 23:34:59  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
