/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.skunk.trace.Debug;

/**
 * a property can be registered with the state monitor so that when the 
 * a given application property changes state, so will the registered property,
 * according to a mapping of property values.  If the mapping is null,
 * the application property value is used directly to set the target property value.
 */
public class StateMonitor
{
//     private static Map applicationValues;
    private static List monitors;

    private Object target;
    private String targetProperty;
    private String applicationProperty;
    private Mapper valueMapper;
    private DomainMatcher domainMatcher;

    private Method setter;

    static
    {
	//applicationValues = Collections.synchronizedMap(new HashMap());
	monitors=Collections.synchronizedList(new ArrayList());
    }

    /**
     *  set a target object's property value to vary according 
     *  to changes in the value of the application property, which
     *  is maintained statically by the StateMonitor class.  The
     *  target property's value is set to whatever the valueMapper
     *  returns, or, if the valueMapper is null, to the value of
     *  the application property.
     *  @param target the Object whose property will be set
     *  @param targetProperty the property of the Object that will be set
     *  @param applicationProperty the application property according to which the target property will vary
     *  @param valueMapper the Mapper that maps application values to target values, or null 
     *  @param domainMatcher the DomainMatcher that tests domainKeys, or null
     */
    public static void registerProperty(Object target, 
					String targetProperty, 
					String applicationProperty,
					Mapper valueMapper,
					DomainMatcher domainMatcher)
    {
	unregisterProperty(target, 
			   targetProperty, 
			   applicationProperty);
	monitors.add(new StateMonitor(target, 
				      targetProperty, 
				      applicationProperty, 
				      valueMapper,
				      domainMatcher));
    }

    public static void unregisterProperty(Object target, 
					  String targetProperty, 
					  String applicationProperty)
    {
	for (Iterator it=monitors.iterator();it.hasNext();)
	{
	    StateMonitor sm=(StateMonitor) it.next();
	    if (sm.target.equals(target)
		&& sm.targetProperty.equals(targetProperty)
		&& sm.applicationProperty.equals(applicationProperty))
	    {
		monitors.remove(sm);
		break; 
	    }
	}
    }

    public static final void setProperty(String applicationProperty, Object value, Object domainKey)
    {
// 	applicationValues.put(applicationProperty, value);
	for (Iterator it=monitors.iterator();it.hasNext();)
	{
	    StateMonitor nextMonitor=(StateMonitor)it.next();
	    if (nextMonitor.applicationProperty.equals(applicationProperty))
	    {
		DomainMatcher dm=nextMonitor.domainMatcher;
		if (dm==null || dm.domainMatches(domainKey))
		    nextMonitor.configureTarget(value);
	    }
	}
    }

    public interface DomainMatcher
    {
	boolean domainMatches(Object domainKey);
    }
    
    /**
     *  maps values of the application property to values of the target property 
     */
    public interface Mapper
    {
	/**
	 *  @param applicationValue the application value of the registered application property
	 *  @return target value that the StateMonitor should set registered target properties 
	 */
	Object getTargetValue(Object applicationValue);

	/**
	 * @return whether or not the Mapper can map this value
	 */
	boolean canMapValue(Object applicationValue);
    }

    /**
     * @return a Mapper object that simply wraps a Map object
     */
    public static Mapper getMapper(final Map m)
    {
	return new Mapper() {
		public Object getTargetValue(Object applicationValue)
		{
		    return (m.containsKey(applicationValue))
			? m.get(applicationValue)
			: null;
		}
		public boolean canMapValue(Object applicationValue)
		{
		    return m.containsKey(applicationValue);
		}
	    };
    }
			
    private StateMonitor(Object target, 
			 String targetProperty, 
			 String applicationProperty, 
			 Mapper valueMapper, 
			 DomainMatcher domainMatcher)
    {
	this.target=target;
	this.targetProperty=targetProperty;
	this.applicationProperty=applicationProperty;
	this.valueMapper=valueMapper;
	this.domainMatcher=domainMatcher;
    }

    private void configureTarget(Object applicationValue)
    {
	if (valueMapper==null || valueMapper.canMapValue(applicationValue))
	{
	    try
	    {
		if (setter==null) //lazy initialization of setter method
		{
		    BeanInfo info=Introspector.getBeanInfo(target.getClass());
		    PropertyDescriptor[] propDescriptors=info.getPropertyDescriptors();
		    for (int i=0;i<propDescriptors.length;i++)
		    {
			PropertyDescriptor pd=propDescriptors[i];
			String propName=pd.getName();
			if (propName.equals(targetProperty))
			{
			    setter=pd.getWriteMethod();
			    break;
			}
		    }
		}
		if (setter!=null)
		{
		    Object targetValue= (valueMapper!=null)
			? valueMapper.getTargetValue(applicationValue)
			: applicationValue;
		    setter.invoke(target, 
				  new Object[] { targetValue });
		}
		else 
		{
		    Debug.trace(this, Debug.DP2, "*** Error: setter not found for " + targetProperty);
		}

	    }
	    catch (IntrospectionException intro)
	    {
		Debug.trace(this, Debug.DP2, intro);
	    }
	    catch (IllegalAccessException illAcc)
	    {
		Debug.trace(this, Debug.DP2, illAcc);
	    }
	    catch (IllegalArgumentException illArg)
	    {
		Debug.trace(this, Debug.DP2, illArg);
	    }
	    catch (InvocationTargetException invoTarg)
	    {
		Debug.trace(this, Debug.DP2, invoTarg);
	    }
	}
    }
}

/* $Log: StateMonitor.java,v $
/* Revision 1.7  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.6  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.5  2000/11/28 00:01:39  smullyan
/* added a status bar/minibuffer, with a location field showing the current line and
/* column number (for the SimpleTextEditor and kin only).
/*
/* Revision 1.4  2000/11/23 04:37:32  smullyan
/* editor and explorer now synch their files using the StateMonitor, which has
/* been improved significantly
/*
/* Revision 1.3  2000/11/22 00:11:27  smullyan
/* editor now locks and unlocks, more or less appropriately.
/* */
