/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.Component;
import java.awt.Point;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.table.TableModel;
import org.skunk.assertF.Assertion;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.Lock;
import org.skunk.dav.client.gui.action.CopyAction;
import org.skunk.dav.client.gui.action.DeleteAction;
import org.skunk.dav.client.gui.action.DownloadAction;
import org.skunk.dav.client.gui.action.LockAction;
import org.skunk.dav.client.gui.action.MoveAction;
import org.skunk.dav.client.gui.action.NewCollectionAction;
import org.skunk.dav.client.gui.action.NewFileAction;
import org.skunk.dav.client.gui.action.OpenAction;
import org.skunk.dav.client.gui.action.PropertiesAction;
import org.skunk.dav.client.gui.action.RefreshAction;
import org.skunk.dav.client.gui.action.RenameAction;
import org.skunk.dav.client.gui.action.StealLockAction;
import org.skunk.dav.client.gui.action.UnlockAction;
import org.skunk.dav.client.gui.action.UploadAction;
import org.skunk.trace.Debug;

/**
 * this listens for mouse events on both the Explorer's table 
 * and the scrollpane that contains it.
 */
public class TableMouser extends MouseAdapter
{
    private Explorer ex;
    
    //the six possible relevant lock states for a file
    private static final int UNLOCKED=0;
    private static final int OWN_SHARED_LOCK=1;
    private static final int OTHER_SHARED_LOCK=2;
    private static final int OWN_EXCLUSIVE_LOCK=3;
    private static final int OTHER_STEALABLE_EXCLUSIVE_LOCK=4;
    private static final int OTHER_UNSTEALABLE_EXCLUSIVE_LOCK=5;
    

    public TableMouser(Explorer ex)
    {
	super();
	this.ex=ex;
    }

    public void mousePressed(MouseEvent me)
    {
	if (SwingUtilities.isRightMouseButton(me) 
	    && ex.getSelectedNode()!=null)
	    handlePopup(me);
    }
	

    public void mouseClicked(MouseEvent me)
    {
	if ((me.getSource() instanceof JTable)  
	    && me.getClickCount()>1)
	{
	    handleDoubleClick(me);
	}
    }

    private DAVFile getDAVFileFor(MouseEvent me)
    {
	Object source=me.getSource();
	Assertion.assertF((source instanceof JTable), 
			 "source is a JTable");
	JTable table=(JTable) source;
	Point p=new Point(me.getX(), me.getY());
	int row=table.rowAtPoint(p);
	TableModel tm=table.getModel();
	Assertion.assertF((tm instanceof DAVTableModel), 
			 "TableModel is a DAVTableModel");
	DAVTableModel dtm=(DAVTableModel) tm;
	DAVFile file=dtm.getDAVFile(row);
	return file;
    }	

    private void handlePopup(MouseEvent me)
    {
	Object sourceObj=me.getSource();
	JPopupMenu popper;
	if (sourceObj instanceof JTable)
	{
	    DAVFile file=getDAVFileFor(me);
	    Point p=new Point(me.getX(), me.getY());
	    popper=getPopup(file);
	}
	else popper=getNoFilePopup();   
	Assertion.assertF((sourceObj instanceof Component), 
			 "{0} is a Component", 
			 new Object[] {sourceObj});
	popper.show((Component) sourceObj, me.getX(), me.getY());
    }

    private void handleDoubleClick(MouseEvent me)
    {
	DAVFile file=getDAVFileFor(me);
	if (file.isCollection())
	{
	    Debug.trace(this, Debug.DP3, "now opening collection " + file.getFileName());
	    ex.displayCollection(file.getFileName());
	}
	else 
	{
	    new OpenAction(ex, file).actionPerformed(null);
	}
    }

    private int getLockState(DAVFile file)
    {
	Assertion.assertF((file!=null), "file is not null");
	if (file.isLocked())
	{
	    String lockOwner=file.getExclusiveLockOwner();
	    ServerData sd=ex.getSelectedConnection();
	    String localLockOwner=sd.getOwner();
	    if (localLockOwner==null)
		localLockOwner=sd.getUsername();
	    if (file.isExclusiveLocked())
	    {
		if (lockOwner!=null && lockOwner.equals(localLockOwner))
		{
		    return OWN_EXCLUSIVE_LOCK;
		}
		else
		{
		    if (sd.getPermitsLockStealing())
		    {
			return OTHER_STEALABLE_EXCLUSIVE_LOCK;
		    }
		    else
		    {
			return OTHER_UNSTEALABLE_EXCLUSIVE_LOCK;
		    }
		}
	    }
	    else
	    {
		Lock sharedLock=file.getLock(localLockOwner);
		if (sharedLock!=null)
		{
		    return OWN_SHARED_LOCK;
		}
		else
		{
		    return OTHER_SHARED_LOCK;
		}
	    }
	}
	else return UNLOCKED;
    }

    private JPopupMenu getPopup(DAVFile file)
    {
	if (file==null)
	    return getNoFilePopup();
	int lockState=getLockState(file);
	JPopupMenu popper=new JPopupMenu();
	/*
	  Open 
	  Copy...
	  Delete
	  Move...
	  Rename
	  Lock/Unlock/Steal Lock
	  Upload
	  Download
	  Properties
	 */
	JMenuItem handleItem=new JMenuItem();
	handleItem.setEnabled(false);
	popper.add(handleItem);
	JMenuItem temp=popper.add(new OpenAction(ex, file));
	temp.setText(ResourceManager.getMessage(ResourceManager.OPEN_KEY));
	temp=popper.add(new CopyAction(ex, file));
	temp.setText(ResourceManager.getMessage(ResourceManager.COPY_KEY));
	temp=popper.add(new DeleteAction(ex, file.getFileName()));
	//if file is locked, disable delete action, but still display it 
	temp.setText(ResourceManager.getMessage(ResourceManager.DELETE_KEY));
	if (lockState!=UNLOCKED) temp.setEnabled(false);

	temp=popper.add(new MoveAction(ex, file));
	temp.setText(ResourceManager.getMessage(ResourceManager.MOVE_KEY));	
	if (lockState==OTHER_UNSTEALABLE_EXCLUSIVE_LOCK) temp.setEnabled(false);

	temp=popper.add(new RenameAction(ex, file));
	temp.setText(ResourceManager.getMessage(ResourceManager.RENAME_KEY));
	if (lockState==OTHER_UNSTEALABLE_EXCLUSIVE_LOCK) temp.setEnabled(false);

	    /*
	      if exclusive:
	         check lock owner and compare with the owner/username in serverdata;
	         if owner is not same: 
		    show Steal Lock; 
		 else: 
		    show Unlock.
	      else:
	         if there is a shared lock matching the current principal:
	            include an Unlock 
		 else: 
		    include a Lock
	    */

	if (lockState==OWN_EXCLUSIVE_LOCK || lockState==OWN_SHARED_LOCK)
	{
	    //principals match
	    popper.add(new UnlockAction(ex, file))
		.setText(ResourceManager.getMessage(ResourceManager.UNLOCK_KEY));	
	}
	else if (lockState==OTHER_STEALABLE_EXCLUSIVE_LOCK)
	{
	    temp=popper.add(new StealLockAction(ex, file));
	    temp.setText(ResourceManager.getMessage(ResourceManager.STEAL_LOCK_KEY));
	}
	else if (lockState==UNLOCKED || lockState==OTHER_SHARED_LOCK)
	{
	    popper.add(new LockAction(ex, file))
		.setText(ResourceManager.getMessage(ResourceManager.LOCK_KEY));
	}	    
	if (!file.isCollection())
	{
	    popper.add(new JSeparator());
	    temp=popper.add(new UploadAction(ex, file));
	    temp.setText(ResourceManager.getMessage(ResourceManager.UPLOAD_KEY));
	    if (lockState==OTHER_UNSTEALABLE_EXCLUSIVE_LOCK) temp.setEnabled(false);
	    popper.add(new DownloadAction(ex, file))
		.setText(ResourceManager.getMessage(ResourceManager.DOWNLOAD_KEY));
	}
	popper.add(new JSeparator());
	popper.add(new PropertiesAction(ex, file))
	    .setText(ResourceManager.getMessage(ResourceManager.PROPERTIES_KEY));
	return popper;
    }

    private JPopupMenu getNoFilePopup()
    {
	JPopupMenu popper=new JPopupMenu();
	/*
	  Refresh
	  New File
	  New Directory
	*/	
	JMenuItem handleItem=new JMenuItem();
	handleItem.setEnabled(false);
	popper.add(handleItem);
	String label=ResourceManager.getMessage(ResourceManager.NEW_FILE_KEY);
	popper.add(new NewFileAction(ex)).setText(label);
	label=ResourceManager.getMessage(ResourceManager.NEW_COLLECTION_KEY);
	popper.add(new NewCollectionAction(ex)).setText(label);
	label=ResourceManager.getMessage(ResourceManager.REFRESH_KEY);
	popper.add(new RefreshAction(ex)).setText(label);
	popper.add(new JSeparator());
	label=ResourceManager.getMessage(ResourceManager.PROPERTIES_KEY);
	popper.add(new PropertiesAction(ex)).setText(label);
	return popper;
    }
}

/* $Log: TableMouser.java,v $
/* Revision 1.16  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.15  2000/12/04 23:51:16  smulloni
/* added ImageViewer; fixed word in SimpleTextEditor
/*
/* Revision 1.14  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.13  2000/11/22 00:11:28  smullyan
/* editor now locks and unlocks, more or less appropriately.
/*
/* Revision 1.12  2000/11/15 19:45:36  smullyan
/* beginning of revamp of application to include multiple buffers.
/*
/* Revision 1.11  2000/11/13 04:56:42  smullyan
/* improved the organization of the code, which had become quasi-monstrous.
/*
/* Revision 1.10  2000/11/10 22:40:07  smullyan
/* added icon to table for resource type; fixes to copy and move; disabling of
/* menu items that are inappropriate.
/*
/* Revision 1.9  2000/11/09 23:35:00  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
