/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import javax.swing.AbstractListModel;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.ComboBoxModel;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.border.TitledBorder;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import org.skunk.assertF.Assertion;
import org.skunk.config.Configurator;
import org.skunk.config.ConfigStore;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.gui.editor.ConfigurableEditor;
import org.skunk.dav.client.gui.editor.SimpleTextEditor;
import org.skunk.swing.ColorChoiceButton;
import org.skunk.swing.text.EntryDocument;
import org.skunk.swing.text.PositiveIntegerEntryFilter;
import org.skunk.swing.text.TextEditorPane;
import org.skunk.swing.text.syntax.FileMode;
import org.skunk.swing.text.syntax.SyntaxDocument;
import org.skunk.swing.text.syntax.SyntaxEditorKit;
import org.skunk.swing.text.syntax.SyntaxStyle;
import org.skunk.trace.Debug;

/**
 * customizes the editor's appearance and behavior.
 * 
 * @author Jacob Smullyan ($Author: smulloni $) <smulloni@krinst.org>
 * @version $Revision: 1.7 $
 */
public class TextEditorCustomizer extends PreferencesCustomizer
{
    private static HashMap previewHash;

    private JButton                     removeExtensionButton;
    private JPanel                      customizer;
    private PreviewPane                 previewPane;
    private ColorChoiceButton           globalBackgroundButton, 
	                                globalForegroundButton, 
	                                globalCaretColorButton,
                                        styleColorButton;
    private JCheckBox                   highlightModeCheckBox,
                                        boldCheckBox,
                                        italicCheckBox;
    private JComboBox                   fontDropdown,
                                        styleDropdown;
    private StyleComboBoxModel          styleModel;
    private FileModeExtensionListModel  extModel;
    private JList                       extensionsList, 
                                        fileModeList;
    private JTextField                  fontSizeField, 
                                        tabSizeField;
    private Color                       globalForeground,
	                                lastForeground,
	                                globalBackground, 
                                        lastBackground,
	                                globalCaretColor, 
                                        lastCaretColor;
    private Font                        defaultFont,
	                                lastFont;
    private int                         tabSize,
	                                lastTabSize;
    private String                      lastStyleConfigData;

    private SyntaxStyle.SyntaxStyleSet  configSet;
    private FileMode.ModeMapContainer   modeSet;

    /* private String lineSeparator; */ //for future release

    private static final int USE_DEFAULT=0;
    private static final int USE_REVERT=1;
    private static final int USE_CONFIG=2;

    static
    {
	previewHash=new HashMap();
    }

    private static String slurp(String resourceName) throws IOException
    {
	InputStream in=TextEditorCustomizer.class.getResourceAsStream(resourceName);
	if (in ==null) throw new IOException("resourceName not found");
	Reader r=new InputStreamReader(in);
	StringWriter sw=new StringWriter();
	int charsRead;
	int offset=0;
	char[] buffer=new char[1<<12];
	while ((charsRead=r.read(buffer, offset, buffer.length))>0)
	{
	    sw.write(buffer, offset, charsRead);
	}
	sw.flush();
	r.close();
	sw.close();
	return sw.toString();
    }

    private static String getPreviewText(String mode)
    {
	if (previewHash.containsKey(mode))
	    return (String) previewHash.get(mode);
	else 
	{
	    try
	    {
		String s=slurp(mode+".example");
		previewHash.put(mode, s);
		return s;
	    }
	    catch (IOException oyVeh)
	    {
		return ResourceManager.getMessage(ResourceManager.TEXT_EDITOR_CUSTOMIZER_IO_EXCEPTION, 
						  new Object[] {mode, oyVeh.getMessage()});
	    }
	}
    }

    /**
     * constructor
     * @param title
     * @param configurator
     */
    public TextEditorCustomizer(String title, Configurator configurator)
    {
	super(title, configurator);
	initConfig();
	initComponents();
    }

    private Object getConfiguredValue(Class propertyClass, String propertyName, boolean useRevert)
    {
	ConfigStore store=(useRevert) 
	    ? getConfigurator().getReversionStore()
	    : getConfigurator().getStore();
	return store.getConfigValue(propertyClass, propertyName);
    }

    private Color getConfiguredGlobalBackground(int mode)
    {
	if (mode!=USE_DEFAULT)
	{
	    Object val=getConfiguredValue(SimpleTextEditor.class, 
					  SimpleTextEditor.BACKGROUND_PROPERTY, 
					  (mode==USE_REVERT));
	    if (mode==USE_REVERT && val==null) 
	    {
		globalBackground=lastBackground;
		return globalBackground;
	    }
	    lastBackground=globalBackground;
	    this.globalBackground=(val!=null && val instanceof Color) 
	    ? (Color) val
	    : SyntaxStyle.DEFAULT_BACKGROUND;
	}
	else
	{
	    this.globalBackground=SyntaxStyle.DEFAULT_BACKGROUND;
	    lastBackground=globalBackground;
	}
	if (lastBackground==null) lastBackground=globalBackground;
	return globalBackground;
    }

    private Color getConfiguredGlobalForeground(int mode)
    {
	if (mode!=USE_DEFAULT)
	{
	    Object val=getConfiguredValue(SimpleTextEditor.class, 
					  SimpleTextEditor.FOREGROUND_PROPERTY,
					  mode==USE_REVERT);
	    if (mode==USE_REVERT && val==null) 
	    {
		globalForeground=lastForeground;
		return globalForeground;
	    }
	    lastForeground=globalForeground;
	    this.globalForeground=(val!=null && val instanceof Color) 
		? (Color) val
		: SyntaxStyle.DEFAULT_FOREGROUND;
	}
	else 
	{
	    this.globalForeground=SyntaxStyle.DEFAULT_FOREGROUND;
	    lastForeground=globalForeground;
	}
	if (lastForeground==null) lastForeground=globalForeground;
	return globalForeground;
    }

    private Color getConfiguredCaretColor(int mode)
    {
	if (mode!=USE_DEFAULT)
	{
	    Object val=getConfiguredValue(SimpleTextEditor.class, 
					  SimpleTextEditor.CARET_COLOR_PROPERTY,
					  mode==USE_REVERT);
	    if (mode==USE_REVERT && val==null) 
	    {
		globalCaretColor=lastCaretColor;
		return globalCaretColor;
	    }
	    lastCaretColor=globalCaretColor;
	    this.globalCaretColor=(val!=null && val instanceof Color)
		? (Color) val
		: SimpleTextEditor.DEFAULT_CARET_COLOR;
	}
	else 
	{
	    this.globalCaretColor=SimpleTextEditor.DEFAULT_CARET_COLOR;
	    lastCaretColor=globalCaretColor;
	}
	if (lastCaretColor==null) lastCaretColor=globalCaretColor;
	return globalCaretColor;
    }

    private Font getConfiguredFont(int mode)
    {
	if (mode!=USE_DEFAULT)
	{
	    Object val=getConfiguredValue(SimpleTextEditor.class, 
					  SimpleTextEditor.FONT_PROPERTY,
					  mode==USE_REVERT);
	    if (mode==USE_REVERT && val==null) 
	    {
		defaultFont=lastFont;
		return defaultFont;
	    }
	    lastFont=defaultFont;
	    boolean gotAFont=val!=null && (val instanceof Font);
	    defaultFont=(gotAFont) ? (Font) val : SyntaxStyle.DEFAULT_FONT;
	    
	}
	else 
	{
	    defaultFont=SyntaxStyle.DEFAULT_FONT;
	    lastFont=defaultFont;
	}
	if (lastFont==null) lastFont=defaultFont;
	return defaultFont;
    }

    private int getConfiguredTabSize(int mode)
    {
	if (mode!=USE_DEFAULT)
	{
	    Object val=getConfiguredValue(SimpleTextEditor.class, 
					  SimpleTextEditor.TAB_SIZE_PROPERTY,
					  mode==USE_REVERT);
	    if (mode==USE_REVERT && val==null) 
	    {
		tabSize=lastTabSize;
		return tabSize;
	    }
	    lastTabSize=tabSize;
	    this.tabSize=(val!=null && val instanceof Integer)
		? ((Integer)val).intValue()
		: SimpleTextEditor.DEFAULT_TAB_SIZE;
	}
	else 
	{
	    this.tabSize=SimpleTextEditor.DEFAULT_TAB_SIZE;
	    lastTabSize=tabSize;
	}
	if (lastTabSize<=0) lastTabSize=tabSize;
	return this.tabSize;
    }

    private SyntaxStyle.SyntaxStyleSet getConfiguredStyleSet(int mode)
    {
	if (this.configSet==null) 
	    this.configSet=new SyntaxStyle.SyntaxStyleSet();
	switch (mode)
	{
	case USE_CONFIG:
	    lastStyleConfigData=configSet.getConfigData();
	    getConfigurator().configure(configSet);
	    break;
	case USE_REVERT:
	    Object val=getConfiguredValue(SyntaxStyle.SyntaxStyleSet.class,
					  SyntaxStyle.SyntaxStyleSet.CONFIG_DATA_PROPERTY,
					  true);
	    if (val!=null)
	    {
		if (Assertion.ASSERT) 
		    Assertion.assertF((val instanceof String), 
				     "config data is a string");
		configSet.setConfigData((String) val);
	    }
	    else
	    {
		configSet.setConfigData(lastStyleConfigData);
	    }
	    break;
	case USE_DEFAULT:
	default:
	    SyntaxStyle.initDefaultStyles(configSet);
	}
	return this.configSet;
    }

    private FileMode.ModeMapContainer getConfiguredModeSet(int mode)
    {
	if (modeSet==null) 
	    modeSet=new FileMode.ModeMapContainer();
	switch (mode)
	{
	case USE_CONFIG:
	    getConfigurator().configure(modeSet);
	    break;
	case USE_REVERT:
	    /*
	      I'm not checking for a null value in revert here, because
	      I'm not currently updating the editor display on the basis of the
	      these values.  Should I do so, I would have to use the same
	      kludge here as I am doing elsewhere.  JS
	     */
	    getConfigurator().revert(modeSet);
	    break;
	case USE_DEFAULT:
	default:
	    FileMode.initDefaultModes(modeSet);
	}
	if (Debug.DEBUG) Debug.trace(this, Debug.DP5, "modeSet: "+modeSet.getModeMap());
	return this.modeSet;
    }

    private void addEditorTargets()
    {
	AppContext context=ExplorerApp.getAppContext();
	if (context!=null)
	{
	    for (Iterator it1=context.views(); it1.hasNext();)
	    {
		View v=(View) it1.next();
		for (Iterator it2=v.getDockedBuffers();it2.hasNext();)
		{
		    Object buffer=it2.next();
		    if (buffer instanceof SimpleTextEditor)
		    {
			if (Debug.DEBUG) Debug.trace(this, Debug.DP4, "text editor found");
			addConfigTarget(buffer);
		    }
		}
	    }
	}
    }

    private void initConfig()
    {
	addEditorTargets();
	addConfigTarget(FileMode.getModeMapContainer());
	addConfigTarget(SyntaxStyle.getDefaultStyleSet());
	//this initialization deals with the case where there are no saved preferences
	getConfiguredStyleSet(USE_DEFAULT);
	getConfiguredModeSet(USE_DEFAULT);

	getConfiguredGlobalBackground(USE_CONFIG);
	getConfiguredGlobalForeground(USE_CONFIG);
	getConfiguredCaretColor(USE_CONFIG);
	getConfiguredFont(USE_CONFIG);
	getConfiguredStyleSet(USE_CONFIG);
	getConfiguredTabSize(USE_CONFIG);
	getConfiguredModeSet(USE_CONFIG);
    }

    private void _setConfigValues()
    {
	setConfigValue(SimpleTextEditor.class, 
		       SimpleTextEditor.FOREGROUND_PROPERTY,
		       globalForeground);
	setConfigValue(SimpleTextEditor.class,
		       SimpleTextEditor.BACKGROUND_PROPERTY,
		       globalBackground);
	setConfigValue(SimpleTextEditor.class,
		       SimpleTextEditor.CARET_COLOR_PROPERTY,
		       globalCaretColor);
	setConfigValue(SimpleTextEditor.class,
		       SimpleTextEditor.FONT_PROPERTY,
		       defaultFont);
	setConfigValue(SyntaxStyle.SyntaxStyleSet.class,
		       SyntaxStyle.SyntaxStyleSet.DEFAULT_FONT_PROPERTY,
		       defaultFont);
	setConfigValue(SimpleTextEditor.class,
		       SimpleTextEditor.TAB_SIZE_PROPERTY,
		       new Integer(tabSize));
	setConfigValue(FileMode.ModeMapContainer.class,
		       FileMode.ModeMapContainer.CONFIG_DATA_PROPERTY,
		       modeSet.getConfigData());
	setConfigValue(SyntaxStyle.SyntaxStyleSet.class,
		       SyntaxStyle.SyntaxStyleSet.CONFIG_DATA_PROPERTY,
		       configSet.getConfigData());
    }

    /**
     * workaround for a weird bug on IBM JDK1.2 for Linux  -- calling getFamily may 
     * yield a name entirely in lowercase,
     * but list of all family names are capitalized. 
     */
    private String getFontName(Font f)
    {
	String fName=f.getFamily(Locale.getDefault());

	if (!Character.isUpperCase(fName.charAt(0)))
	{
	    StringBuffer sb=new StringBuffer();
	    sb.append(Character.toUpperCase(fName.charAt(0)));
	    sb.append(fName.substring(1));
	    fName=sb.toString();
	}
	return fName;
    }

    public void ok()
    {
	_setConfigValues();
	addEditorTargets();
	super.ok();
	lastBackground=globalBackground;
	lastForeground=globalForeground;
	lastFont=defaultFont;
	lastCaretColor=globalCaretColor;
	lastTabSize=tabSize;
    }

    public void tryOn()
    {
	_setConfigValues();
	addEditorTargets();	
	super.tryOn();
    }

    public void revert()
    {
	synchConfig(USE_REVERT);
	synchGlobalControls();
	_setConfigValues();
	addEditorTargets();
	super.revert();
	//getConfigurator().tryOff(SimpleTextEditor.class);
	synchFileModes();
	synchStyles();
	synchPreview();	
    }

    public void revertToDefault()
    {
	synchConfig(USE_DEFAULT);
	synchGlobalControls();
	_setConfigValues();
	addEditorTargets();
	super.revertToDefault();
	synchFileModes();
	synchStyles();
	synchPreview();
    }

    private void synchFileModes()
    {
	((FileModeListModel)fileModeList.getModel()).refresh();
	FileMode mode=modeSet.getMode((String)fileModeList.getSelectedValue());
	((SyntaxDocument)previewPane.getDocument()).setFileMode(mode);
	extModel.setFileMode(mode);
	if (extModel.getSize()>0) 
	{
	    extensionsList.setSelectedIndex(0);
	    removeExtensionButton.setEnabled(true);
	}
	else removeExtensionButton.setEnabled(false);
	highlightModeCheckBox.setSelected(mode.getShouldHighlight());
	highlightModeCheckBox.setEnabled(mode.getCanHighlight());
    }

    private void synchStyles()
    {
	Object item=styleDropdown.getSelectedItem();
	styleModel.refresh();
	styleDropdown.setSelectedItem(item);
	SyntaxStyle ss=configSet.getStyle((String) item);
	styleColorButton.setColor(ss.getForegroundColor());
	boldCheckBox.setSelected(ss.isBold());
	italicCheckBox.setSelected(ss.isItalic());	
    }

    private void synchPreview()
    {
	previewPane.setBackground(globalBackground);
	previewPane.setForeground(globalForeground);
	previewPane.setCaretColor(globalCaretColor);
	previewPane.setFont(defaultFont);
	previewPane.setTabSize(tabSize);
	previewPane.repaint();
    }

    private void synchGlobalControls()
    {
	globalBackgroundButton.setBackground(globalBackground);
	globalForegroundButton.setBackground(globalForeground);
	globalCaretColorButton.setBackground(globalCaretColor);
	fontDropdown.setSelectedItem(getFontName(defaultFont));
	fontSizeField.setText(Integer.toString(defaultFont.getSize()));
	tabSizeField.setText(Integer.toString(tabSize));
    }

    private void synchConfig(int mode)
    {
	getConfiguredGlobalBackground(mode);
 	getConfiguredGlobalForeground(mode);
	getConfiguredCaretColor(mode);
	getConfiguredTabSize(mode);
	getConfiguredFont(mode);
	getConfiguredStyleSet(mode);
	getConfiguredModeSet(mode);
    }

    private GridBagConstraints initGBC()
    {
	GridBagConstraints gbc=new GridBagConstraints();
	gbc.gridx=0;
	gbc.gridy=0;
	gbc.gridwidth=1;
	gbc.gridheight=1;
	gbc.insets=new Insets(2, 2, 2, 2);
	gbc.fill=GridBagConstraints.NONE;
	gbc.ipadx=1;
	gbc.ipady=1;
	gbc.weightx=0.33;
	gbc.weighty=0.0;
	gbc.anchor=GridBagConstraints.WEST;
	return gbc;
    }

    private  JPanel initGlobal()
    {
	String globalPropertiesStr=ResourceManager.getMessage(ResourceManager.GLOBAL_PROPERTIES);
	JLabel globalBackgroundLbl=new JLabel(ResourceManager.getMessage(ResourceManager.GLOBAL_BACKGROUND));
	JLabel globalForegroundLbl=new JLabel(ResourceManager.getMessage(ResourceManager.GLOBAL_FOREGROUND));
	JLabel globalCaretColorLbl=new JLabel(ResourceManager.getMessage(ResourceManager.GLOBAL_CARET_COLOR));
	JLabel fontNameLbl=new JLabel(ResourceManager.getMessage(ResourceManager.FONT_NAME));
	JLabel fontSizeLbl=new JLabel(ResourceManager.getMessage(ResourceManager.FONT_SIZE));
	JLabel tabSizeLbl=new JLabel(ResourceManager.getMessage(ResourceManager.TAB_SIZE));
	/* 
	//not yet used
	JLabel lineSeparatorLbl=new JLabel(ResourceManager.getMessage(ResourceManager.LINE_SEPARATOR));
	String[] separatorOptions=new String[] 
	{
	ResourceManager.getMessage(ResourceManager.LF_LINE_SEPARATOR),
	ResourceManager.getMessage(ResourceManager.CRLF_LINE_SEPARATOR),
	ResourceManager.getMessage(ResourceManager.CR_LINE_SEPARATOR)
	};
	*/
	globalForegroundButton=new ColorChoiceButton();
	globalForegroundButton.setColor(globalForeground);
	globalForegroundButton.addPropertyChangeListener(new PropertyChangeListener()
	    {
		public void propertyChange(PropertyChangeEvent cheese)
		{
		    if (cheese.getPropertyName().equals(ColorChoiceButton.COLOR_PROPERTY))
		    {
			Color c=globalForegroundButton.getColor();
			if (c.equals(TextEditorCustomizer.this.globalForeground))
			    return;
			TextEditorCustomizer.this.globalForeground=c;
			TextEditorCustomizer.this.previewPane.setForeground(globalForeground);
			TextEditorCustomizer.this.setDirty(true);
		    }
		}
	    });

	globalBackgroundButton=new ColorChoiceButton();
	globalBackgroundButton.setColor(globalBackground);
	globalBackgroundButton.addPropertyChangeListener(new PropertyChangeListener()
	    {
		public void propertyChange(PropertyChangeEvent cheese)
		{
		    if (cheese.getPropertyName().equals(ColorChoiceButton.COLOR_PROPERTY))
		    {
			Color c=globalBackgroundButton.getColor();
			if (TextEditorCustomizer.this.globalBackground.equals(c)) return;
			TextEditorCustomizer.this.globalBackground=c;
			TextEditorCustomizer.this.previewPane.setBackground(globalBackground);
			TextEditorCustomizer.this.setDirty(true);
		    }
		}
	    });
	globalCaretColorButton=new ColorChoiceButton();
	globalCaretColorButton.setColor(globalCaretColor);
	globalCaretColorButton.addPropertyChangeListener(new PropertyChangeListener()
	    {
		public void propertyChange(PropertyChangeEvent cheese)
		{
		    if (cheese.getPropertyName().equals(ColorChoiceButton.COLOR_PROPERTY))
		    {
			Color c=globalCaretColorButton.getColor();
			if (TextEditorCustomizer.this.globalCaretColor.equals(c)) return;
			TextEditorCustomizer.this.globalCaretColor=c;
			TextEditorCustomizer.this.previewPane.setCaretColor(globalCaretColor);
			TextEditorCustomizer.this.setDirty(true);
		    }
		}
	    });
	String[] fontNames=GraphicsEnvironment.getLocalGraphicsEnvironment()
	    .getAvailableFontFamilyNames(Locale.getDefault());
	Arrays.sort(fontNames);
	fontDropdown=new JComboBox(fontNames);
	/*
	 * on IBM JDK1.3 for Linux, a blank item is returned, for some reason.  
	 * Maybe it the font of our dreams, but I don't want it.
	 */
	fontDropdown.removeItem("");
	fontDropdown.setSelectedItem(getFontName(defaultFont));

	fontDropdown.addItemListener(new ItemListener()
	    {
		public void itemStateChanged(ItemEvent ie)
		{
		    if (ie.getStateChange()==ItemEvent.SELECTED)
		    {
			Object o=ie.getItem();
			String fontName=(String) o;
			int fontSize=Integer.parseInt(fontSizeField.getText());
			TextEditorCustomizer.this.defaultFont=new Font(fontName, 
								       Font.PLAIN, 
								       fontSize);
			if (TextEditorCustomizer.this.configSet.getDefaultFont().equals(defaultFont))
			    return;
			TextEditorCustomizer.this.configSet.setDefaultFont(defaultFont);
			TextEditorCustomizer.this.previewPane.setFont(defaultFont);			
			setDirty(true);
		    }
		}
	    });
	fontSizeField=new JTextField(3);
	fontSizeField.setDocument(new EntryDocument(new PositiveIntegerEntryFilter(1, 72)));
	fontSizeField.setText(Integer.toString(defaultFont.getSize()));
	tabSizeField=new JTextField(2);
	tabSizeField.setDocument(new EntryDocument(new PositiveIntegerEntryFilter(1, 12)));
	tabSizeField.setText(Integer.toString(tabSize));
	//listener for both tabSize  and fontSize
	ActionListener tabAndFontSizeListener=new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    try
		    {
			Integer tabInt=new Integer(tabSizeField.getText());
			tabSize=tabInt.intValue();
			previewPane.setTabSize(tabSize); 
			int fontSize=Integer.parseInt(fontSizeField.getText());
			defaultFont=new Font(getFontName(defaultFont), Font.PLAIN, fontSize);
			previewPane.setFont(defaultFont);
			configSet.setDefaultFont(defaultFont);
			setDirty(true);
		    }
		    catch (NumberFormatException nafta)
		    {
			//nothing
		    }
		}
	    };
	    
	fontSizeField.addActionListener(tabAndFontSizeListener);
	tabSizeField.addActionListener(tabAndFontSizeListener);

	GridBagConstraints gbc=initGBC();
	JPanel tmp=new JPanel(new GridBagLayout());

	tmp.add(globalBackgroundLbl, gbc);
	gbc.gridx++;
	tmp.add(globalBackgroundButton, gbc);
	gbc.gridx++;
	
	tmp.add(globalForegroundLbl, gbc);
	gbc.gridx++;	
	tmp.add(globalForegroundButton, gbc);
	gbc.gridx++;

	tmp.add(globalCaretColorLbl, gbc);
	gbc.gridx++;
	tmp.add(globalCaretColorButton, gbc);
	gbc.gridx=0;
	gbc.gridy++;
	
	tmp.add(fontNameLbl, gbc);
	gbc.gridx++;
	tmp.add(fontDropdown, gbc);
	gbc.gridx++;

	tmp.add(fontSizeLbl, gbc);
	gbc.gridx++;
	tmp.add(fontSizeField, gbc);

	gbc.gridx++;
	tmp.add(tabSizeLbl, gbc);
	gbc.gridx++;
	tmp.add(tabSizeField, gbc);
	
	tmp.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
						       globalPropertiesStr,
						       TitledBorder.LEFT,
						       TitledBorder.TOP));
	return tmp;
    }

    private JPanel initFileModes()
    {
	String fileModePropertiesStr=ResourceManager.getMessage(ResourceManager.FILE_MODE_PROPERTIES);
	JLabel fileModesLbl=new JLabel(ResourceManager.getMessage(ResourceManager.FILE_MODES));
	JLabel extensionsLbl=new JLabel(ResourceManager.getMessage(ResourceManager.EXTENSIONS));

	final String newExtensionDialogTitle=ResourceManager.getMessage(ResourceManager.NEW_EXTENSION_DIALOG_TITLE);
	fileModeList=new JList(new FileModeListModel(modeSet));
	fileModeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	extModel=new FileModeExtensionListModel();
	extensionsList=new JList(extModel);	
	
	highlightModeCheckBox=new JCheckBox(ResourceManager.getMessage(ResourceManager.HIGHLIGHT_THIS_MODE));
	highlightModeCheckBox.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    FileMode mode=modeSet.getMode((String)fileModeList.getSelectedValue());
		    mode.setShouldHighlight(highlightModeCheckBox.isSelected());
		    SyntaxDocument sdoc=(SyntaxDocument)previewPane.getDocument();
		    sdoc.setFileMode(mode);
		    sdoc.retokenizeAll();
		    TextEditorCustomizer.this.previewPane.repaint();
		    TextEditorCustomizer.this.setDirty(true);
		}
	    });
	removeExtensionButton=new JButton(ResourceManager.getMessage(ResourceManager.REMOVE_EXTENSION));
	fileModeList.addListSelectionListener(new ListSelectionListener()
	    {
		public void valueChanged(ListSelectionEvent lassie)
		{
		    final FileMode mode=modeSet.getMode((String)fileModeList.getSelectedValue());
		    extModel.setFileMode(mode);
		    if (extModel.getSize()>0) 
		    {
			extensionsList.setSelectedIndex(0);
			removeExtensionButton.setEnabled(true);
		    }
		    else removeExtensionButton.setEnabled(false);
		    highlightModeCheckBox.setSelected(mode.getCanHighlight() && mode.getShouldHighlight());
		    highlightModeCheckBox.setEnabled(mode.getCanHighlight());
		    final SyntaxDocument sdoc=(SyntaxDocument)previewPane.getDocument();
		    sdoc.setFileMode(mode);
		    SwingUtilities.invokeLater(new Runnable() 
			{
			    public void run()
			    {
				previewPane.setText(getPreviewText(mode.getName()));
				sdoc.retokenizeAll();
				previewPane.repaint();
			    }
			});
		}
	    });
	fileModeList.setSelectedValue("plain", true);
	final JButton addExtensionButton=new JButton(ResourceManager.getMessage(ResourceManager.ADD_EXTENSION));
	addExtensionButton.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    FileMode mode=modeSet.getMode((String)fileModeList.getSelectedValue());
		    String message=ResourceManager.getMessage(ResourceManager.NEW_EXTENSION_DIALOG_PROMPT, 
							      new Object[] { mode.getName() });
		    String newExt=JOptionPane.showInputDialog(TextEditorCustomizer.this.getCustomizer(), 
							      message,
							      newExtensionDialogTitle,
							      JOptionPane.QUESTION_MESSAGE);
		    if (newExt!=null)
		    {
			mode.addExtension(newExt);
			extModel.setFileMode(mode);
			if (extModel.getSize()>0) 
			{
			    extensionsList.setSelectedIndex(0);
			    removeExtensionButton.setEnabled(true);
			}
			else removeExtensionButton.setEnabled(false);
			TextEditorCustomizer.this.setDirty(true);
		    }
		}
	    });

	removeExtensionButton.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    FileMode mode=modeSet.getMode((String)fileModeList.getSelectedValue());
		    Object o=extensionsList.getSelectedValue();
		    if (o!=null) 
		    {
			mode.removeExtension((String)o);
			extModel.setFileMode(mode);
			if (extModel.getSize()>0) 
			{
			    extensionsList.setSelectedIndex(0);
			    removeExtensionButton.setEnabled(true);
			}
			else removeExtensionButton.setEnabled(false);
			TextEditorCustomizer.this.setDirty(true);
		    }
		}
	    });

	//layout file modes sub-panel
	JPanel tmp=new JPanel(new GridBagLayout());
	GridBagConstraints gbc=initGBC();
	gbc.fill=GridBagConstraints.VERTICAL;
	JScrollPane scroll=new JScrollPane(fileModeList);
	scroll.setPreferredSize(new Dimension(48, 70));
	Box tmpBox=Box.createVerticalBox();
	tmpBox.add(fileModesLbl);
	tmpBox.add(scroll);
	tmp.add(tmpBox, gbc);
	gbc.gridx++;
	scroll=new JScrollPane(extensionsList);
	scroll.setPreferredSize(new Dimension(48, 70));
	tmpBox=Box.createVerticalBox();
	tmpBox.add(extensionsLbl);
	tmpBox.add(scroll);
	tmp.add(tmpBox, gbc);
	gbc.gridx++;
	tmpBox=Box.createVerticalBox();
	tmpBox.add(Box.createVerticalGlue());
	tmpBox.add(addExtensionButton);
	tmpBox.add(removeExtensionButton);
	tmp.add(tmpBox, gbc);
	gbc.gridx++;
	tmp.add(highlightModeCheckBox, gbc);

	tmp.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
						       fileModePropertiesStr,
						       TitledBorder.LEFT,
						       TitledBorder.TOP));
	return tmp;
    }

    private JPanel initStyles()
    {
	String stylePropertiesStr=ResourceManager.getMessage(ResourceManager.STYLE_PROPERTIES);
	boldCheckBox=new JCheckBox(ResourceManager.getMessage(ResourceManager.BOLD));
	boldCheckBox.setHorizontalTextPosition(SwingConstants.LEFT);
	italicCheckBox=new JCheckBox(ResourceManager.getMessage(ResourceManager.ITALIC));
	italicCheckBox.setHorizontalTextPosition(SwingConstants.LEFT);
	JLabel colorLbl=new JLabel(ResourceManager.getMessage(ResourceManager.COLOR));
	styleModel=new StyleComboBoxModel();
	styleDropdown=new JComboBox(styleModel);
	styleColorButton=new ColorChoiceButton();
	styleColorButton.addPropertyChangeListener(new PropertyChangeListener()
	    {
		public void propertyChange(PropertyChangeEvent cheese)
		{
		    if (cheese.getPropertyName().equals(ColorChoiceButton.COLOR_PROPERTY))
		    {
			Color styleColor=styleColorButton.getColor();
			SyntaxStyle ss=configSet.getStyle((String)styleDropdown.getSelectedItem());
			if (ss.getForegroundColor().equals(styleColor))
			    return;
			ss.setForegroundColor(styleColor);
			previewPane.repaint();
			TextEditorCustomizer.this.setDirty(true);
		    }
		}
	    });
	styleDropdown.addItemListener(new ItemListener()
	    {
		public void itemStateChanged(ItemEvent ie)
		{
		    if (ie.getStateChange()==ItemEvent.SELECTED)
		    {
			Object o=ie.getItem();
			if (Assertion.ASSERT) 
			    Assertion.assertF((o instanceof String), "selected item is a string");
			SyntaxStyle ss=configSet.getStyle((String) o);
			styleColorButton.setColor(ss.getForegroundColor());
			boldCheckBox.setSelected(ss.isBold());
			italicCheckBox.setSelected(ss.isItalic());
		    }
		}
	    });
	styleDropdown.setSelectedIndex(0);

	boldCheckBox.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    SyntaxStyle ss=configSet.getStyle((String)styleDropdown.getSelectedItem());
		    ss.setBold(boldCheckBox.isSelected());
		    previewPane.repaint();
		    TextEditorCustomizer.this.setDirty(true);
		}
	    });
	italicCheckBox.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent ae)
		{
		    SyntaxStyle ss=configSet.getStyle((String)styleDropdown.getSelectedItem());
		    ss.setItalic(italicCheckBox.isSelected());
		    previewPane.repaint();
		    TextEditorCustomizer.this.setDirty(true);
		}
	    });
	GridBagConstraints gbc=initGBC();
	JPanel tmp=new JPanel(new GridBagLayout());
	
	tmp.add(styleDropdown, gbc);
	gbc.gridx++;
	tmp.add(boldCheckBox, gbc);
	gbc.gridx++;
	tmp.add(italicCheckBox, gbc);
	gbc.gridx++;
	tmp.add(colorLbl, gbc);
	gbc.gridx++;
	tmp.add(styleColorButton, gbc);

	tmp.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
						       stylePropertiesStr,
						       TitledBorder.LEFT,
						       TitledBorder.TOP));
	return tmp;
    }

    private JScrollPane initPreview()
    {
	String previewStr=ResourceManager.getMessage(ResourceManager.PREVIEW);
	previewPane.setBackground(globalBackground); 
	previewPane.setForeground(globalForeground);
	previewPane.setCaretColor(globalCaretColor);
	previewPane.setFont(defaultFont);
	previewPane.setBorder(BorderFactory.createEtchedBorder());
	//margin isn't configurable, for now
	previewPane.setMargin(new Insets(8, 8, 8, 8));
	previewPane.setTabSize(tabSize);
	previewPane.setText(getPreviewText("plain"));
	JScrollPane scroller =new JScrollPane(previewPane);
	scroller.setPreferredSize(new Dimension(600, 280));
	scroller.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
							    previewStr,
							    TitledBorder.LEFT,
							    TitledBorder.TOP));
	return scroller;
    }

    private void initComponents()
    {
	customizer=new JPanel(new GridBagLayout());
	previewPane=new PreviewPane();
	previewPane.setEditorKit(new SyntaxEditorKit(configSet));
	GridBagConstraints gbc=initGBC();
	gbc.fill=GridBagConstraints.HORIZONTAL;
	customizer.add(initGlobal(), gbc);
	gbc.gridy=1;
	customizer.add(initFileModes(), gbc);
	gbc.gridy=2;
	customizer.add(initStyles(), gbc);
	gbc.gridy=3;
	customizer.add(initPreview(), gbc);

    } 

    public JComponent getCustomizer()
    {
	return customizer;
    }      

    private class FileModeListModel extends AbstractListModel
    {
	private ArrayList modeList;
	private FileMode.ModeMapContainer modeSet;

	FileModeListModel(FileMode.ModeMapContainer modeSet)
	{
	    super();
	    this.modeSet=modeSet;
	    modeList=new ArrayList();
	    refresh();
	}

	public void refresh()
	{
	    modeList.clear();
	    for (Iterator it=modeSet.modeNames();it.hasNext();)
		modeList.add(it.next());
	    Collections.sort(modeList);
	}

	public Object getElementAt(int index)
	{
	    return modeList.get(index);
	}

	public int getSize()
	{
	    return modeList.size();
	}
    }

    private class FileModeExtensionListModel extends AbstractListModel
    {
	private String[] extensions;

	FileModeExtensionListModel()
	{
	    super();
	}

	void setFileMode(FileMode mode)
	{
	    if (mode!=null) 
	    {
		this.extensions=mode.getExtensions();
		fireContentsChanged(this, 0, extensions.length);
	    }
	    else 
	    {
		this.extensions=null;
		fireContentsChanged(this, 0, 0);
	    }
	   
	}

	public int getSize()
	{
	    if (extensions==null)
		return 0;
	    else return extensions.length;
	}

	public Object getElementAt(int index)
	{
	    return extensions[index];
	}
    }

    private class StyleComboBoxModel extends DefaultComboBoxModel
    {
	ArrayList styleList;

	StyleComboBoxModel()
	{
	    super();
	    styleList=new ArrayList();
	    refresh();
	}

	public void refresh()
	{
	    styleList.clear();
	    for (Iterator it=configSet.styles();it.hasNext();)
	    {
		styleList.add(((SyntaxStyle)it.next()).getName());
	    }
	    Collections.sort(styleList);
	}

	public int getSize()
	{
	    return styleList.size();
	}

	public Object getElementAt(int index)
	{
	    return styleList.get(index);
	}
    }

    public static class PreviewPane extends TextEditorPane 
	implements ConfigurableEditor
    {
	int tabSize=DEFAULT_TAB_SIZE;

	PreviewPane()
	{
	    super();
	}

	public int getTabSize()
	{
	    return tabSize;
	}

	public void setTabSize(int tabSize)
	{
	    this.tabSize=tabSize;
	    getDocument().putProperty("tabSize", new Integer(tabSize));
	}
    }
}

/* $Log: TextEditorCustomizer.java,v $
/* Revision 1.7  2001/02/20 18:04:26  smulloni
/* fixed bug in TextEditorCustomizer that prevented file extensions from being
/* removed from file modes.
/*
/* Revision 1.6  2001/02/16 19:57:36  smulloni
/* fixed problem with reverting customizations to the syntax style set.
/*
/* Revision 1.5  2001/02/16 18:15:10  smulloni
/* many fixes to TextEditorCustomizer.  FileMode and SyntaxStyle now have a
/* configData property (they will probably be made into sibling classes).
/*
/* Revision 1.4  2001/02/08 23:48:50  smulloni
/* now load default styles from a config file.  Styles have been renamed to be
/* more general (comment3 instead of docComment) so future modes won't need to
/* invent more styles.
/*
/* Revision 1.3  2001/02/06 22:13:40  smulloni
/* first more-or-less working version of syntax highlighting, with customization.
/*
/* Revision 1.2  2001/02/06 00:11:18  smulloni
/* struggle, perhaps futile, with the TextEditorCustomizer and other implicated
/* classes
/*
   /* Revision 1.1  2001/02/02 23:30:33  smulloni
   /* adding customization features to the text editor.
   /* */
