/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/


package org.skunk.dav.client.gui.action;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.util.StringTokenizer;
import javax.swing.AbstractAction;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import org.skunk.dav.client.gui.ResourceManager;
import org.skunk.swing.SplashScreen;

/**
 * an Action for the "About" menu item.  Shows a splash screen.
 *
 * @author Jacob Smullyan
 */

public class AboutAction extends AbstractAction
{
    private static final String DEFAULT_ICON_KEY=ResourceManager.SPLASH_SCREEN_ICON;
    private SplashScreen splash;
    private String iconKey;
    private String splashScreenText;
    private SplashListener listener;
    
    public AboutAction()
    {
	this(DEFAULT_ICON_KEY);
    }

    public AboutAction(String iconKey)
    {
	this.iconKey=iconKey;
	initSplash(iconKey);
    }

    private void initSplash(String iconKey)
    {
	listener=new SplashListener();
	String releaseVersion=ResourceManager.getMessage(ResourceManager.RELEASE_VERSION);
	splashScreenText=ResourceManager.getMessage(ResourceManager.SPLASH_SCREEN_TEXT, 
						    new Object[] {releaseVersion});
	splash=new SplashScreen(getImageIcon(iconKey));
	splash.addMouseListener(listener);
	splash.addKeyListener(listener);
    }

    /**
     * obtains the image from the ResourceManager with the given key, and draws text on it.
     * the drawing is entirely hardcoded, and expects the splash screen text to be in a pipe-delimited
     * format with exactly five fields. 
     * @param iconKey the key for the ResourceManager
     */
    protected ImageIcon getImageIcon(String iconKey)
    {
	ImageIcon ii=ResourceManager.getImageIcon(iconKey);
	if (ii==null) ii=ResourceManager.getImageIcon(DEFAULT_ICON_KEY);
	if (ii==null) return null;
	//do custom painting on image.
	Image im=ii.getImage();
	//in order to call a components createImage() method, 
	//I need one that has been fully constructed,
	//or I may get a null pointer.
	JLabel label=new JLabel(); 
	BufferedImage bim=(BufferedImage)label.createImage(im.getWidth(label), im.getHeight(label));
	if (bim==null) // I don't know why this happens, but it does.
	    bim=new BufferedImage(im.getWidth(label), im.getHeight(label), BufferedImage.TYPE_INT_RGB);
	Graphics2D g2d=bim.createGraphics();
	g2d.drawImage(im, 0, 0, label);
	Font bigFont=new Font("Dialog", Font.BOLD, 18);
	Font mediumFont=new Font("Dialog", Font.BOLD, 12);
	Font littleFont=new Font("SansSerif", Font.PLAIN, 9);

	StringTokenizer st=new StringTokenizer(getSplashScreenText(), "|");
	String title=st.nextToken();
	String subtitle=st.nextToken();
	String copyright=st.nextToken();
	String acknowledgements1=st.nextToken();
	String acknowledgements2=st.nextToken();
	g2d.setFont(bigFont);
	g2d.setColor(Color.red);
	g2d.drawString(title, 10, 30);
	g2d.setFont(mediumFont);
	g2d.drawString(subtitle, 10, 50);
	g2d.setFont(littleFont);
	g2d.setColor(Color.white);
	g2d.drawString(copyright, 10, 160);
	g2d.drawString(acknowledgements1, 10, 175);
	g2d.drawString(acknowledgements2, 10, 190);
	return new ImageIcon(bim);
    }

    public void actionPerformed(ActionEvent ae)
    {
	splash.setVisible(true);
    }

    public String getSplashScreenText()
    {
	return splashScreenText;
    }

    public void setSplashScreenText(String splashScreenText)
    {
	this.splashScreenText=splashScreenText;
	splash.removeMouseListener(listener);
	splash.removeKeyListener(listener);
	this.splash=new SplashScreen(getImageIcon(this.iconKey));
	splash.addMouseListener(listener);
	splash.addKeyListener(listener);
    }

    private class SplashListener extends MouseAdapter implements KeyListener
    {
	public void mouseClicked(MouseEvent mao)
	{
	    splashAbout();
	}
	public void keyPressed(KeyEvent keanu)
	{
	    splashAbout();
	}
	public void keyReleased(KeyEvent keanu) {}
	public void keyTyped(KeyEvent keanu) {}

	private void splashAbout()
	{
	    if (splash.isVisible())
		splash.setVisible(false);
	}
    }
}

/* $Log: AboutAction.java,v $
/* Revision 1.4  2001/07/22 07:04:24  smulloni
/* changed the handling of release number.
/*
/* Revision 1.3  2000/12/21 18:53:13  smulloni
/* cosmetic improvements.
/*
/* Revision 1.2  2000/12/19 22:06:15  smulloni
/* adding documentation.
/* */
