/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui.action;

import java.awt.Component;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.util.Iterator;
import javax.swing.AbstractAction;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.gui.Buffer;
import org.skunk.dav.client.gui.DAVTreeNode;
import org.skunk.dav.client.gui.Explorer;
import org.skunk.dav.client.gui.ExplorerApp;
import org.skunk.dav.client.gui.ResourceManager;
import org.skunk.dav.client.gui.ServerData;
import org.skunk.dav.client.gui.StateMonitor;
import org.skunk.dav.client.gui.StateProperties;
import org.skunk.dav.client.gui.View;
import org.skunk.dav.client.gui.editor.DAVEditor;
import org.skunk.trace.Debug;

/**
 * an Action that closes a buffer.
 * @author Jacob Smullyan
 */
public class CloseBufferAction extends AbstractAction
{
    private Buffer buffer;
    private boolean noprompt=false;

    /**
     * @param buffer the buffer to be closed
     */
    public CloseBufferAction(Buffer buffer)
    {
	this.buffer=buffer;
    }

    /**
     * @param buffer the buffer to be closed
     * @param noprompt whether or not to suppress dialog messages to the client
     */
    public CloseBufferAction(Buffer buffer, boolean noprompt)
    {
	this(buffer);
	this.noprompt=noprompt;
    }

    /**
     * constructor that closes the current buffer
     */
    public CloseBufferAction()
    {
	this(null);
    }

    public void actionPerformed(ActionEvent ae)
    {
	Buffer localBuffer=(buffer==null) 
	    ? ExplorerApp.getAppContext().getCurrentView().getFocussedBuffer()
	    : buffer;
	Debug.trace(this, Debug.DP3, "buffer to be closed is "+localBuffer);
	if (localBuffer!=null)
	{
	    if (localBuffer instanceof DAVEditor)
	    {
		DAVEditor editor=(DAVEditor)localBuffer;
		if (editor.isDirty())
		{
		    String title=ResourceManager.getMessage(ResourceManager.CLOSE_BUFFER_KEY);
		    String message=ResourceManager.getMessage(ResourceManager.CLOSE_DIRTY_BUFFER_WARNING, 
							      new Object[] {localBuffer.getName()});
		    if (!noprompt)
		    {
			int option=JOptionPane.showConfirmDialog(localBuffer.getComponent(),
								 message, 
								 title, 
								 JOptionPane.OK_CANCEL_OPTION,
								 JOptionPane.WARNING_MESSAGE);
			if (option==JOptionPane.CANCEL_OPTION)
			    return;
		    }
		}
		DAVFile file=editor.getDAVFile();
		Explorer ex=ExplorerApp.getExplorerForBuffer(localBuffer);
		DAVTreeNode node=ex.getNodeForFile(file);
		Debug.trace(this, Debug.DP5, "explorer is {0}, node is {1}, file is {2}",
			    new Object[] {ex, node, file});
		if (ex!=null)
		{
		    if (file.isLocked())
		    {
			if (file.isExclusiveLocked())
			{
			    String lockOwner=file.getExclusiveLockOwner();
			    ServerData sd=ex.getConnection(node);
			    String localLockOwner=sd.getOwner();
			    if (localLockOwner==null)
				localLockOwner=sd.getUsername();
			    if (localLockOwner.equals(lockOwner))
				ex.unlock(node, file);
			}
		    }
		}
	    }
	    View v=ExplorerApp.getViewForBuffer(localBuffer);
	    if (v!=null)
		v.undock(localBuffer);
	    StateMonitor.unregisterProperty(localBuffer, "DAVFile", StateProperties.RELOAD);
	}
    }
}

/* $Log: CloseBufferAction.java,v $
/* Revision 1.7  2000/12/19 22:06:15  smulloni
/* adding documentation.
/*
/* Revision 1.6  2000/12/19 19:22:22  smulloni
/* some tweaks: ExitAction now cleans up the app's buffers.  The app now does
/* not resize every time a buffer is added or removed.
/*
/* Revision 1.5  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.4  2000/11/23 04:37:34  smullyan
/* editor and explorer now synch their files using the StateMonitor, which has
/* been improved significantly
/*
/* Revision 1.3  2000/11/22 00:11:28  smullyan
/* editor now locks and unlocks, more or less appropriately.
/*
/* Revision 1.2  2000/11/20 23:30:22  smullyan
/* more editor integration work.
/*
/* Revision 1.1  2000/11/16 20:45:19  smullyan
/* the start of editor integration.
/* */
