/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui.action;

import java.awt.event.ActionEvent;
import java.util.Iterator;
import javax.swing.AbstractAction;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.gui.Buffer;
import org.skunk.dav.client.gui.DAVTreeNode;
import org.skunk.dav.client.gui.Explorer;
import org.skunk.dav.client.gui.ExplorerApp;
import org.skunk.dav.client.gui.ServerData;
import org.skunk.dav.client.gui.View;
import org.skunk.dav.client.gui.editor.CannotLoadException;
import org.skunk.dav.client.gui.editor.DAVEditor;
import org.skunk.dav.client.gui.editor.DAVEditorFactory;
import org.skunk.trace.Debug;

public class OpenAction extends AbstractAction
{
    private Explorer ex;
    private DAVTreeNode node;
    private DAVFile file;

    public OpenAction(Explorer ex)
    {
	this(ex, null, null);
    }

    public OpenAction(Explorer ex, DAVFile file)
    {
	this(ex, null, file);
    }

    public OpenAction(Explorer ex, DAVTreeNode node, DAVFile file)
    {
	this.ex=ex;
	this.node=node;
	this.file=file;
    }

    public void actionPerformed(ActionEvent ae)
    {
	final DAVTreeNode localNode=(node==null) ? ex.getSelectedNode() : node;
	
	if (localNode==null)
	{
	    Debug.trace(this, Debug.DP3, "null node in OpenAction");
	    return;
	}
	final DAVFile localFile=(file==null) ? ex.getSelectedTableFile() : file;
	if (localFile==null)
	{
	    Debug.trace(this, Debug.DP3, "null file in OpenAction");
	    return;
	}
	if (localFile.isCollection())
	{
	    ex.displayCollection(localNode, localFile.getFileName());
	}
	else
	{
	    //determine if this file is already being edited
	    View v=ExplorerApp.getViewForBuffer(ex);
	    for (Iterator it=v.getDockedBuffers();it.hasNext();)
	    {
		Object o=it.next();
		if (o instanceof DAVEditor)
		{
		    DAVEditor ed=(DAVEditor)o;
		    if (ed.getResourceName().equals(localFile.getFullName()))
		    {
			v.focus(ed);
			return;
		    }
		}
	    }
	    Debug.trace(this, Debug.DP3, 
			"invoking associated editor for file " 
			+ localFile.getName());
	    final Explorer finalEx=ex;
	    if (localFile.isLocked())
	    {
		if (localFile.isExclusiveLocked())
		{
		    String lockOwner=localFile.getExclusiveLockOwner();
		    ServerData sd=ex.getConnection(localNode);
		    String localLockOwner=sd.getOwner();
		    if (localLockOwner==null)
			localLockOwner=sd.getUsername();
		    openFileInEditor(localFile, 
				     (lockOwner!=null && lockOwner.equals(localLockOwner)));
		}
		//currently no support for shared locks
		else openFileInEditor(localFile, false);
	    }
	    else 
	    {
		ex.lock(localNode, localFile, new Runnable() 
		    {
			public void run()
			{
			    openFileInEditor(finalEx.getFile(localNode, localFile.getFileName()), true);
			}
		    });
	    }
	}
    }

    private void openFileInEditor(DAVFile file, boolean writeable)
    {
	DAVEditor editor=DAVEditorFactory.getEditor(file);
	editor.addEditListener(ex);
	byte[] body=ex.get(file);
	editor.setResourceBody(body);
	try
	{
	    editor.load();
	}
	catch (CannotLoadException canal)
	{
	    editor.removeEditListener(ex);
	    editor=null;
	    editor=DAVEditorFactory.getDefaultEditor(file);
	    editor.addEditListener(ex);
	    editor.setResourceBody(body);
	    try
	    {
		editor.load();
	    }
	    catch (CannotLoadException canal2)
	    {
		//should not happen
		Debug.trace(this, Debug.DP2, canal2);
	    }
	}
	editor.setWriteable(writeable);
	View v=ExplorerApp.getViewForBuffer(ex);
	if (v!=null)
	    v.dock(editor);
    }
}

/* $Log: OpenAction.java,v $
/* Revision 1.15  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.14  2000/12/08 05:50:30  smulloni
/* fixed MessageCatalogEditor.  The spi features are now a special build option,
/* and editors are loaded through reflection.
/*
/* Revision 1.13  2000/12/04 23:51:16  smulloni
/* added ImageViewer; fixed word in SimpleTextEditor
/*
/* Revision 1.12  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.11  2000/11/23 04:37:34  smullyan
/* editor and explorer now synch their files using the StateMonitor, which has
/* been improved significantly
/*
/* Revision 1.10  2000/11/22 00:11:29  smullyan
/* editor now locks and unlocks, more or less appropriately.
/*
/* Revision 1.9  2000/11/20 23:30:22  smullyan
/* more editor integration work.
/*
/* Revision 1.8  2000/11/18 04:36:05  smullyan
/* work on StateMonitor and related functionality.
/*
/* Revision 1.7  2000/11/16 23:16:47  smullyan
/* more preliminary work on editor integration
/*
/* Revision 1.6  2000/11/16 20:45:20  smullyan
/* the start of editor integration.
/*
/* Revision 1.5  2000/11/15 20:17:04  smullyan
/* added a Buffer interface, which is a wrapper around a displayable component.
/*
/* Revision 1.4  2000/11/15 19:45:37  smullyan
/* beginning of revamp of application to include multiple buffers.
/*
/* Revision 1.3  2000/11/09 23:35:04  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
