/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui.editor;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import javax.swing.JComponent;
import javax.swing.SwingUtilities;
import org.skunk.assertF.Assertion; 
import org.skunk.dav.client.DAVConnection;
import org.skunk.dav.client.DAVConnectionPool;
import org.skunk.dav.client.DAVConstants;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.gui.DAVTreeModel; 
import org.skunk.dav.client.gui.DAVTreeNode;
import org.skunk.dav.client.gui.Explorer;
import org.skunk.dav.client.gui.ExplorerApp;
import org.skunk.dav.client.gui.ExplorerView;
import org.skunk.dav.client.gui.StateMonitor;
import org.skunk.dav.client.gui.StateProperties;
import org.skunk.dav.client.gui.View;
import org.skunk.trace.Debug;

public abstract class AbstractDAVEditor implements DAVEditor
{
    private String name;
    private byte[] resourceBody;
    private boolean dirty;
    private JComponent editingComponent;
    private String resourceName;
    private boolean writeable;
    private List editListeners=Collections.synchronizedList(new ArrayList());
    private DAVFile file;

    public AbstractDAVEditor(DAVFile file)
    {
	this.file=file;
	if (file!=null) 
	{
	    setName(file.getFileName());
	    setResourceName(file.getFullName());
	}
	setDirty(false);
	//this synchronizes the editor's DAVFile reference to the file in the explorer's TreeModel (our VFS)
	StateMonitor.registerProperty(this, 
				      "DAVFile", 
				      StateProperties.RELOAD,
				      getMapper(), 
				      getDomainMatcher());				      
    }

    private StateMonitor.Mapper getMapper()
    {
	return new StateMonitor.Mapper()
	    {
		public Object getTargetValue(Object applicationValue)
		{
		    Debug.trace(this, Debug.DP3, "in getTargetValue() with applicationValue of "+applicationValue);
		    Assertion.assertF((applicationValue instanceof DAVTreeModel), 
				     "we have a DAVTreeModel");
		    DAVTreeModel model=(DAVTreeModel) applicationValue;
		    //obtain appropriate DAVFile from treeModel
		    String fullPath=AbstractDAVEditor.this.getResourceName();
		    int lastSlash=fullPath.lastIndexOf("/");
		    String parentPath=fullPath.substring(0, lastSlash);
		    String filename=fullPath.substring(lastSlash+1);
		    DAVTreeNode node=model.getNodeMatchingPath(parentPath);
		    Debug.trace(this, Debug.DP3, "parentPath: {0}; filename: {1}; node: {2}",
				new Object[] {parentPath, filename, node});
		    if (node!=null)
		    {
			DAVFile parentFile=node.getDAVFile();
			return parentFile.getChildNamed(AbstractDAVEditor.this.getDAVFile().getName());
		    }
		    return null;
		}

		public boolean canMapValue(Object applicationValue)
		{
		    return (applicationValue instanceof DAVTreeModel);
		}
	    };
    }

    private StateMonitor.DomainMatcher getDomainMatcher()
    {
	return new StateMonitor.DomainMatcher()
	    {
		public boolean domainMatches(Object domainKey)
		{
		    Debug.trace(this, Debug.DP3, "testing domainKey "+domainKey);
		    if (domainKey==null) return false;
		    String pathString=domainKey.toString();
		    String resourceString=AbstractDAVEditor.this.resourceName;
		    Debug.trace(this, Debug.DP3, "resourceString is "+resourceString);
		    if (resourceString.equals(pathString))
		    {
			Debug.trace(this, Debug.DP3, "exact match");
			return true;
		    }
		    else if (resourceString.startsWith(pathString))
		    {
			Debug.trace(this, Debug.DP3, "resourceString {0} may match {1}",
				    new Object[] {resourceString, pathString});
			//compare the last path element in the key with the corresponding 
			//element in the resource name
			int pathLength=pathString.length();
			pathLength=(pathString.endsWith("/"))
			    ? pathLength-1
			    : pathLength;
			//this prevents pathString /foo/bar/ from matching with resourceString /foo/bar2/giblet
			boolean result= (resourceString.charAt(pathLength)=='/');
			Debug.trace(this, Debug.DP3, "did the domainKey match? "+result);
			return result;
		    }
		    else return false;
		}
	    };
    }

    public void docking()
    {
    }

    public void undocking()
    {
	StateMonitor.unregisterProperty(this, 
				      "DAVFile", 
					StateProperties.RELOAD);
    }

    public void setDAVFile(DAVFile file)
    {
	if (file==null)
	{
	    handleNulledFile();
	    return;
	}
	Debug.trace(this, Debug.DP3, "DAVFile property now being set to "+file);
	this.file=file;
    }

    protected void handleNulledFile()
    {
	Debug.trace(this, Debug.DP3, "in handleNulledFile()");
	//should do something here, put up a dialog, perhaps -- TO BE DONE
    }

    public DAVFile getDAVFile()
    {
	return this.file;
    }

    public void setName(String name)
    {
	this.name=name;
    }

    public String getName()
    {
	return name;
    }

    public byte[] getResourceBody() 
    { 
	return this.resourceBody; 
    }

    public void setResourceBody(byte[] resourceBody)
    {
	this.resourceBody=resourceBody;
    }

    public String getResourceName()
    {
	return this.resourceName;
    }

    public void setResourceName(String resourceName)
    {
	this.resourceName=resourceName;
    }

    public boolean isDirty()
    {
	return this.dirty;
    }

    public void setDirty(boolean dirty)
    {
	this.dirty=dirty;
	View v=ExplorerApp.getViewForBuffer(this);
	if (v instanceof ExplorerView)
	{
	    ((ExplorerView) v).bufferDirtinessChanged(this);
	}
    }

    public boolean isWriteable()
    {
	return this.writeable;
    }

    public void setWriteable(boolean writeable)
    {
	this.writeable=writeable;
    }

    public abstract JComponent getComponent();

    public void save()
    {
	Debug.trace(this, Debug.DP3, "in save()");
	EditEvent event=new EditEvent(this, EditEvent.SAVE);
	fireEditEvent(event);
    }

    public final void saveAs()
    {
	Debug.trace(this, Debug.DP3, "in saveAs()");
	EditEvent event=new EditEvent(this, EditEvent.SAVE_AS);
	fireEditEvent(event);
    }

    public final void saveAs(String filename)
    {
	setResourceName(filename);
	setName(filename);
	save();
    }

    public abstract void load() throws CannotLoadException;


    public void addEditListener(EditListener ed)
    {
	editListeners.add(ed);
    }

    public void removeEditListener(EditListener ed)
    {
	editListeners.remove(ed);
    }

    private void fireEditEvent(EditEvent event)
    {
	for (Iterator it=editListeners.iterator();it.hasNext();)
	{
	    ((EditListener)it.next()).editPerformed(event);
	}
    }

    
}

/* $Log: AbstractDAVEditor.java,v $
/* Revision 1.16  2001/02/16 18:15:10  smulloni
/* many fixes to TextEditorCustomizer.  FileMode and SyntaxStyle now have a
/* configData property (they will probably be made into sibling classes).
/*
/* Revision 1.15  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.14  2000/12/08 05:50:30  smulloni
/* fixed MessageCatalogEditor.  The spi features are now a special build option,
/* and editors are loaded through reflection.
/*
/* Revision 1.13  2000/12/07 00:01:34  smulloni
/* adding an additional editor, still hard-coded into the DAVEditorFactory.
/*
/* Revision 1.12  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.11  2000/12/01 16:25:54  smullyan
/* improvements to look and feel; fixed NPE in DAVFile; new actions for text
/* editor
/*
/* Revision 1.10  2000/11/28 19:36:19  smullyan
/* some bug fixes; notes.
/*
/* Revision 1.9  2000/11/28 00:01:40  smullyan
/* added a status bar/minibuffer, with a location field showing the current line and
/* column number (for the SimpleTextEditor and kin only).
/*
/* Revision 1.8  2000/11/23 04:37:35  smullyan
/* editor and explorer now synch their files using the StateMonitor, which has
/* been improved significantly
/*
/* Revision 1.7  2000/11/20 23:30:23  smullyan
/* more editor integration work.
/*
/* Revision 1.6  2000/11/18 04:36:07  smullyan
/* work on StateMonitor and related functionality.
/*
/* Revision 1.5  2000/11/17 20:25:07  smullyan
/* new SaveAction; a StateMonitor being added to handle application state.
/*
/* Revision 1.4  2000/11/16 23:16:48  smullyan
/* more preliminary work on editor integration
/*
/* Revision 1.3  2000/11/15 20:17:05  smullyan
/* added a Buffer interface, which is a wrapper around a displayable component.
/*
/* Revision 1.2  2000/11/15 19:45:37  smullyan
/* beginning of revamp of application to include multiple buffers.
/*
/* Revision 1.1  2000/11/14 23:14:39  smullyan
/* first cut at editing framework
/* */
