/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui.editor;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Properties;
import java.util.StringTokenizer;
import org.skunk.config.Configurator;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.gui.ExplorerApp;
import org.skunk.swing.text.syntax.FileMode;
import org.skunk.swing.text.syntax.SyntaxStyle;
import org.skunk.trace.Debug;

public abstract class DAVEditorFactory
{
    public static Properties editorProps;
    public static final String EDITOR_PROPERTIES="editor.properties";
    public static final String EXTENSION_PREFIX="extension-";
    public static final String EXTENSION_SEPARATOR="/";
    private static HashMap extensionHash;
    private static Configurator configurator;
    
    static
    {
	editorProps=new Properties();
	extensionHash=new HashMap();
	//this will only be adequate while all the editor associations are hard-coded.
	try
	{
	    editorProps.load(DAVEditorFactory.class.getResourceAsStream(EDITOR_PROPERTIES));
	    if (Debug.DEBUG) Debug.trace(DAVEditorFactory.class, Debug.DP3, editorProps);
	    makeExtensionHash();
	}
	catch (IOException oyVeh)
	{
	    if (Debug.DEBUG) Debug.trace(DAVEditorFactory.class, Debug.DP1, oyVeh);
	}
	//configure the FileMode mode maps
	configurator=ExplorerApp.getAppContext().getConfigurator();
	configurator.configure(FileMode.getModeMapContainer());
	configurator.configure(SyntaxStyle.getDefaultStyleSet());
    }
    
    private static void makeExtensionHash()
    {
	for (Enumeration enum=editorProps.keys();enum.hasMoreElements();)
	{
	    String key=enum.nextElement().toString();
	    if (Debug.DEBUG) Debug.trace(DAVEditorFactory.class, Debug.DP3, "processing key: "+key);
	    if (key.startsWith(EXTENSION_PREFIX))
	    {
		StringTokenizer st=new StringTokenizer(key.substring(EXTENSION_PREFIX.length()), 
						       EXTENSION_SEPARATOR);
		String classStr=editorProps.getProperty(key);
		while (st.hasMoreTokens())
		{
		    extensionHash.put(st.nextToken().toLowerCase().trim(), classStr);
		}
	    }
	}
	if (Debug.DEBUG) Debug.trace(DAVEditorFactory.class, Debug.DP3, 
		    "extensionHash initialized: {0}", extensionHash);
    }

    public static final DAVEditor getEditor(DAVFile file)
    {
	String filename=file.getFileName();
	int lastDot=filename.lastIndexOf(".")+1;
	if (lastDot>0 && (lastDot)<filename.length())
	{
	    String extension=filename.substring(lastDot).toLowerCase().trim();
	    if (extensionHash.containsKey(extension))
	    {
		String editorName=extensionHash.get(extension).toString();
		//try to create it through reflection, using the constructor that takes a DAVFile
		try
		{
		    Class editorClass=Class.forName(editorName);
		    Constructor editorConstructor=editorClass.getConstructor(new Class[] {DAVFile.class});
		    Object editorObject=editorConstructor.newInstance(new Object[] {file});
		    if (editorObject instanceof DAVEditor) 
		    {
			configurator.configure(editorObject);
			return (DAVEditor) editorObject;
		    }
		}
		catch (Exception e)
		{
		    if (Debug.DEBUG) Debug.trace(DAVEditorFactory.class, Debug.DP3, e);
		}
	    }
	}
	return getDefaultEditor(file);
    }

    public static final DAVEditor getDefaultEditor(DAVFile file)
    {
	DAVEditor ed=new SimpleTextEditor(file);
	configurator.configure(ed);
	return ed;
    }
}
