/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui.editor;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Insets;
import java.awt.Rectangle;
import java.util.ArrayList;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JScrollPane;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Caret;
import javax.swing.text.Document;
import javax.swing.text.Element;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.gui.AppContext;
import org.skunk.dav.client.gui.Buffer;
import org.skunk.dav.client.gui.BusinessTracker;
import org.skunk.dav.client.gui.Explorer;
import org.skunk.dav.client.gui.ExplorerApp;
import org.skunk.dav.client.gui.StateMonitor;
import org.skunk.dav.client.gui.StateProperties;
import org.skunk.dav.client.gui.TextPoint;
import org.skunk.dav.client.gui.View;
import org.skunk.swing.text.TextEditorPane;
import org.skunk.swing.text.syntax.FileMode;
import org.skunk.swing.text.syntax.SyntaxDocument;
import org.skunk.swing.text.syntax.SyntaxEditorKit;
import org.skunk.swing.text.syntax.SyntaxStyle;
import org.skunk.trace.Debug;

/**
 * text editor supporting search, navigation, and syntax highlighting. 
 */
public class SimpleTextEditor extends AbstractDAVEditor implements ConfigurableEditor
{
    private TextEditorPane editingPane;
    private JScrollPane edcomp;
    private DAVEditorUndoManager undoManager;
    private View currentView;
    private int tabSize;

    public SimpleTextEditor(DAVFile file)
    {
	super(file);
	editingPane=new TextEditorPane();
	editingPane.setEditorKit(new SyntaxEditorKit());
	/* I am commenting this out, because the corresponding code in the Flexicizer introduces an
	   undesirable dependency into the build */
// 	if (Debug.DEBUG && Debug.isDebug(org.skunk.swing.text.syntax.SyntaxDocument.class, Debug.DP8))
// 	{
// 	    getSyntaxDocument().putProperty("editor", this);
// 	}

	//defaults of configurable properties
	editingPane.setFont(SyntaxStyle.getDefaultFont());
	setMargin(new Insets(15, 15, 15, 30)); 
	setBackground(SyntaxStyle.DEFAULT_BACKGROUND);
	setForeground(SyntaxStyle.DEFAULT_FOREGROUND);
	setCaretColor(DEFAULT_CARET_COLOR);
	setTokenizing(false);
	setTabSize(DEFAULT_TAB_SIZE);

	KeyMapper.loadKeymap(editingPane);
	edcomp=new JScrollPane(editingPane);
	undoManager=new DAVEditorUndoManager();
	initListeners();

    }

    public Insets getMargin()
    {
	return editingPane.getMargin();
    }

    public void setMargin(Insets margin)
    {
	editingPane.setMargin(margin);
    }

    public Color getBackground()
    {
	return editingPane.getBackground();
    }

    public void setBackground(Color background)
    {
	if (Debug.DEBUG) Debug.trace(this, Debug.DP4, "in setBackground({0}", background);
	editingPane.setBackground(background);
	editingPane.repaint();
    }

    public Color getForeground()
    {
	return editingPane.getForeground();
    }

    public void setForeground(Color foreground)
    {
	editingPane.setForeground(foreground);
    }

    public Color getCaretColor()
    {
	return editingPane.getCaretColor();
    }

    public void setCaretColor(Color caretColor)
    {
	editingPane.setCaretColor(caretColor);
    }

    public void setFont(Font defaultFont)
    {
	editingPane.setFont(defaultFont);
    }

    public int getTabSize()
    {
	return this.tabSize;
    }

    public void setTabSize(int tabSize)
    {
	this.tabSize=tabSize;
	//undocumented, but this is how documents store the tab size.
	editingPane.getDocument().putProperty("tabSize", new Integer(tabSize));
    }

    public boolean isTokenizing()
    {
	return getSyntaxDocument().isTokenizing();
     }

    public void setTokenizing(boolean tokenize)
    {
	getSyntaxDocument().setTokenizing(tokenize);
    }

    public FileMode getFileMode()
    {
	return getSyntaxDocument().getFileMode();
    }

    public void setFileMode(FileMode mode)
    {
	getSyntaxDocument().setFileMode(mode);
	if (mode==null ||  !mode.getCanHighlight())
	    setTokenizing(false);
	else setTokenizing(mode.getShouldHighlight());
    }


    public boolean isWordWrap()
    {
	return this.editingPane.getWordWrap();
    }

    public void setWordWrap(boolean wordWrap)
    {
	this.editingPane.setWordWrap(wordWrap);
	this.editingPane.revalidate();
    }	

    private void initListeners()
    {
	editingPane.getDocument().addDocumentListener(new DocumentListener()
	    {
		public void changedUpdate(DocumentEvent dee)
		{
		    //do nothing
		}

		public void insertUpdate(DocumentEvent dee)
		{
		    setDirty(true);
		}

		public void removeUpdate(DocumentEvent dee)
		{
		    setDirty(true);
		}
	    });
	editingPane.addCaretListener(new CaretListener() 
	    {
		public void caretUpdate(CaretEvent caries)
		{
		    publishCaretUpdate(caries.getDot());
		}
	    });

	editingPane.getDocument().addUndoableEditListener(undoManager);

	StateMonitor.registerProperty(this, "editorInFocus", StateProperties.EDITOR_IN_FOCUS, null, null);
    }

    public void setDirty(boolean dirty)
    {
	super.setDirty(dirty);
	if (!dirty && undoManager!=null && undoManager.canUndo())
	    undoManager.discardAllEdits();
    }

    public DAVEditorUndoManager getUndoManager()
    {
	return undoManager;
    }

    private void publishCaretUpdate(int dot)
    {
	TextPoint tp=getLineAndColumnForOffset(dot);
	Object o= (tp==null) ? (Object)"anything" : tp;
	StateMonitor.setProperty(StateProperties.CARET_POSITION, 
				 o, 
				 this.currentView);
    }

    public void setEditorInFocus(boolean focussed)
    {
	if (focussed)
	{
	    View v=ExplorerApp.getViewForBuffer(this);
	    if (v==null) 
	    {
		if (Debug.DEBUG) Debug.trace(this, Debug.DP3, "view for this buffer is null");
		return;
	    }
	    this.currentView=v;
	    Buffer b=v.getFocussedBuffer();
	    if (b!=null && this.equals(b))
	    {
		publishCaretUpdate(editingPane.getCaret().getDot());
	    }
	    //if another text editor has focus, do nothing, it will take care of it
	    else if (!(b instanceof SimpleTextEditor)) 
		publishCaretUpdate(-1);
	}
	// multiple text editors may do this redundantly
	else publishCaretUpdate(-1);
    }

    public void copy()
    {
	editingPane.copy();
    }

    public void cut()
    {
	editingPane.cut();
    }

    public void paste()
    {
	editingPane.paste();
    }

    public void replaceSelection(String content)
    {
	getUndoManager().beginUpdate();
	editingPane.replaceSelection(content);
	getUndoManager().endUpdate();
    }

    private int getLineCount()
    {
	Document doc=editingPane.getDocument();
	if (doc==null) return 0;
	Element mainElem=doc.getDefaultRootElement();
	if (mainElem==null) return 0;
	return mainElem.getElementCount();
    }

    private int getLineForOffset(int offset)
    {
	Document doc=editingPane.getDocument();
	if (doc==null) return -1;
	Element mainElem=doc.getDefaultRootElement();
	if (mainElem==null)
	    return -1;
	return mainElem.getElementIndex(offset);
    }

    private int getOffsetForLine(int line)
    {
	Document doc=editingPane.getDocument();
	if (doc==null) return -1;
	Element mainElem=doc.getDefaultRootElement();
	Element lineElem=mainElem.getElement(line);
	return lineElem.getStartOffset();
    }

    private TextPoint getLineAndColumnForOffset(int offset)
    {
	if (offset<0) return null;
	Document doc=editingPane.getDocument();
	if (doc==null) return null;
	Element mainElem=doc.getDefaultRootElement();
	if (mainElem==null)
	    return null;
	int x=mainElem.getElementIndex(offset);
	int y=offset-mainElem.getElement(x).getStartOffset();
	return new TextPoint(x, y);
    }

    public void select(int dot, int mark)
    {
	editingPane.setCaretPosition(dot);
	editingPane.moveCaretPosition(mark);
	//on Windows sometimes the selection is not highlighted; 
	//this is an attempt to fix that
	editingPane.getCaret().setSelectionVisible(true);
	scrollTo(mark);
    }

    public void selectAll()
    {
	Document doc=editingPane.getDocument();
	int dot=0;
	int mark=doc.getLength();
	select(dot, mark);
    }

    public void gotoLine(int line)
    {
	if (line<0) line=0;
	int max=getLineCount();
	if (max>0) max--;
	if (line>max) line=max;
	int offset=getOffsetForLine(line);
	if (offset!=-1)
	{
	    editingPane.setCaretPosition(offset);
	    scrollTo(offset);
	}
    }

    public void scrollTo(int offset)
    {
	try
	{
	    Rectangle r=editingPane.modelToView(offset);
	    if (r!=null) editingPane.scrollRectToVisible(r);
	    editingPane.getCaret().setVisible(true);
	}
	catch (BadLocationException badLuck)
	{
	    if (Debug.DEBUG) Debug.trace(this, Debug.DP2, badLuck);
	}
    }	

    public SyntaxDocument getSyntaxDocument()
    {
	return (SyntaxDocument) editingPane.getDocument();
    }

    public Document getDocument()
    {
	return editingPane.getDocument();
    }

    public Caret getCaret()
    {
	return editingPane.getCaret();
    }

    public JComponent getComponent()
    {
	return edcomp;
    }

    public void setWriteable(boolean writeable)
    {
	super.setWriteable(writeable);
	editingPane.setEditable(writeable);
    }

    public void save()
    {
	BusinessTracker.addBusy(this);
	setResourceBody(editingPane.getText().getBytes());
	super.save();
	undoManager.discardAllEdits();
	BusinessTracker.removeBusy(this);
    }

    public void load()
    {
	BusinessTracker.addBusy(this);
	//determine FileMode
	FileMode fm=FileMode.getModeForFilename(getResourceName());

	if (fm!=null)
	{
	    setFileMode(fm);
	}
	    
	byte[] body=getResourceBody();
	if (body!=null)
	{
	    boolean savedTokenizationState=isTokenizing();
	    if (savedTokenizationState) setTokenizing(false);
	    editingPane.setText(new String(body));
	    if (savedTokenizationState) 
	    {
		    setTokenizing(true);
		    getSyntaxDocument().retokenizeAll();
	    }
	}
	scrollTo(0);
	editingPane.requestFocus();
	editingPane.setCaretPosition(0);
	editingPane.getCaret().setVisible(true);
	undoManager.discardAllEdits();
	setDirty(false);
	BusinessTracker.removeBusy(this);
    }

}

/* $Log: SimpleTextEditor.java,v $
/* Revision 1.28  2001/02/16 18:15:10  smulloni
/* many fixes to TextEditorCustomizer.  FileMode and SyntaxStyle now have a
/* configData property (they will probably be made into sibling classes).
/*
/* Revision 1.27  2001/02/13 22:53:50  smulloni
/* adding a syntax highlighting mode for STML (Skunk Template Markup Language);
/* fixed a bug in SyntaxStyle in reading default.styles.
/*
/* Revision 1.26  2001/02/09 20:00:00  smulloni
/* fixed particularly nasty bug in GappedIntArray.set(int, int[]), and other
/* bugs in the syntax highlighting system.
/*
/* Revision 1.25  2001/02/06 23:15:53  smulloni
/* added syntax highlight and selectAll actions to the menu
/*
/* Revision 1.24  2001/02/06 22:13:41  smulloni
/* first more-or-less working version of syntax highlighting, with customization.
/*
/* Revision 1.23  2001/02/06 00:11:18  smulloni
/* struggle, perhaps futile, with the TextEditorCustomizer and other implicated
/* classes
/*
/* Revision 1.22  2001/02/02 23:30:33  smulloni
/* adding customization features to the text editor.
/*
/* Revision 1.21  2001/01/30 23:03:19  smulloni
/* beginning of integration of syntax highlighting into SimpleTextEditor.
/*
/* Revision 1.20  2000/12/21 18:53:13  smulloni
/* cosmetic improvements.
/*
/* Revision 1.19  2000/12/19 22:36:05  smulloni
/* adjustments to preamble.
/*
/* Revision 1.18  2000/12/15 23:22:08  smulloni
/* added word wrap to text editor.
/*
/* Revision 1.17  2000/12/14 23:09:17  smulloni
/* fixes to search and replace; partial fix to adding custom properties bug.
/*
/* Revision 1.16  2000/12/14 06:36:26  smulloni
/* changes to search and replace in text editor.
/*
/* Revision 1.15  2000/12/13 17:59:50  smulloni
/* added replaceSelection(String s)
/*
/* Revision 1.14  2000/12/08 16:36:48  smulloni
/* fixed bug that didn't update the Location field correctly when switching
/* buffers to an editor other than SimpleTextEditor.
/*
/* Revision 1.13  2000/12/04 23:51:16  smulloni
/* added ImageViewer; fixed word in SimpleTextEditor
/*
/* Revision 1.12  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.11  2000/12/03 20:40:06  smulloni
/* reconciling the minos.skunk.org cvs repository with sourceforge.
/*
/* Revision 1.11  2000/12/01 22:54:14  smullyan
/* added undo and basic editing functionality to text editor; defined some
/* keybindings in an inflexible way that will have to be changed soon.
/*
/* Revision 1.10  2000/12/01 16:25:54  smullyan
/* improvements to look and feel; fixed NPE in DAVFile; new actions for text
/* editor
/*
/* Revision 1.9  2000/11/29 23:16:06  smullyan
/* adding first rough cut of search capability to the text editor.  View
/* is being updated to allow components to be docked into the status bar.
/*
/* Revision 1.8  2000/11/28 19:36:20  smullyan
/* some bug fixes; notes.
/*
/* Revision 1.7  2000/11/28 00:17:31  smullyan
/* a fix that is meant to prevent a null pointer exception I noticed in the logs.
/*
/* Revision 1.6  2000/11/28 00:01:41  smullyan
/* added a status bar/minibuffer, with a location field showing the current line and
/* column number (for the SimpleTextEditor and kin only).
/*
/* Revision 1.5  2000/11/20 23:30:25  smullyan
/* more editor integration work.
/*
/* Revision 1.4  2000/11/18 04:36:07  smullyan
/* work on StateMonitor and related functionality.
/*
/* Revision 1.3  2000/11/16 23:16:48  smullyan
/* more preliminary work on editor integration
/*
/* Revision 1.2  2000/11/15 20:17:05  smullyan
/* added a Buffer interface, which is a wrapper around a displayable component.
/*
/* Revision 1.1  2000/11/15 19:45:37  smullyan
/* beginning of revamp of application to include multiple buffers.
/* */
