/*
 *  Copyright (c) 2000-2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.gui.editor.action;

import cmp.Boyer.Boyer;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.JOptionPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.Caret;
import javax.swing.text.Document;
import javax.swing.text.Segment;
import gnu.regexp.RE;
import gnu.regexp.REException;
import gnu.regexp.REMatch;
import org.skunk.assertF.Assertion;
import org.skunk.dav.client.gui.Buffer;
import org.skunk.dav.client.gui.ExplorerApp;
import org.skunk.dav.client.gui.ResourceManager;
import org.skunk.dav.client.gui.StateMonitor;
import org.skunk.dav.client.gui.View;
import org.skunk.dav.client.gui.editor.IncrementalSearchPanel;
import org.skunk.dav.client.gui.editor.ISearchPanel;
import org.skunk.dav.client.gui.editor.SimpleTextEditor;
import org.skunk.trace.Debug;


public class SearchAction extends AbstractAction
{
    public static final String SEARCH_IN_PROGRESS_PROPERTY="search_in_progress";
    
    private int matchLength;

    public void actionPerformed(ActionEvent ae)
    {
	Object source=ae.getSource();
	Assertion.assertF((source!=null), "source of ActionEvent is not null");

	View currentView=ExplorerApp.getAppContext().getCurrentView();
	Assertion.assertF((currentView!=null), "current view is not null");
	
	Buffer buffy=currentView.getFocussedBuffer();
	Assertion.assertF((buffy!=null), "focussed buffer is not null");
	if (!(buffy instanceof SimpleTextEditor))
	{
	    Debug.trace(this, Debug.DP3, "search invoked on inappropriate buffer");
	    return;
	}
	if (source instanceof ISearchPanel)
	{
	    //obtain search parameters
	    ISearchPanel panel=(ISearchPanel)source;
	    String command=ae.getActionCommand();
	    SimpleTextEditor editor=(SimpleTextEditor)buffy;
	    actOnCommand(command, panel, editor);
	}
	else 
	{
	    Debug.trace(this, Debug.DP3, "creating search panel");
	    ISearchPanel isp=createSearchPanel();
	    isp.focus();
	    StateMonitor.setProperty(SEARCH_IN_PROGRESS_PROPERTY, new Boolean(true), null);
	    showSearchPanel(currentView, isp);
	}
    }

    protected void showSearchPanel(View currentView, ISearchPanel panel)
    {
	currentView.dockStatus(panel.getComponent());
    }

    protected void hideSearchPanel(View currentView, ISearchPanel panel)
    {
	currentView.undockStatus(panel.getComponent());
    }

    private void endSearch(ISearchPanel panel, SimpleTextEditor editor)
    {
	Debug.trace(this, Debug.DP3, "search panel should now be undocked");
	hideSearchPanel(ExplorerApp.getAppContext().getCurrentView(), panel);
	panel.removeActionListener(this);
	panel=null;
	editor.getComponent().requestFocus();
	StateMonitor.setProperty(SEARCH_IN_PROGRESS_PROPERTY, new Boolean(false), null);
    }

    protected void actOnCommand(String command, ISearchPanel panel, SimpleTextEditor editor)
    {
	//obtain search parameters
	String searchString=panel.getSearchText();
	if (ISearchPanel.END_SEARCH.equals(command) 
	    || searchString==null 
	    || searchString.length()==0)
	{
	    endSearch(panel, editor);
	    return;
	}
	else Debug.trace(this, Debug.DP3, "searchString is "+searchString);
	Document doc=editor.getDocument();
	Caret caret=editor.getCaret();	
	int searchMode=panel.getSearchMode();
	boolean reverse=panel.isReverse();
	boolean replacingNext=ISearchPanel.REPLACE_NEXT.equals(command);
	boolean replacingAll=ISearchPanel.REPLACE_ALL.equals(command);

	do
	{
	    int dot=Math.min(caret.getDot(), caret.getMark());
	    int mark=Math.max(caret.getDot(), caret.getMark());
	    if (replacingNext || replacingAll)
	    {
		String replaceText=panel.getReplaceText();
		editor.replaceSelection(replaceText);
		mark=dot+replaceText.length();
		//editor.select(dot, mark);
		command=(reverse) 
		    ? ISearchPanel.SEARCH_FROM_CARET
		    : ISearchPanel.SEARCH_FROM_SELECTION;		
	    }
	    
	    //obtain object of search operation
	    
	    int offset, length;
	    Segment segment=new Segment();
	    
	    Assertion.assertF((command!=null), "action command for search is not null");
	    boolean isCaretSearch=command.equals(ISearchPanel.SEARCH_FROM_CARET);
	    boolean isSelectionSearch=command.equals(ISearchPanel.SEARCH_FROM_SELECTION);
	    boolean isStartOrEndSearch=command.equals(ISearchPanel.SEARCH_FROM_START_OR_END);
	    if (isCaretSearch || isSelectionSearch || isStartOrEndSearch)
	    {
		if (!reverse)
		{
		    Debug.trace(this, Debug.DP3, "calculating segment for forward search");
		    if (isCaretSearch)
		    {
			offset=dot;
		    }
		    else if (isSelectionSearch)
		    {
			offset=mark;
		    }
		    else offset=0;
		    Debug.trace(this, Debug.DP3, "offset is "+offset);
		    length=doc.getLength() - offset;
		    Debug.trace(this, Debug.DP3, "length of segment is "+length);
		}
		else
		{
		    Debug.trace(this, Debug.DP3, "calculating segment for backward search");
		    Debug.trace(this, Debug.DP3, "dot: "+dot + "; mark: "+mark);
		    if (isCaretSearch)
		    {
			Debug.trace(this, Debug.DP3, "caret search");
			length=dot;
		    }
		    else if (isSelectionSearch)
		    {
			Debug.trace(this, Debug.DP3, "selection search");
			length=mark;
		    }
		    else
		    {
			Debug.trace(this, Debug.DP3, "start or end search");
			length=doc.getLength();
		    }
		    offset=0;
		    Debug.trace(this, Debug.DP3, "offset is "+offset);
		    Debug.trace(this, Debug.DP3, "length of segment is "+length);
		}
		int[] found;
		if (length<=0) 
		    found=new int[] {-1, 0};
		else
		{
		    try
		    {
			doc.getText(offset, length, segment);
		    }
		    catch (BadLocationException badLuck)
		    {
			Debug.trace(this, Debug.DP2, badLuck);
		    }
		    //now performs search according to those parameters and manipulates the buffer component accordingly
		    found=search(searchString, searchMode, reverse, segment);
		}
		if (found[0]>=0)
		{
		    found[0]=found[0]+offset;
		    
		    handleMatch(editor, found);
		}
		else 
		{
		    handleNoMatch(editor);
		    replacingAll=false;
		}
	    }
	}
	while (replacingAll);
    }

    protected void handleMatch(SimpleTextEditor editor, int[] found)
    {
	editor.select(found[0], found[0]+found[1]);
    }

    protected void handleNoMatch(SimpleTextEditor editor)
    {
	String message=ResourceManager.getMessage(ResourceManager.SEARCH_FAILED_MESSAGE);
	String title=ResourceManager.getMessage(ResourceManager.SEARCH_FAILED_TITLE);
	JOptionPane.showMessageDialog(editor.getComponent(), 
				      message,
				      title,
				      JOptionPane.INFORMATION_MESSAGE);
    }

    protected ISearchPanel createSearchPanel()
    {
	ISearchPanel isp=new IncrementalSearchPanel();
	isp.addActionListener(new SearchAction());
	return isp;
    }

    private int[] search(String searchString, int searchMode, boolean reverse, Segment segment)
    {
	if (searchMode==ISearchPanel.REGEXP_MODE)
	{
	    try
	    {
		RE mrRegex=new RE(searchString);
		if (reverse)
		{
		    Debug.trace(this, Debug.DP3, "in reverse regex");
		    char[] text=getSegmentChars(segment);
		    Debug.trace(this, Debug.DP3, "\n\tsearchString: {0}\n\ttext: {1}",
				new Object[] {searchString, new String(text)});
		    
		    REMatch[] allMatches=mrRegex.getAllMatches(text);
		    if (allMatches!=null && allMatches.length>0)
		    {
			REMatch lastMatch=allMatches[allMatches.length-1];
			int start=lastMatch.getStartIndex();
			int length=lastMatch.getEndIndex()-start;
			return new int[] { start, length };
		    }
		}
		REMatch match=mrRegex.getMatch(getSegmentChars(segment));
		if (match!=null)
		{
		    int start=match.getStartIndex();
		    int length=match.getEndIndex()-start;
		    return new int[] {start, length};
		}
	    }
	    catch (REException rex)
	    {
		Toolkit.getDefaultToolkit().beep();
	    }
		    
	    return new int[] {-1, 0}; 
	}
	else 
	{
	    int matchLength=searchString.length();
	    if (reverse)
		return new int[] {reverseSearch(searchString, searchMode, getSegmentChars(segment)), matchLength};
	    else return new int[] {forwardSearch(searchString, searchMode, getSegmentChars(segment)), matchLength};
	}
    }

    private char[] getSegmentChars(Segment segment)
    {
	char[] retArray=new char[segment.count];
	System.arraycopy(segment.array, segment.offset, retArray, 0, segment.count);
	return retArray;
    }

    private int boyerMoore(String searchString, char[] text)
    {
	int i= Boyer.indexOf(text, searchString);
	Debug.trace(this, Debug.DP3, "searchString {0} found at {1}",
		    new Object[] {searchString, new Integer(i)});
	return i;
    }

    private int forwardSearch(String searchString, int searchMode, char[] text)
    {
	switch (searchMode)
	{
	case ISearchPanel.LITERAL_MODE:
	    return boyerMoore(searchString, text);
	    
	case ISearchPanel.CASE_MODE:
	    return boyerMoore(searchString.toLowerCase(), toLowerCase(text));
	    
	case ISearchPanel.REGEXP_MODE:
	    //fall through
	    
	default:
	    Assertion.assertF(false, "this is not reached");
	    return -1;
	}
    }

    private char[] toLowerCase(char[] text)
    {
	char[] lowered=new char[text.length];
	for (int i=0;i<text.length;i++)
	    lowered[i]=Character.toLowerCase(text[i]);
	return lowered;
    }

    private char[] reverse(char[] text)
    {
	int len=text.length;
	char[] clonedText=new char[len];	
	for (int i=0;i<len;i++)
	    clonedText[i]=text[len-1-i];
	return clonedText;
    }

    private String reverse(String s)
    {
	StringBuffer sb=new StringBuffer(s);
	sb.reverse();
	return sb.toString();
    }

    private int reverseSearch(String searchString, int searchMode, char[] text)
    {
	Assertion.assertF((searchMode!=ISearchPanel.REGEXP_MODE), "not in regexp mode");
	int reverseFind=forwardSearch(reverse(searchString), searchMode, reverse(text));
	if (reverseFind>=0)
	    reverseFind=text.length-reverseFind-searchString.length();
	return reverseFind;
	    
    }
}
/* $Log: SearchAction.java,v $
/* Revision 1.8  2001/07/22 19:11:07  smulloni
/* replaced the ORO regexp package with gnu-regexp
/*
/* Revision 1.7  2000/12/19 22:36:06  smulloni
/* adjustments to preamble.
/*
/* Revision 1.6  2000/12/14 23:09:17  smulloni
/* fixes to search and replace; partial fix to adding custom properties bug.
/*
/* Revision 1.5  2000/12/14 06:36:26  smulloni
/* changes to search and replace in text editor.
/*
/* Revision 1.4  2000/12/08 22:25:06  smulloni
/* added non-incremental search, and forward and reverse regex search, using
/* the org.apache.oro regex package.
/* */
