/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.method;

import java.io.IOException;
import java.util.Map;
import org.skunk.dav.client.AbstractWriteDAVMethod;
import org.skunk.dav.client.DAVConstants;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.DAVMethodName;
import org.skunk.dav.client.Depth;
import org.skunk.dav.client.IfHeader;
import org.skunk.minixml.MalformedXMLException;
import org.skunk.minixml.XMLDocument;
import org.skunk.minixml.XMLElement;
import org.skunk.minixml.XMLParser;
import org.skunk.minixml.XMLViewer;
import org.skunk.trace.Debug;

public class CopyMethod extends AbstractWriteDAVMethod
{
    private boolean overwrite;
    private String destination;
    private Depth depth;
    private String sourceLockToken;
    private String destLockToken;

    public CopyMethod(String source, 
		      String destination, 
		      boolean overwrite, 
		      Depth depth)
    {
	super(source);
	this.destination=destination;
	this.overwrite=overwrite;
	if (depth.equals(Depth.ONE))
	    throw new IllegalArgumentException("Depth must be 0 or Infinity");
	this.depth=depth;
    }

    public CopyMethod(String source, 
		      String destination)
    {
	this(source, destination, true, Depth.INFINITY);
    }

    public DAVMethodName getRequestMethodName()
    {
	return DAVMethodName.COPY;
    }

    /**
     * convenience function for the case when a locktoken is
     * associated with the source resource.  Incompatible with
     * accessing the IfHeader directly; may be deprecated in a future
     * version.
     */
    public void setSourceLockToken(String sourceLockToken)
    {
	this.sourceLockToken=sourceLockToken;
    }

    /**
     * convenience function for the case when a locktoken is
     * associated with the destination resource.  Incompatible with
     * accessing the IfHeader directly; may be deprecated in a future
     * version.
     */
    public void setDestLockToken(String destLockToken)
    {
	this.destLockToken=destLockToken;
    }

    public void processRequestHeaders()
    {
	Map headers=getRequestHeaders();
	headers.put(DAVConstants.DESTINATION_HEADER, destination);
	if (!overwrite)
	    headers.put(DAVConstants.OVERWRITE_HEADER, "F");
	if (depth.equals(Depth.ZERO))
	    headers.put(DAVConstants.DEPTH_HEADER, depth);

	boolean sourceNull=sourceLockToken==null;
	boolean destNull=destLockToken==null;
	if (! (sourceNull && destNull))
	{
	    IfHeader iffy=new IfHeader(true);
	    if (!sourceNull)
	    {
		iffy.addStateToken(getRequestURL(), 
				   sourceLockToken);
	    }
	    if (!destNull)
	    {
		iffy.addStateToken(destination,
				   destLockToken);
	    }
	    this.setIfHeader(iffy);
	}
	super.processRequestHeaders();
    }
}

//     public void processRequestBody()
//     {
// 	//PROPERTYBEHAVIOR stuff should go here, if specified.
// 	//note that apache mod_dav does not honor this; consequently, low priority task
//     }

/* $Log: CopyMethod.java,v $
/* Revision 1.6  2001/07/17 04:41:32  smulloni
/* integrated IfHeader more closely into method package; added lock test.
/*
/* Revision 1.5  2001/07/15 18:16:37  smulloni
/* fixed some copyright notices.
/*
/* Revision 1.4  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.3  2000/11/09 23:35:06  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
