/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/


package org.skunk.dav.client.method;

import org.skunk.dav.client.AbstractDAVMethod;
import org.skunk.dav.client.DAVConstants;
import org.skunk.dav.client.DAVMethodName;
import org.skunk.dav.client.Depth;
import org.skunk.dav.client.IfHeader;
import org.skunk.dav.client.LockScope;
import org.skunk.dav.client.PropertyHandler;
import org.skunk.dav.client.Timeout;
import org.skunk.minixml.MalformedXMLException;
import org.skunk.minixml.XMLDocument;
import org.skunk.minixml.XMLElement;
import org.skunk.minixml.XMLParser;
import org.skunk.minixml.XMLViewer;
import org.skunk.trace.Debug;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ListIterator;
import java.util.Map;


public class LockMethod extends AbstractDAVMethod
{
    private String lockToken;
    private Depth depth=Depth.ZERO;
    private Timeout timeout;
    private LockScope lockScope;
    private boolean isRefresh=false;
    private String userHref;
    private IfHeader ifHeader=null;
    
    private static final String lockRequestXML=createLockRequestXML();
    
    public LockMethod(String url, String userHref)
    {
	this(url, Depth.ZERO, null, LockScope.EXCLUSIVE, userHref);
    }
    
    public LockMethod(String url, Depth depth, Timeout timeout, LockScope lockScope, String userHref)
    {
	super(url);
	this.depth=depth;
	this.timeout=timeout;
	this.lockScope=lockScope;
	this.userHref=userHref;
    }
    
    public void setDepth(Depth depth)
    {
	if (depth==Depth.ONE)
	    throw new IllegalArgumentException("lock depth must be zero or infinity");
	else this.depth=depth;
    }
    
    public Depth getDepth()
    {
	return depth;
    }
    
    public Timeout getTimeout()
    {
	return timeout;
    }
    
    public void setTimeout(Timeout timeout)
    {
	this.timeout=timeout;
    }

    public LockScope getLockScope()
    {
	return this.lockScope;
    }

    public void setLockScope(LockScope lockScope)
    {
	this.lockScope=lockScope;
    }

    private XMLElement getLockScopeElement()
    {
	return new XMLElement((this.lockScope==LockScope.EXCLUSIVE)
			      ? DAVConstants.EXCLUSIVE_ELEM
			      : DAVConstants.SHARED_ELEM);
    }

    public void setIfHeader(IfHeader ifHeader)
    {
	this.ifHeader=ifHeader;
    }

    public IfHeader getIfHeader()
    {
	return this.ifHeader;
    }

    
    public void processRequestHeaders()
    {
	Map headers=getRequestHeaders();
	headers.put(DAVConstants.DEPTH_HEADER, depth);
	if (timeout!=null)
	    headers.put(DAVConstants.TIMEOUT_HEADER, timeout);
	if (!isRefresh()) headers.put(DAVConstants.CONTENT_TYPE, "text/xml");
	if (ifHeader!=null)
	    headers.put(DAVConstants.IF_HEADER, ifHeader.toString());
    }
    
    public DAVMethodName getRequestMethodName()
    {
	return DAVMethodName.LOCK;
    }
    
    public void processResponseHeaders()
    {
	//get lock-token
	int s=getStatus();
	if (s>=200 && s>300)
	{
	    Map headers=getResponseHeaders();
	    if (headers.containsKey(DAVConstants.LOCK_TOKEN_HEADER))
	    {
		Object lockTokenObj=headers.get(DAVConstants.LOCK_TOKEN_HEADER);
		this.lockToken=lockTokenObj.toString();
	    }
	}
    }
    
    public void processResponseBody() throws MalformedXMLException
    {
	int status=getStatus();
	if (status==207)
	{
	    super.processResponseBody();
	    
	}
	else if (status==200)
	{
	    byte[] body=getResponseBody();
	    if (body==null || getStatus()>=400) 
		return;
	    String s=new String(body).trim();
	    XMLParser parser=new XMLParser();
	    try
	    {
		XMLDocument doc=parser.parse(s);
		if (Debug.isDebug(this, Debug.DP8))
		{
		    XMLViewer.viewDocumentDialog(doc);
		}
		//capture locktoken if we don't already have it from the header
		if (lockToken==null)
		{
		    XMLElement elem=doc.getElement("prop/lockdiscovery/activelock/locktoken/href");
		    Object bodyLock=(elem!=null) ? elem.getChild(0) : null;
		    if (bodyLock!=null) lockToken=bodyLock.toString();
		}
		//populate DAVFile
		XMLElement propElem=doc.getElement("prop");
		if (propElem==null) return;
		for (ListIterator lit=propElem.children(); lit.hasNext();)
		{
		    Object propChile=lit.next();
		    if (propChile instanceof XMLElement)
			PropertyHandler.handleProperty(getDAVFile(), (XMLElement) propChile);
		}
	    }
	    catch (IOException oyVeh)
	    {
		oyVeh.printStackTrace();
		throw new MalformedXMLException(oyVeh.getMessage());
	    }
	}
    }
    
    public void setRefresh(boolean isRefresh)
    {
	this.isRefresh=isRefresh;
    }
    
    public boolean isRefresh()
    {
	return isRefresh;
    }
    
    public void processRequestBody()
    {
	if (!isRefresh) //for refreshes, do not include request body
	{
	    String s = MessageFormat.format(getLockRequestXML(),
					    new Object[] { getLockScopeElement(),
							   getUserHref() });
	    setRequestBody(s.getBytes());
	}
    }

    private static String getLockRequestXML()
    {
	return lockRequestXML;
    }
    
    private static String createLockRequestXML()
    {
	try
	{
	    XMLElement lockInfoElem=new XMLElement(DAVConstants.LOCKINFO_ELEM)
		.setAttribute(DAVConstants.XML_NAMESPACE_ATTR,
			      DAVConstants.DAV_NAMESPACE);
	    XMLElement lockScopeElem=new XMLElement(DAVConstants.LOCKSCOPE_ELEM)
		.addChild("{0}"); //new XMLElement(DAVConstants.EXCLUSIVE_ELEM));
	    XMLElement lockTypeElem=new XMLElement(DAVConstants.LOCKTYPE_ELEM)
		.addChild(new XMLElement(DAVConstants.WRITE_ELEM));
	    XMLElement ownerElem=new XMLElement(DAVConstants.OWNER_ELEM)
		.addChild(new XMLElement(DAVConstants.HREF_ELEM)
			  .addChild("{1}")); //message format code
	    lockInfoElem.addChild(lockScopeElem);
	    lockInfoElem.addChild(lockTypeElem);
	    lockInfoElem.addChild(ownerElem);
	    XMLDocument minidoc=new XMLDocument();
	    minidoc.addElement(DAVConstants.XML_BOILERPLATE);
	    minidoc.addElement(lockInfoElem);
	    return minidoc.toString()+DAVConstants.CRLF;
	}
	catch (MalformedXMLException malex) //only thrown in development
	{
	    Debug.trace(LockMethod.class, Debug.DP1, malex);
	    return null;
	}
    }
    
    public String getLockToken()
    {
	return lockToken;
    }
    
    public String getUserHref()
    {
	return userHref;
    }
}

/* $Log: LockMethod.java,v $
/* Revision 1.13  2001/07/18 20:20:53  smulloni
/* added IfHeader to LockMethod; fixed peculiar DAVFile bug (in getLock());
/* various fixes, new cases to auto.py
/*
/* Revision 1.12  2001/07/15 18:16:37  smulloni
/* fixed some copyright notices.
/*
/* Revision 1.11  2001/01/02 17:43:38  smulloni
/* changed debug level for XMLViewer popups to Debug.DP8.
/*
/* Revision 1.10  2000/12/03 23:53:26  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.9  2000/11/09 23:35:07  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
