/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.dav.client.method;

import org.skunk.dav.client.AbstractDAVMethod;
import org.skunk.dav.client.DAVConstants;
import org.skunk.dav.client.DAVFile;
import org.skunk.dav.client.DAVMethodName;
import org.skunk.dav.client.DAVProperty;
import org.skunk.dav.client.Depth;
import org.skunk.dav.client.PropertyHandler;
import org.skunk.minixml.MalformedXMLException;
import org.skunk.minixml.XMLDocument;
import org.skunk.minixml.XMLElement;
import org.skunk.trace.Debug;
import java.io.IOException;
import java.util.ListIterator;
import java.util.Map;
import java.util.StringTokenizer;

/**
 *  
 */
public class PropFindMethod extends AbstractDAVMethod
{
    private Depth depth;
    private PropFindQueryType queryType;
    private DAVProperty[] queryProps;
    
    public PropFindMethod(String url)
    {
	super(url);
        this.depth=Depth.ONE;
        this.queryType=PropFindQueryType.ALLPROP;
    }

    public PropFindMethod(String url, 
			  Depth depth, 
			  boolean allprop)
    {
	this(url, 
	     depth, 
	     ((allprop) ? PropFindQueryType.ALLPROP : PropFindQueryType.PROPNAME), 
	     null);
    }

    public PropFindMethod(String url, 
			  Depth depth, 
			  PropFindQueryType queryType, 
			  DAVProperty[] queryProps)
    {
        super(url);
        this.depth=depth;
        this.queryType=queryType;
	this.queryProps=queryProps;
    }

    public PropFindMethod(String url, 
			  Depth depth, 
			  DAVProperty[] queryProps)
    {
	this(url, depth, PropFindQueryType.PROP, queryProps);
    }

    public PropFindQueryType getQueryType()
    {
	return queryType;
    }

    public Depth getDepth()
    {
	return depth;
    }
    
    public void processRequestHeaders()
    {
	Map headerMap=getRequestHeaders();
	headerMap.put(DAVConstants.DEPTH_HEADER, depth);
    }
    
    public void processRequestBody()
    {
	try
	{
	    String namespaceCode="D"; //could be anything
	    XMLElement propfindElem=new XMLElement(DAVConstants.PROPFIND_ELEM, 
						   namespaceCode,
						   DAVConstants.DAV_NAMESPACE);
	    XMLElement queryTypeElem=new XMLElement(queryType.toString(), namespaceCode);
	    if (queryType.equals(PropFindQueryType.PROP))
	    {
		//add properties to be queried.
		if (queryProps!=null)
		{
		    for (int i=0;i<queryProps.length;i++)
			queryTypeElem.addChild(queryProps[i]);
		}
	    }
	    propfindElem.addChild(queryTypeElem);
	    XMLDocument doc=new XMLDocument();
	    doc.addElement(DAVConstants.XML_BOILERPLATE);
	    doc.addElement(propfindElem);
	    setRequestBody(new String(doc.toString()+DAVConstants.CRLF).getBytes());
	}
	catch (MalformedXMLException malex)
	{
	    Debug.trace(this, Debug.DP1, malex);
	}
    }
    
    public void processResponseHeaders()
    {
	//nothing to do
    }

    /**
     * by default, passes the properties off to PropertyHandler
     */
    protected void processPropElement(DAVFile file, XMLElement elem)
    {
	if (queryType==PropFindQueryType.PROPNAME)
	{
	    //cycle through children of propElem and store props in available list
	    for (ListIterator lit=elem.children();lit.hasNext();)
	    {
		Object temp=lit.next();
		if (! (temp instanceof XMLElement))
		    continue;
		XMLElement propElem=(XMLElement) temp;
		file.addAvailableProperty(new DAVProperty(propElem));
	    }
	}
	else if (queryType==PropFindQueryType.PROP) //this is rather unnecessary
	{
	    //cycle through queryProps
	    if (queryProps!=null)
	    {
		for (int i=0;i<queryProps.length;i++)
		{
		    DAVProperty dp=queryProps[i];
		    XMLElement propElem=elem.getChild(dp.getElementName(), 
						      dp.getNamespace());
		    if (propElem==null)
		    {
			Debug.trace(this, Debug.DP2,
				    "the property requested "
				    + dp + " was not found in response for " 
				    + file.getName());
		    }
		    else
		    {
			PropertyHandler.handleProperty(file, propElem);
		    }
		}
	    }
	}
	else super.processPropElement(file, elem);
    }	

    public DAVMethodName getRequestMethodName()
    {
	return DAVMethodName.PROPFIND;
    }
}

/* $Log: PropFindMethod.java,v $
/* Revision 1.15  2001/07/15 18:16:37  smulloni
/* fixed some copyright notices.
/*
/* Revision 1.14  2000/12/03 23:53:27  smulloni
/* added license and copyright preamble to java files.
/*
/* Revision 1.13  2000/11/09 23:35:08  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
