/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.minixml;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import org.skunk.trace.Debug;

/**
 *  a minimal tool for constructing Strings that are XML elements.
 *  This is not an all-purpose XML library.
 */
public class XMLElement
{
    private String elementName;
    private String elementNamespaceCode;
    private List attributes=Collections.synchronizedList(new ArrayList());
    private List children=Collections.synchronizedList(new ArrayList());
    private XMLElement parent = null;
    private String defaultNamespace=null;
    private HashMap namespaceHash=new HashMap();

    public static final String XMLNS_ATTR="xmlns";

    public XMLElement(String fullElementName)
    {
	int colonIndex=fullElementName.indexOf(":");
	this.elementName=fullElementName.substring(colonIndex+1);
	this.elementNamespaceCode=(colonIndex==-1) ? "" : fullElementName.substring(0, colonIndex);
    }

    public XMLElement(String elementName, String namespaceCode, String namespace)
    {
	Debug.trace(this, 
		    Debug.DP5, 
		    "elementName: {0}, namespaceCode: {1}, namespace: {2}",
		    new Object[] { elementName, namespaceCode, namespace });
				       
		    
	this.elementName=elementName;
	this.elementNamespaceCode=namespaceCode;
	if (namespaceCode!=null && namespaceCode.length()>0 && namespace!=null && namespace.length()>0)
	{
	    StringBuffer attrBuffer=new StringBuffer(XMLNS_ATTR);
	    attrBuffer.append(':');
	    attrBuffer.append(namespaceCode);
	    setAttribute(attrBuffer.toString(), namespace);
	}
    }

    public XMLElement(String elementName, String namespaceCode)
    {
	this(elementName, namespaceCode, null);
    }

    public String getFullElementName()
    {
	if (elementNamespaceCode!=null && elementNamespaceCode.length()>0)
	{
	    StringBuffer sb=new StringBuffer(elementNamespaceCode);
	    sb.append(':');
	    sb.append(elementName);
	    return sb.toString();
	}
	return elementName;
    }

    public String getElementName()
    {
	return elementName;
    }

    public String getElementNamespaceCode()
    {
	return elementNamespaceCode;
    }
    
    public XMLElement setAttribute(String name, String value)
    {
	if (name==null || value==null)
	    throw new IllegalArgumentException("nulls not supported for xml attributes");
	if (name.startsWith(XMLNS_ATTR))
	{
	    int colonIndex=name.indexOf(":");
	    if (colonIndex==-1)
	    {
		setDefaultNamespace(value);
	    }
	    else setNamespace(name.substring(colonIndex+1), value);
	}	    
	for (ListIterator li=attributes.listIterator();li.hasNext();)
	{
	    XMLAttribute attr=(XMLAttribute)li.next();
	    if (attr.getName().equals(name))
	    {
		attr.setValue(value);
		return this;
	    }
	}
	attributes.add(new XMLAttribute(name, value));
	return this;
    }
    
    public String getAttribute(String name)
    {
	for (ListIterator li=attributes.listIterator();li.hasNext();)
	{
	    XMLAttribute attr=(XMLAttribute)li.next();
	    if (attr.getName().equals(name))
		return attr.getValue();
	}
	return null;
    }
    
    public ListIterator attributes()
    {
	return attributes.listIterator();
    }
    
    public ListIterator children()
    {
	return children.listIterator();
    }
    
    public XMLElement addChild(Object child)
    {
	children.add(child);
	if (child instanceof XMLElement)
	    ((XMLElement)child).setParent(this);
	return this;
    }
    /**
     *  this method ignores namespace
     */
    public XMLElement getChild(String elementName, int index)
    {
	int i=0;
	for (ListIterator lit=children.listIterator();lit.hasNext();)
	{
	    Object child=lit.next();
	    if (child instanceof XMLElement 
		&& ((XMLElement)child).getElementName().equals(elementName))
	    {
		if (i==index) return (XMLElement) child;
		else i++;
	    }
	}
	return null;
    }

    public XMLElement getChildElement(int index)
    {
	if (index>=children.size()) return null;
	int i=0;
	for (ListIterator lit=children.listIterator();lit.hasNext();)
	{
	    Object o=lit.next();
	    if (o instanceof XMLElement)
	    {
		if (index==(i++))
		{
		    return (XMLElement) o;
		}
	    }
	}
	return null;
    }

    public Object getChild(int index)
    {
	if (index<children.size()) return children.get(index);
	return null;
    }

    /**
     *  this method ignores namespace and returns the first matching child
     */
    public XMLElement getChild(String elementName)
    {
	return getChild(elementName, 0);
    }

    /**
     *  returns the first matching child
     */
    public XMLElement getChild(String elementName, String namespace)
    {
	return getChild(elementName, namespace, 0);
    }

    public XMLElement getChild(String elementName, String namespace, int index)
    {
	int i=0;
	for (ListIterator lit=children.listIterator();lit.hasNext();)
	{
	    Object child=lit.next();
	    if (child instanceof XMLElement)
	    {
		Debug.trace(this, Debug.DP5, "child element: {0}", child);
		XMLElement childElem=(XMLElement) child;
		if (childElem.getElementName().equals(elementName))
		{
		    String childNamespace=childElem.getNamespace(childElem.getElementNamespaceCode());
		    Debug.trace(this, Debug.DP5, "child namespace: {0}", childNamespace);
		    if (namespace==null) 
		    {
			if (childNamespace==null)
			{
			    if (i==index) return childElem;
			    else i++;
			}
		    }
		    else if (childNamespace!=null && childNamespace.equals(namespace))
		    {
			if (i==index) return childElem;
			else i++;
		    }
		}
	    }
	}
	return null;
    }
    
    public boolean isEmpty()
    {
	return children.isEmpty();
    }
    
    public String toString()
    {
	StringBuffer sb=new StringBuffer();
	sb.append('<');
	if (elementNamespaceCode.length()>0)
	{
	    sb.append(elementNamespaceCode);
	    sb.append(':');
	}
	sb.append(elementName);
	for (ListIterator li=attributes.listIterator(); li.hasNext();)
	{
	    XMLAttribute attr=(XMLAttribute)li.next();
	    sb.append(' ');
	    sb.append(attr.getName());
	    sb.append("=\"");
	    sb.append(attr.getValue());
	    sb.append("\"");
	}
	if (!isEmpty())
	{
	    sb.append('>');
	    for (ListIterator li=children.listIterator();li.hasNext();)
	    {
		sb.append(li.next());
	    }
	    sb.append("</");
	    if (elementNamespaceCode.length()>0)
	    {
		sb.append(elementNamespaceCode);
		sb.append(':');
	    }
	    sb.append(elementName);
	    sb.append('>');
	}
	else
	{
	    sb.append("/>");
	}
	return sb.toString();	
    }

    private void setDefaultNamespace(String id)
    {
	this.defaultNamespace=id;
    }

    private void setNamespace(String name, String id)
    {
	Debug.trace(this, Debug.DP5, 
		    "in setNamespace with element {0}, name {1}, id {2}",
		    new Object[] {this, name, id});
	this.namespaceHash.put(name, id);
	this.elementNamespaceCode=name;
    }

    public String getDefaultNamespace()
    {
	//elements leave the default namespace null unless 
	//they or their parent have a defined default namespace
	if (this.defaultNamespace==null)
	{
	    if (this.parent!=null)
	    {
		return parent.getDefaultNamespace();
	    }
	    else return "";
	}
	else return this.defaultNamespace;
    }
	    

    public String getNamespace(String namespaceCode)
    {
	if (namespaceHash.containsKey(namespaceCode))
	    return namespaceHash.get(namespaceCode).toString();
	else if (parent!=null)
	    return parent.getNamespace(namespaceCode);
	return null;
    }

    public String getNamespace()
    {
	return getNamespace(getElementNamespaceCode());
    }

    protected void setParent(XMLElement parent)
    {
	this.parent=parent;
    }
}


/* $Log: XMLElement.java,v $
/* Revision 1.11  2001/09/01 18:25:56  smulloni
/* fixes for Zope interoperability; version 1.0.2.4 dev
/*
/* Revision 1.10  2001/07/17 03:00:04  smulloni
/* added XMLCData class to represent CDATA, and fixed CDATA parsing, which was
/* wrong in any case; added license preambles where I had forgotten to put them.
/*
/* Revision 1.9  2000/11/09 23:35:11  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
