/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

/**
 * XMLParser.java
 * a parser with a very minimal feature set
 * Created: Wed Sep 13 17:54:20 2000
 *
 * @author <a href="mailto:smulloni@smullyan.org">Jacob Smullyan</a>
 * @version $Revision: 1.10 $
 */

package org.skunk.minixml;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.StringTokenizer;

public class XMLParser
{
    public static final int OPEN=0;
    public static final int CLOS=1;
    public static final int QUES=2;
    public static final int EXCL=3;
    public static final int SLAS=4;
    public static final int EQUA=5;
    public static final int QUOT=6;
    public static final int DASH=7;
    public static final int COLO=8;
    public static final int WORD=9;
    public static final int EOF=10;
    public static final int UNKN=11;
    public static final int BRAC=12;
    
    public XMLDocument parse(String s) throws MalformedXMLException, IOException
    {
	return parse(s, false);
    }

    public XMLDocument parse(Reader reader) throws MalformedXMLException, IOException
    {
	return parse(reader, false);
    }

    public XMLDocument parse(String s, boolean lenient) throws MalformedXMLException, IOException
    {
	return parse(new StringReader(s), lenient);
    }

    public XMLDocument parse(Reader reader, boolean lenient) throws MalformedXMLException, IOException
    {
	Lexer lex=new Lexer(reader);
	XMLDocument doc=new XMLDocument(lenient);
	Object o;
	while ((o=getNextElement(lex))!=null)
	{
	    if (o.toString().trim().length()>0)
		doc.addElement(o);
	}
	return doc;
    }

    private Object getNextElement(Lexer lex) throws MalformedXMLException, IOException
    {
	int token=lex.getNextToken();
	Object  elem=null;
	switch (token)
	{
	case EOF:
	    //nothing to do, leave elem as null
	    break;
	case OPEN:
	    int nextTok;
	    while ((nextTok=lex.getNextToken())!=EOF)
	    {
		if (nextTok==UNKN)
		{
		    continue;
		}
		if (nextTok==QUES) //an instruction; eat up tokens
		{
		    elem=getProcessingOrXMLInstruction(lex);
		    break;
		}
		else if (nextTok==EXCL)
		{
		    nextTok=lex.getNextToken();
		    if (nextTok==WORD)
		    {
			lex.pushBack();
			elem=getDoctypeInstruction(lex);
			break;
		    }
		    else if (nextTok==DASH)
		    {
			if (lex.getNextToken()==DASH) //discard next dash
			{	    
			    elem=getComment(lex);
			    break;
			}
			else throw new MalformedXMLException("unknown token: \"<!-\"");
		    }
		    else if (nextTok==BRAC)
		    {
			if ((lex.getNextToken()==WORD) && lex.getTokenString().equals("CDATA"))
			{
			    nextTok=lex.getNextToken();
			    if (nextTok==BRAC)
			    {
				elem=getCData(lex);
				break;
			    }
			    else throw new MalformedXMLException("unknown token: \"<![CDATA\" without \"[\"");
			 
			}
			else throw new MalformedXMLException("unknown token: \"<![\" without CDATA");
		    }
		}
		else if (nextTok==WORD)
		{
		    lex.pushBack();
		    elem=getElement(lex);
		    break;
		}
		else if (nextTok==SLAS)
		{
		    break;
		}
	    }
	    break;
	case SLAS:
	    //fall through to case WORD
	case QUOT:
	case EQUA:
	case DASH:
	case COLO:
	case QUES:
	case EXCL:
	case WORD:
	    //leave the word to be gathered by the following routine
	    lex.pushBack();
	    elem=getPCData(lex);
	    break;
	case UNKN:
	    //drop through to default
	default:
	    return getNextElement(lex);
	}
	return elem;
    }
    /**
     *
     */
    private XMLPInstruction getProcessingOrXMLInstruction(Lexer lex) throws MalformedXMLException, IOException
    {
	String s=lex.readUntil("?>");
	if (s!=null) s=s.trim();
	return new XMLPInstruction(s);
    }

    /**
     *
     */
    private String getDoctypeInstruction(Lexer lex) throws MalformedXMLException, IOException
    {
	String s=lex.readUpToCloseTag("<", ">");
	if (s!=null) s=s.trim();
	return "<!"+s;
    }

    /**
     *
     */
    private XMLCData  getCData(Lexer lex) 
	throws MalformedXMLException, IOException
    {
	String data=lex.readUntil("]]>");
	return new XMLCData(data);
	//return "<![CDATA" + lex.readUntil("]]>") + "]]>";
    }

    /**
     *
     */
    private XMLComment getComment(Lexer lex) throws MalformedXMLException, IOException
    {
	String s=lex.readUntil("-->");
	if (s!=null) s=s.trim();
	return new XMLComment(s);
    }

    /**
     *  Precondition: lexer's next token will return a word
     */
    private XMLElement getElement(Lexer lex) throws MalformedXMLException, IOException
    {
	String tag=lex.readUntil(">");
	if (tag.endsWith("/")) //empty element
	{
	    XMLElement elem= parseOpeningTag(tag.substring(0, tag.length()-1));
	    return elem;
	}
	else
	{
	    XMLElement elem=parseOpeningTag(tag);
	    Object o;
	    while ((o=getNextElement(lex))!=null && o.toString().trim().length()>0)
	    {
		elem.addChild(o);
	    }
	    //should now be on a closing tag
	    lex.readUntil(">");
	    return elem;
	}
    }

    /**
     * Returns an XMLElement without children, based on an opening xml tag
     */
    private XMLElement parseOpeningTag(String tag) throws MalformedXMLException, IOException
    {
	XMLElement elem=null;
	StringTokenizer stk=new StringTokenizer(tag);
	if (stk.hasMoreTokens())
	{
	    String tagName=stk.nextToken();
	    elem=new XMLElement(tagName);
	    while (stk.hasMoreTokens())
	    {
		String attrName=stk.nextToken("= \n\r\t");
		String attrValue=null;
		if (stk.hasMoreTokens())
		{
		    attrValue=stk.nextToken();
		    if (attrValue.equals("=") && stk.hasMoreTokens())
		    {
			attrValue=stk.nextToken();
		    }
		    if (attrValue.length()>1
			&& (attrValue.startsWith("\"") &&  attrValue.endsWith("\""))
			|| (attrValue.startsWith("'") && attrValue.endsWith("'")))
		    {
			attrValue=attrValue.substring(1, attrValue.length()-1);
		    }
		}
		if (attrValue!=null)
		    elem.setAttribute(attrName, attrValue);
	    }
	}
	return elem;
    }

    private String getPCData(Lexer lex) throws MalformedXMLException, IOException
    {
	String s=lex.readUntil("<");
	lex.pushBack();
	return s.trim();
    }

    public static void main(String[] args) throws MalformedXMLException, IOException
    {
	if (args.length==0)
	{
	    System.out.println("for testing, pass file or files as arguments to be parsed");
	    System.exit(1);
	}
	XMLParser parser=new XMLParser();
	for (int i=0;i<args.length;i++)
	{
	    XMLDocument doc=parser.parse(new FileReader(args[i]));
	    XMLViewer.viewDocumentDialog(doc);
	}
	System.exit(0);
    }
}

class Lexer
{
    private StreamTokenizer st;

    public Lexer(Reader r)
    {
	if (r instanceof BufferedReader)
	    st=new StreamTokenizer(r);
	else st=new StreamTokenizer(new BufferedReader(r));
	resetSyntax();
    }

    private static final int TT_EOF=StreamTokenizer.TT_EOF;
    private static final int TT_WORD=StreamTokenizer.TT_WORD;
    private static final int TT_EOL=StreamTokenizer.TT_EOL;

    public int getNextToken() throws IOException
    {
	int tok= st.nextToken();
	int retTok=XMLParser.UNKN;
	
	switch (tok)
	{
	case TT_EOF:
	    retTok=XMLParser.EOF;
	    break;
	case '<':
	    retTok=XMLParser.OPEN;
	    break;
	case '>':
	    retTok=XMLParser.CLOS;
	    break;
	case '?':
	    retTok=XMLParser.QUES;
	    break;
	case '!':
	    retTok=XMLParser.EXCL;
	    break;
	case '/':
	    retTok=XMLParser.SLAS;
	    break;
	case ':':
	    retTok=XMLParser.COLO;
	    break;
	case '-':
	    retTok=XMLParser.DASH;
	    break;
	case '=':
	    retTok=XMLParser.EQUA;
	    break;
	case '"':
	    retTok=XMLParser.QUOT;
	    break;
	case '[':
	    retTok=XMLParser.BRAC;
	    break;
	case TT_WORD:
	    retTok=XMLParser.WORD;
	    break;
	case TT_EOL:
	    //do nothing
	    break;
	default:
	    System.err.println("unexpected token: "+tok);
	    return getNextToken();
	}
	return retTok;
    }

    public String getTokenString()
    {
	return st.sval;
    }

    public void pushBack()
    {
	st.pushBack();
    }

    /**
     * this method stops with the first terminator that matches
     */
    public String readUntil(String terminator) throws MalformedXMLException, IOException
    {
	StringBuffer sb=new StringBuffer();
	boolean found=false;
	st.resetSyntax();
	st.eolIsSignificant(false);
	while (st.nextToken()!=st.TT_EOF)
	{
	    if (st.ttype==st.TT_WORD)
		sb.append(st.sval);
	    else sb.append((char)st.ttype);
	    found=endsWith(sb, terminator);
	    if (found)
		break;
	}
	resetSyntax();
	if (!found)
	    throw new MalformedXMLException("terminator "+terminator+" not found before EOF");
	String s=sb.toString();
	return s.substring(0, s.indexOf(terminator));
    }
    
    private boolean endsWith(StringBuffer sb, String terminator)
    {
	if (sb==null || terminator==null) 
	    return false;
	int termlen=terminator.length();
	int sblen=sb.length();
	if (sblen<termlen) 
	    return false;
	boolean matches=true;
	int offset=sblen-termlen;
	for (int i=0;i<termlen;i++)
	{
	    matches &= (terminator.charAt(i)==sb.charAt(offset++));
	    if (!matches) 
		break;
	}
	return matches;
    }
    

    public String readUpToCloseTag(String  openTag, String closeTag) throws MalformedXMLException, IOException
    {
	StringBuffer sb=new StringBuffer();
	String retStr="";
	int tagsOpen=1;
	st.resetSyntax();
	st.eolIsSignificant(false);
	while (st.nextToken()!=st.TT_EOF)
	{
	    if (st.ttype==st.TT_WORD)
		sb.append(st.sval);
	    else sb.append((char)st.ttype);
	    retStr=sb.toString();
	    if (retStr.endsWith(openTag))
		tagsOpen++;
	    else if (retStr.endsWith(closeTag))
		tagsOpen--;
	    if (tagsOpen==0)
		break;
	}
	this.resetSyntax();
	if (tagsOpen!=0) throw new MalformedXMLException("unmatched tags at "+st.lineno());
	return retStr;
    }

    private void resetSyntax()
    {
	st.resetSyntax();
	st.eolIsSignificant(false);
	st.wordChars('a', 'z');
	st.wordChars('A', 'Z');
	st.wordChars('0', '9');
	st.wordChars(35, 38); // # $ % &
	st.wordChars(40, 46); // ( ) * + , - . 
	// thanks to Chris Knight for this patch:
	st.wordChars(91, 96);  // [ \ ] ^ _ `
	// end patch
	st.wordChars(123, 126);
	st.whitespaceChars(0, 32);
    }

    public String toString()
    {
	return st.toString();
    }
}


/* $Log: XMLParser.java,v $
/* Revision 1.10  2001/07/31 22:26:20  smulloni
/* included Chris Knight's patch for underscores and other well-beloved
/* characters.
/*
/* Revision 1.9  2001/07/17 03:00:04  smulloni
/* added XMLCData class to represent CDATA, and fixed CDATA parsing, which was
/* wrong in any case; added license preambles where I had forgotten to put them.
/*
/* Revision 1.8  2001/07/11 02:15:01  smulloni
/* the previous spi build targets have been folded into the main build;
/* also added a javadoc target.
/*
/* Revision 1.7  2000/12/08 05:50:30  smulloni
/* fixed MessageCatalogEditor.  The spi features are now a special build option,
/* and editors are loaded through reflection.
/*
/* Revision 1.6  2000/12/06 23:59:56  smulloni
/* added commented-out code that I mean to check (possible performance
/* improvement)
/*
/* Revision 1.5  2000/11/09 23:35:12  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
