/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.minixml;

import java.awt.BorderLayout;
import java.awt.Component;
import java.util.Enumeration;
import java.util.ListIterator;
import java.util.Vector;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;

/**
 * displays an XMLDocument in a JTree
 */
public class XMLViewer extends JPanel
{
    private JTree xmlTree;
    private static final String[] tooltips=
    {
	"XML element",
	"PI",
	"CDATA",
	"PCDATA",
	"XML comment",
	"DOCTYPE"
    };

    public XMLViewer(XMLDocument doc)
    {
	super();
	TreeNode rootNode=new XMLTreeNode((doc.hasRoot()) 
					  ? (Object)doc.getRootElement() 
					  : (Object)doc);
	xmlTree=new JTree(new DefaultTreeModel(rootNode));
	initComponents();
    }

    
    private void initComponents()
    {
	xmlTree.setCellRenderer(new DefaultTreeCellRenderer() {
		public Component getTreeCellRendererComponent(JTree tree,
							      Object value,
							      boolean sel,
							      boolean expanded,
							      boolean leaf,
							      int row,
							      boolean hasFocus) 
		{
		    Component comp=super.getTreeCellRendererComponent(tree,
								      value, sel,
								      expanded,
								      leaf, row,
								      hasFocus);
		    if (value instanceof XMLTreeNode && comp instanceof JComponent)
		    {
			int type=((XMLTreeNode)value).getType();
			if (type<XMLViewer.tooltips.length)
			    ((JComponent)comp).setToolTipText(XMLViewer.tooltips[type]);
		    }
		    return comp;
		}
	    });
	setLayout(new BorderLayout(2,2));
	add(new JScrollPane(xmlTree), BorderLayout.CENTER);	
    }

    public static void viewDocumentDialog(XMLDocument doc)
    {
	XMLViewer viewer=new XMLViewer(doc);
	JOptionPane.showMessageDialog(null, viewer, "XML Document", JOptionPane.INFORMATION_MESSAGE, null);
    }
}

class XMLTreeNode extends DefaultMutableTreeNode
{
    public static final int ELEMENT_TYPE=0;
    public static final int PI_TYPE=1;
    public static final int CDATA_TYPE=2;
    public static final int PCDATA_TYPE=3;
    public static final int COMMENT_TYPE=4;
    public static final int DOCTYPE_TYPE=5;
    public static final int DOCUMENT_TYPE=6;
    public static final int UNKNOWN_TYPE=7;

    private int type;
    private Vector children=new Vector();
    private Object xmlObject;

    public XMLTreeNode(Object xmlObject)
    {
	this(xmlObject, true);
    }

    public XMLTreeNode(Object xmlObject, boolean allowsChildren)
    {
	super(getUserObject(xmlObject), allowsChildren);
	this.xmlObject=xmlObject;
	setType();
	init();
    }

    protected Object getXMLObject()
    {
	return xmlObject;
    }

    private static Object getUserObject(Object xmlObject)
    {
	if (xmlObject instanceof XMLElement)
	{
	    XMLElement elem=(XMLElement) xmlObject;
	    StringBuffer sb=new StringBuffer();
	    String ns=elem.getElementNamespaceCode();
	    if (ns!=null && ns.length()>0)
	    {
		sb.append(ns.trim());
		sb.append(':');
	    }
	    sb.append(elem.getElementName());
	    for (ListIterator lit=elem.attributes();lit.hasNext();)
	    {
		sb.append(" ");
		XMLAttribute attr=(XMLAttribute)lit.next();
		sb.append(attr.getName().trim());
		sb.append("=");
		sb.append(attr.getValue().trim());
	    }
	    return sb.toString();
	}
	else if (xmlObject instanceof XMLDocument)
	    return "XML Document";
	else return xmlObject;	    
    }
    
    private void init()
    {
	switch (type)
	{
	case ELEMENT_TYPE:
	    XMLElement elem=(XMLElement)xmlObject;
	    for (ListIterator lit=elem.children();lit.hasNext();)
		children.addElement(new XMLTreeNode(lit.next()));
	    break;
	case DOCUMENT_TYPE:
	    XMLDocument doc=(XMLDocument)xmlObject;
	    for (ListIterator lit=doc.contents();lit.hasNext();)
		children.addElement(new XMLTreeNode(lit.next()));
	    break;
	default:
	    //do nothing
	}
    }

    public TreeNode getChildAt(int index)
    {
	if (index>=0 && index<children.size())
	    return (TreeNode) children.elementAt(index);
	else return null;
    }

    public Enumeration children()
    {
	return children.elements();
    }

    public int getChildCount()
    {
	return children.size();
    }

    private void setType()
    {
	if (xmlObject instanceof XMLElement)
	    type=ELEMENT_TYPE;
	else if (xmlObject instanceof XMLPInstruction)
	    type=PI_TYPE;
	else if (xmlObject instanceof XMLComment)
	    type=COMMENT_TYPE;
	else if (xmlObject instanceof String)
	{
	    if (((String)xmlObject).startsWith("<!DOCTYPE"))
		type=DOCTYPE_TYPE;
	    else if (((String)xmlObject).startsWith("<![CDATA"))
		type=CDATA_TYPE;
	    else type=PCDATA_TYPE;
	}
	else if (xmlObject instanceof XMLDocument)
	    type=DOCUMENT_TYPE;
	else type=UNKNOWN_TYPE;
    }

    public int getType()
    {
	return type;
    }
	    
}

    
			  

/* $Log: XMLViewer.java,v $
/* Revision 1.3  2001/07/17 03:00:04  smulloni
/* added XMLCData class to represent CDATA, and fixed CDATA parsing, which was
/* wrong in any case; added license preambles where I had forgotten to put them.
/*
/* Revision 1.2  2000/11/09 23:35:12  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
