package org.skunk.spi;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.skunk.assertF.Assertion;
import org.skunk.trace.Debug;

/**
 *  parses a python dictionary with the following characteristics:
 *    a. all keys are strings.
 *    b. some values are dictionaries are dictionaries whose keys are 
 *       strings and whose values are also strings.
 */
public class DictParser
{
    public static final boolean ENCODE=true;

    public static String toString(Map hm)
    {
	StringBuffer sb=new StringBuffer("{");
	for (Iterator it=hm.keySet().iterator();it.hasNext();)
	{
	    Object key=it.next();
	    Object value=hm.get(key);
	    String valStr =(value instanceof Map) 
		? toString((Map)value)
		: quote(value);
	    sb.append(quote(key))
		.append(": ")
		.append(valStr);
	    if (it.hasNext()) sb.append(",\n ");
	}
	sb.append("}");
	return sb.toString();
    }

    private static char[] hexdigit = "0123456789ABCDEF".toCharArray();

    /**
     * the following is modified from Jython source code: 
     *   @see org.python.core.PyString.unicodeescape()
     */
    private static String quote(Object s) 
    {
	String str=s.toString();
        int size = str.length();
        StringBuffer v = new StringBuffer(size);
        boolean unicode = false;  
	char quote = str.indexOf('\'') >= 0 && 
	    str.indexOf('"') == -1 ? '"' : '\'';
	v.append(quote);

        for (int i = 0; size-- > 0; ) 
	{
            int ch = str.charAt(i++);
            /* Map 16-bit characters to '\\uxxxx' */       
	    if (ENCODE && ch >= 256) 
	    {
                if (!unicode) 
		{
                   v.insert(0, 'u');
                   unicode = true;
                }
                v.append('\\');
                v.append('u');
                v.append(hexdigit[(ch >> 12) & 0xf]);
                v.append(hexdigit[(ch >> 8) & 0xf]);
                v.append(hexdigit[(ch >> 4) & 0xf]);
                v.append(hexdigit[ch & 15]);
            }
	    if (ENCODE && ch < ' ' || ch >= 128) 
	    {
                v.append('\\');
                v.append(hexdigit[(ch >> 6) & 7]);
                v.append(hexdigit[(ch >> 3) & 7]);
                v.append(hexdigit[ch & 7]);
	    }
	    else if (ch == '\n') v.append("\\n");
            else if (ch == '\t') v.append("\\t");
            else if (ch == '\b') v.append("\\b");
            else if (ch == '\f') v.append("\\f");
            else if (ch == '\r') v.append("\\r");
	    else if (ch=='\\')
	    {
		if (ENCODE && testEscape(str, i))
		    v.append('\\');
		else
		    v.append("\\\\");
	    }
            /* Copy everything else as-is */
            else
                v.append((char) ch);
        }
	v.append(quote);
        return v.toString();
    }

    private static boolean testEscape(String s, int index)
    {
	if (s.length()>(index+1))
	{
	    char c=s.charAt(index+1);
	    if (c=='n'
		|| c=='t'
		|| c=='b'
		|| c=='f'
		|| c=='r'
		)
		return true;
	}
	if (s.length()>=(index+3))
	{
	    boolean b=true;
	    for (int i=0;i<3;i++)
	    {
		char c=s.charAt(i+index);		
		b=b && (Character.isDigit(c) );
// 			|| (i==0 && c=='x'));
		if (!b) break;
	    }
	    return b;
	}
	return false;
    }

    /**
     * parses a string representation of a Python dictionary and returns a HashMap
     */
    public static HashMap fromString(String s)
    {
	return fromReader(new StringReader(s));
    }

    public static HashMap fromReader(Reader reader)
    {
	DictLexer lexer=new DictLexer(reader);
	try
	{
	    return lexer.dictionary();
	}
	catch (IOException oyVeh)
	{
	    Debug.trace(DictParser.class, Debug.DP1, oyVeh);
	    return null;
	}
    }

    static class DictLexer
    {
	private StreamTokenizer st;

	static final int SINGLE_QUOTE='\'';
	static final int DOUBLE_QUOTE='"';
	static final int BACKSLASH='\\';
	static final int COMMA=',';
	static final int BRACELEFT='{';
	static final int BRACERIGHT='}';
	static final int WORD=-56;
	static final int COLON=':';
	static final int UNKNOWN=-99;

	DictLexer(Reader reader)
	{
	    if (reader instanceof BufferedReader)
		st=new StreamTokenizer(reader);
	    else st=new StreamTokenizer(new BufferedReader(reader));
	    resetSyntax();
	}

	int next() throws IOException
	{
// 	    Debug.trace(DictLexer.class, Debug.DP4, "in next()");
	    int tok=st.nextToken();
	    int retTok=UNKNOWN;
	    switch (tok)
	    {
	    case ',':
		retTok = COMMA;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found comma");
		break;
	    case '{':
		retTok = BRACELEFT;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found braceleft");
		break;
	    case '}':
		retTok = BRACERIGHT;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found braceright");
		break;
	    case '\'':
		retTok = SINGLE_QUOTE;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found single quote");
		break;
	    case '"':
		retTok = DOUBLE_QUOTE;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found double quote");
		break;
	    case '\\':
		retTok = BACKSLASH;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found backslash");
		break;
	    case ':':
		retTok=COLON;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found colon");
		break;
	    case StreamTokenizer.TT_WORD:
		retTok = WORD;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found word: "+st.sval);
		break;
	    case StreamTokenizer.TT_EOL:
		retTok = UNKNOWN;
// 		Debug.trace(DictLexer.class, Debug.DP4, "found unknown!");
		break;
	    default:
// 		Debug.trace(DictLexer.class, Debug.DP4, "dropped through to default!");
		retTok = next();
	    }
	    return retTok;
	}

	int peek() throws IOException
	{
// 	    Debug.trace(DictLexer.class, Debug.DP4, "in peek()");
	    int i=st.nextToken();
	    st.pushBack();
	    return i;
	}

	HashMap dictionary() throws IOException
	{
//  	    Debug.trace(DictLexer.class, Debug.DP4, "in dictionary()");
	    HashMap hm=new HashMap();
// 	    Debug.trace(DictLexer.class, Debug.DP4, "about to consume a braceleft...");
	    expect(BRACELEFT, next());
// 	    Debug.trace(DictLexer.class, Debug.DP4, "...braceleft consumed");
	    while (true)
	    {
		if (peek()==BRACERIGHT)
		    break;
		Object key, value;
		key=string();
// 		Debug.trace(DictLexer.class, Debug.DP4, "key found: "+key);
// 		Debug.trace(DictLexer.class, Debug.DP4, "about to consume a colon...");
		expect(COLON, next());
// 		Debug.trace(DictLexer.class, Debug.DP4, "...colon consumed");
		value=(peek()==BRACELEFT) ? (Object) dictionary() : (Object) string();
// 		Debug.trace(DictLexer.class, Debug.DP4, "value found: "+value);
		hm.put(key, value);
		if (peek()==COMMA)
		{
// 		    Debug.trace(DictLexer.class, Debug.DP4, "found comma in the right place, skipping");
		    next();
		}
	    }
// 	    Debug.trace(DictLexer.class, Debug.DP4, "about to consume a braceright...");
	    expect(BRACERIGHT, next());
// 	    Debug.trace(DictLexer.class, Debug.DP4, "...braceright consumed");
	    return hm;
	}

	String string() throws IOException
	{
// 	    Debug.trace(DictLexer.class, Debug.DP4, "in string()");
	    int tok=next();
	    Assertion.assertF((tok==SINGLE_QUOTE || tok==DOUBLE_QUOTE), "token is a quote: "+(char)tok);
	    String s=readUntil((char)tok);
	    expect(tok, next());
// 	    Debug.trace(this, Debug.DP4, "returning string: "+s);
	    return s;
	}

	String readUntil(char terminator) throws IOException
	{
	    StringBuffer sb=new StringBuffer();
	    boolean found=false;
	    boolean escape=false;
	    st.resetSyntax();
	    st.eolIsSignificant(false);
	    while (st.nextToken()!=st.TT_EOF)
	    {
		if (st.ttype==st.TT_WORD)
		    sb.append(st.sval);
		else 
		{
		    escape= (st.ttype=='\\');
		    sb.append((char)st.ttype);
		}
		found=(!escape) && (sb.charAt(sb.length()-1)==terminator);
		if (escape)
		{
		    if (peek()=='\\')
		    {
			next();
		    }
		    escape=false;
		}
		
		if (found)
		{
		    Debug.trace(DictLexer.class, Debug.DP4, "found terminator: " + terminator);
		    st.pushBack();
		    break;
		}
	    }
	    resetSyntax();
	    Assertion.assertF(found, "terminator found: "+terminator);
	    sb.setLength(sb.length()-1);
	    String s=unescape(sb.toString());
// 	    Debug.trace(DictLexer.class, Debug.DP4, "returning from readUntil(): "+s);
	    return s;
	}

	String unescape(String inputString)
	{
	    int len=inputString.length();
	    StringBuffer sb=new StringBuffer(len);
	    int offset=0;
	    int slash;
	    while((slash=inputString.indexOf("\\", offset))>=0)
	    {
		sb.append(inputString.substring(offset, slash));
		offset=slash+1;
		if (len>=slash+4)
		{
		    String possibleOctal=inputString.substring(slash+1, slash+4);
		    try
		    {
			char c=(char) Integer.parseInt(possibleOctal, 8);
			sb.append(c);
			offset=offset+3;
		    }
		    catch (NumberFormatException nafta)
		    {
			sb.append("\\");
			sb.append(possibleOctal);
			offset=offset+3;
		    }
		}
		else sb.append("\\");
	    }
	    if (offset<=len-1)
	    {
		sb.append(inputString.substring(offset));
	    }
	    return sb.toString();
	}

	void expect(int expected, int received)
	{
	    Assertion.assertF((expected==received), expected + " equals " + received);
	}

	private void resetSyntax()
	{
	    st.resetSyntax();
	    st.eolIsSignificant(false);
	    st.wordChars('a', 'z');
	    st.wordChars('A', 'Z');
	    st.wordChars('0', '9');
	    st.whitespaceChars(0, 32);
	}
    }

    /**
     *  test method.
     */
    public static void main(String[] args) throws IOException
    {
	String filename=args[0];
	FileReader fr=new FileReader(filename);
	HashMap hm=fromReader(fr);
	System.out.println(hm);
    }	
}

/* $Log: DictParser.java,v $
/* Revision 1.7  2001/02/20 23:55:43  smulloni
/* fencepost error in DictParser fixed
/*
/* Revision 1.6  2000/12/21 18:53:13  smulloni
/* cosmetic improvements.
/*
/* Revision 1.5  2000/12/13 22:25:33  smulloni
/* now save catalogs explicitly in 8859_1 (ISO Latin-1) encoding
/*
/* Revision 1.4  2000/12/13 15:28:24  smulloni
/* removing improper check for hex escape
/*
/* Revision 1.3  2000/12/11 22:36:13  smulloni
/* improved support for input of UTF.
/*
/* Revision 1.2  2000/12/08 05:50:30  smulloni
/* fixed MessageCatalogEditor.  The spi features are now a special build option,
/* and editors are loaded through reflection.
/* */
