package org.skunk.swing;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.ListIterator;
import javax.swing.Box;
import javax.swing.DefaultListModel;
import javax.swing.DefaultListSelectionModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import org.skunk.trace.Debug;

/**
 * <P>a widget for making an ordered selection from a list of items.
 * Two JLists are separated vertically by a stack of four buttons,
 * add, remove, up and down.  The add and remove buttons move
 * items from the left Jlist to the right and vice versa, respectively.
 * The up and down buttons move selected items of the right JList up 
 * and down the list.
 *
 * <P>Internationalized applications should set the label properties.
 */
public class DualListPanel extends JPanel
{
    /**
     * keeps track of all elements in both panes
     */
    private DefaultListModel listModel;

    /**
     * internal list models for the two JLists
     */
    private DefaultListModel sourceModel, sinkModel;

    /**
     * 	selection model keeps track of which goes in which pane 
     */
    private DefaultListSelectionModel selectionModel;

    private JList sourceList, sinkList;
    private JButton addButton, removeButton, upButton, downButton;

    /**
     * reference to all items, so when they
     * bounce back and forth between the two lists they don't get reordered
     */
    private ArrayList allItems;
    private Object[] sourceItems, sinkItems;

    private JLabel sourceListLabel, sinkListLabel;
    private String sourceListLabelText, sinkListLabelText;

    //default values for button labels -- should not be used in real life, not i18n!!!!!
    private static final String ADD="-->";
    private static final String REMOVE="<--";
    private static final String UP="up";
    private static final String DOWN="down";
    private static final String POSSIBLE_VALUES="Possible Values";
    private static final String SELECTED_VALUES="Selected Values";
    
    /**
     * constructs a DualListPanel with the given source and sink values.
     * @param sourceItems the source values (will appear in the left JList)
     * @param sinkItems the sink values (will appear in the right JList)
     */
    public DualListPanel(Object[] sourceItems, Object[] sinkItems)
    {
	this(sourceItems, sinkItems, POSSIBLE_VALUES, SELECTED_VALUES);
    }

    /**
     * constructs a DualListPanel with the given source and sink values and labels.
     * @param sourceItems the source values (will appear in the left JList)
     * @param sinkItems the sink values (will appear in the right JList)
     * @param sourceLabel the label text for the left JList
     * @param sinkLabel the label text for the right JList
     */
    public DualListPanel(Object[] sourceItems, Object[] sinkItems, String sourceLabel, String sinkLabel)
    {
	super();
	this.allItems=new ArrayList();
	sourceModel=new DefaultListModel();
	sinkModel=new DefaultListModel();
	this.sourceListLabelText=sourceLabel;
	this.sinkListLabelText=sinkLabel;
	this.sourceItems=sourceItems;
	this.sinkItems=sinkItems;
	//add initial data to models
	initModels();
	initComponents();
	initLayout();

    }

    private  void _addListItem(Object item, boolean chosen)
    {
	synchronized (allItems)
	{
	    if (!allItems.contains(item)) allItems.add(item);
	    int itemIndex=allItems.size()-1;
	}
	DefaultListModel tmp=(chosen) ? sinkModel : sourceModel;
	if (!tmp.contains(item)) tmp.addElement(item);
	tmp=(!chosen) ? sinkModel : sourceModel;
	if (tmp.contains(item)) tmp.removeElement(item);
    }

    /* addListItem() and removeListItem() need to be written */

//     public void addListItem(Object item, boolean chosen)
//     {
//     }

//     public void removeListItem(Object item)
//     {
//     }

    /**
     * returns panel and models to original state at time of construction.
     */
    public void reset()
    {
	Debug.trace(this, Debug.DP3, "in reset()");
	sourceModel.clear();
	sinkModel.clear();
	initModels();
	repaint();
    }

    /**
     * populates the JLists with the given source and sink values.
     * @param sourceItems the new contents of the left JList
     * @param sinkITems the new contents of the right JList
     */
    public void reset(Object[] sourceItems, Object[] sinkItems)
    {
	this.sourceItems=sourceItems;
	this.sinkItems=sinkItems;
	reset();
    }

    /**
     * sets the contents of the right JList to the given values,
     * and places other values currently in the DualListPanel in the left JList
     * @param chosenItems the items the right JList should contain
     */
    public void reset(Object[] chosenItems)
    {
	Debug.trace(this, Debug.DP3, "in reset(Object[])");
	ArrayList tmp=new ArrayList();
	for (ListIterator lit=allItems.listIterator();lit.hasNext();)
	{
	    Object nextObj=lit.next();
	    if (!inArray(chosenItems, nextObj))
	    {
		tmp.add(nextObj);
	    }
	}
	reset(tmp.toArray(), chosenItems);
    }

    private static boolean inArray(Object[] objArray, Object thing)
    {
	for (int i=0;i<objArray.length;i++)
	{
	    if (objArray[i].equals(thing))
		return true;
	}
	return false;
    }


    private void initModels()
    {
	if (sourceItems!=null)
	{
	    for (int i=0;i<sourceItems.length;i++)
	    {
		_addListItem(sourceItems[i], false);
	    }
	}
	if (sinkItems!=null)
	{
	    for (int i=0;i<sinkItems.length;i++)
	    {
		_addListItem(sinkItems[i], true);
	    }
	}
    }

    /**
     * returns the items in the right JList
     * @return the chosen items
     */
    public Object[] getChosenItems()
    {
	return sinkModel.toArray();
    }

    /**
     * add a listener to the data model of the left JList
     * @param ldl the listener
     */
    public void addListDataListener(ListDataListener ldl)
    {
	sinkModel.addListDataListener(ldl);
    }

    /**
     * remove a listener from the data model of the left JList
     * @param ldl the listener
     */
    public void removeListDataListener(ListDataListener ldl)
    {
	sinkModel.removeListDataListener(ldl);
    }

    private void initComponents()
    {
	//create the JLists

	sourceList=new JList(sourceModel);
	sourceList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	sourceList.addListSelectionListener(new SourceSelectionListener());

	sinkList=new JList(sinkModel);
	sinkList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	sinkList.addListSelectionListener(new SinkSelectionListener());

	//labels
	sourceListLabel=new JLabel(sourceListLabelText);
	sinkListLabel=new JLabel(sinkListLabelText);

	//create buttons
	addButton=new JButton(ADD);
	addButton.setEnabled(false);
	removeButton=new JButton(REMOVE);
	removeButton.setEnabled(false);
	upButton=new JButton(UP);
	upButton.setEnabled(false);
	downButton=new JButton(DOWN);
	downButton.setEnabled(false);

	//add listeners
	addButton.addActionListener(new AddAction());
	removeButton.addActionListener(new RemoveAction());
	upButton.addActionListener(new UpAction());
	downButton.addActionListener(new DownAction());		
    }

    private void initLayout()
    {
	GridBagConstraints gbc=new GridBagConstraints();
	gbc.gridx=0;
	gbc.gridy=0;
	gbc.gridwidth=1;
	gbc.gridheight=1;
	gbc.fill=gbc.NONE;
	gbc.insets=new Insets(2, 2, 2, 2);
	gbc.anchor=gbc.CENTER;

	setLayout(new GridBagLayout());
	gbc.weighty=0.1;
	add(sourceListLabel, gbc);
	gbc.weighty=0.0;
	gbc.gridx=2;
	add(sinkListLabel, gbc);

	gbc.gridy++;
	gbc.gridx=gbc.RELATIVE;
	gbc.weightx=1.0;
	gbc.weighty=0.9;	
	add(new JScrollPane(sourceList), gbc);

	JPanel buttonBox=new JPanel(new GridLayout(4, 1));
	buttonBox.add(addButton);
	buttonBox.add(removeButton);
	buttonBox.add(upButton);
	buttonBox.add(downButton);
	
	gbc.weightx=0.5;
	gbc.weighty=0.0;
	add(buttonBox, gbc);
	
	gbc.weightx=1.0;
	add(new JScrollPane(sinkList), gbc);	
    }
    
    /**
     * provides access to the add button
     * @return the add button
     */
    public JButton getAddButton() { return addButton; }

    /**
     * provides access to the remove button
     * @return the remove button
     */
    public JButton getRemoveButton() { return removeButton; }

    /**
     * provides access to the up button
     * @return the up button
     */
    public JButton getUpButton() { return upButton; }

    /**
     * provides access to the down button
     * @return the down button
     */
    public JButton getDownButton() { return downButton; }

    /**
     * sets the preferred size of both lists.
     * @param d the preferred size
     */
    public void setListPreferredSize(Dimension d)
    {
	sourceList.setPreferredSize(d);
	sinkList.setPreferredSize(d);
	revalidate();
    }

    /**
     * sets the label of the add button.
     * @param addLabel the new label
     */
    public void setAddLabel(String addLabel)
    {
	addButton.setText(addLabel);
    }

    /**
     * sets the label of the remove button
     * @param removeLabel the new label
     */
    public void setRemoveLabel(String removeLabel)
    {
	removeButton.setText(removeLabel);
    }

    /**
     * sets the label of the up button
     * @param upLabel the new label
     */
    public void setUpLabel(String upLabel)
    {
	upButton.setText(upLabel);
    }

    /**
     * sets the label of the down button
     * @param downLabel the new label
     */
    public void setDownLabel(String downLabel)
    {
	downButton.setText(downLabel);
    }

    /**
     * sets the label of the left JList
     * @param sourceListLabelText the new label
     */
    public void setSourceListLabelText(String sourceListLabelText)
    {
	this.sourceListLabel.setText(sourceListLabelText);
    }

    /**
     * sets the label of the right JList
     * @param sinkListLabelText the new label
     */
    public void setSinkListLabelText(String sinkListLabelText)
    {
	this.sinkListLabel.setText(sinkListLabelText);
    }

    private int getOriginalSourceItemIndex(Object sourceItem)
    {
	return allItems.indexOf(sourceItem);
    }

    private int getInsertIndex(Object sourceItem)
    {
	int originalIndex=getOriginalSourceItemIndex(sourceItem);
	int i=0;
	while (i<sourceModel.getSize())
	{
	    if (getOriginalSourceItemIndex(sourceModel.getElementAt(i))>originalIndex)
	    {
		break;
	    }
	    i++;
	}
	return i;	
    }

    private void checkAddButton()
    {
	addButton.setEnabled(sourceList.getSelectedIndices().length!=0);
    }

    private void checkRemoveButton()
    {
	removeButton.setEnabled(sinkList.getSelectedIndices().length!=0);	 
    }

    private void checkUpAndDownButtons()
    {
	int[] selected=sinkList.getSelectedIndices();
	boolean enableUp=false;
	boolean enableDown=false;
	int max=sinkModel.getSize()-1;
	for (int i=0;i<selected.length;i++)
	{
	    int index=selected[i];
	    if (index>0) enableUp=true;
	    if (index<max) enableDown=true;
	    if (enableUp && enableDown) break;
	}
	upButton.setEnabled(enableUp);
	downButton.setEnabled(enableDown);
    }

    class SourceSelectionListener implements ListSelectionListener
    {
	public void valueChanged(ListSelectionEvent lousy)
	{
	    checkAddButton();
	}
    }

    class SinkSelectionListener implements ListSelectionListener
    {
	public void valueChanged(ListSelectionEvent lousy)
	{
	    checkRemoveButton();
	    checkUpAndDownButtons();
	}
    }

    class UpAction implements ActionListener
    {
	public void actionPerformed(ActionEvent ae)
	{
	    Debug.trace(this, Debug.DP3, "in UpAction");
	    int[] selected=sinkList.getSelectedIndices();
	    for (int i=0;i<selected.length;i++)
	    {
		int oldIndex=selected[i];
		if (oldIndex>0)
		{
		    int newIndex=oldIndex-1;
		    sinkModel.insertElementAt(sinkModel.remove(oldIndex), newIndex);
		    sinkList.setSelectedIndex(newIndex);
		}		    
	    }
	    checkUpAndDownButtons();
	    checkRemoveButton();
	}
    }

    class DownAction implements ActionListener
    {
	public void actionPerformed(ActionEvent ae)
	{
	    Debug.trace(this, Debug.DP3, "in DownAction");
	    int[] selected=sinkList.getSelectedIndices();
	    for (int i=0;i<selected.length;i++)
	    {
		int oldIndex=selected[i];
		if (oldIndex<sinkModel.getSize()-1)
		{
		    int newIndex=oldIndex+1;
		    sinkModel.insertElementAt(sinkModel.remove(oldIndex), newIndex);
		    sinkList.setSelectedIndex(newIndex);
		}
	    }	   
	    checkUpAndDownButtons();
	}
    }

    class AddAction implements ActionListener
    {
	public void actionPerformed(ActionEvent ae)
	{
	    Debug.trace(this, Debug.DP3, "in AddAction");
	    //take selected item in source and move it into sink
	    int[] selected=sourceList.getSelectedIndices();
	    for (int i=0;i<selected.length;i++)
	    {
		Object o=sourceModel.remove(selected[i]);
		sinkModel.addElement(o);
	    }
	    checkAddButton();
	    checkUpAndDownButtons();
	}
    }

    class RemoveAction implements ActionListener
    {
	public void actionPerformed(ActionEvent ae)
	{
	    Debug.trace(this, Debug.DP3, "in RemoveAction");
	    //take selected item in sink and move it into source
	    int[] selected=sinkList.getSelectedIndices();
	    for (int i=0;i<selected.length;i++)
	    {
		Object o=sinkModel.remove(selected[i]);
		
		sourceModel.insertElementAt(o,getInsertIndex(o) );
	    }
	    checkRemoveButton();
	    checkUpAndDownButtons();
	}
    }
}


    

/* $Log: DualListPanel.java,v $
/* Revision 1.5  2001/01/04 06:02:49  smulloni
/* added more javadoc documentation.
/*
/* Revision 1.4  2000/11/09 23:35:13  smullyan
/* log added to every Java file, with the help of python.  Lock stealing
/* implemented, and treatment of locks made more robust.
/* */
