/*
 *  Copyright (c) 2000, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV  is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package org.skunk.swing.filechooser;

import java.awt.Component;
import java.io.File;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Vector;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import org.skunk.swing.TreeNodeChooser;

/**
 * a file chooser similar to GTK+'s GtkFileSelection widget
 */
public class LocalFileChooser extends TreeNodeChooser
{
    /**
     * constructor with a specified file root.
     * @param fileRoot the root of the desired portion of the file system
     */
    public LocalFileChooser(String fileRoot)
    {
	super(new DefaultTreeModel(new FileNode(new File(fileRoot))));
	setComboBoxCellRenderer(new FileComboRenderer());
    }

    /**
     * constructor with the system file roots.
     */
    public LocalFileChooser()
    {
	super(new DefaultTreeModel(FileNode.getRootNode()));
	if (FileNode.listRoots().length>1)
	    setRootVisible(false);
	setComboBoxCellRenderer(new FileComboRenderer());
    }

    /**
     * a test of the file chooser, for development purposes.
     */
    public static class Test
    {
	TreeNodeChooser chooser;
	public Test(String fileRoot)
	{
	    initModel(fileRoot);  
	}

	private void initModel(String fileRoot)
	{
	    if (fileRoot!=null)
		chooser=new LocalFileChooser(fileRoot);
	    else chooser=new LocalFileChooser();
	}

	public void showDialog()
	{
	    int option=JOptionPane.showConfirmDialog(null, 
						     chooser, 
						     "choose a file", 
						     JOptionPane.OK_CANCEL_OPTION,
						     JOptionPane.QUESTION_MESSAGE,
						     null);
	    if (option==JOptionPane.OK_OPTION)
	    {
		System.out.println(chooser.getEntryFieldText());
	    }
	    else System.out.println("CANCELLED");
	}

	public static void main(String[] args)
	{
	    //parse args
	    String initialPath=null;
	    SelectionMode  mode=null;
	    if (args.length>=1)
		initialPath=args[0];
	  
	    if (args.length>=2)
	    {
		if (args[1].equalsIgnoreCase("branch"))
		    mode=SelectionMode.BRANCH_ONLY;
		else if (args[1].equalsIgnoreCase("leaf"))
		    mode=SelectionMode.LEAF_ONLY;
		else if (args[1].equalsIgnoreCase("both"))
		    mode=SelectionMode.LEAF_AND_BRANCH;
	    }
	    Test test=new Test(initialPath);
	
	    if (mode!=null) 
		test.chooser.setSelectionMode(mode);

	    test.showDialog();
	    System.exit(0);
	}
   }

    /**
     * renderer for the combo box
     */
    protected class FileComboRenderer extends DefaultListCellRenderer
    {
	protected FileComboRenderer()
	{
	    super();
	}

	public Component getListCellRendererComponent(JList list,
						      Object value,
						      int index,
						      boolean isSelected,
						      boolean cellHasFocus)
	{
	    Object path;
	    if (value instanceof FileNode)
	    {
		File f=((FileNode)value).getFile();
		if (f!=null)
		    path=f.getAbsolutePath();
		else path="what is this?";
	    }
	    else path=value;
	    return super.getListCellRendererComponent(list,
						      path,
						      index,
						      isSelected,
						      cellHasFocus);
	}
    }
    

    public static class FileNode extends DefaultMutableTreeNode implements Comparable
    {
	private File file;
	private Vector children;
	private static FileNode rootNode;
	private static File[] roots=File.listRoots();

	public static final boolean isRoot(File f)
	{
	    for (int i=0;i<roots.length;i++)
	    {
		if (roots[i].equals(f))
		    return true;
	    }
	    return false;
	}

	/**
	 * equivalent to calling File.listRoots(), but retrieves statically cached information.
	 * on Windows, calling File.listRoots() causes a dialog to appear if there is no disk in a
	 * removable drive! Aaargh.
	 */
	public static File[] listRoots()
	{
	    return roots;
	}

	public static FileNode getRootNode()
	{
	    if (rootNode==null)
	    {
		if (roots.length>2)
		    rootNode=new FileNode(null);
		else rootNode=new FileNode(roots[0]);	       
	    }
	    return rootNode;
	}
	
	public FileNode(File f)
	{
	    super();
	    if (f==null) setUserObject(null);
	    else 
	    {
		if (isRoot(f))
		{
		    String s=f.toString();
		    setUserObject(s);
		}
		else
		{
		    setUserObject(f.getName());
		}
	    }
	    this.file=f;	
	}
	
	private void initChildren()
	{
	    children=new Vector();
	    File[] files=(file!=null) 
		    ? file.listFiles()
		    : roots;
	    if (files!=null)
	    {
		Arrays.sort(files);
		for (int i=0;i<files.length;i++)
		{
		    children.add(new FileNode(files[i]));
		}
	    }
	}
	
	public int getChildCount()
	{
	    if (children==null)
		initChildren();
	    return children.size();
	}
	
	public boolean isLeaf()
	{
	    return file!=null && !file.isDirectory();
	}
	
	public TreeNode getChildAt(int index)
	{
	    if (children==null)
		initChildren();
	    if (index>=0 && index<getChildCount())
		return (TreeNode)children.elementAt(index);
	    return null;
	}
	
	public int getIndex(TreeNode child)
	{
	    if (children==null)
		initChildren();
	    if (child instanceof FileNode)
	    {
		return children.indexOf(child);
	    }
	    return -1;
	}
	
	public Enumeration children()
	{
	    if (children==null)
		initChildren();
	    return children.elements();
	}
	
	public File getFile()
	{
	    return this.file;
	}

	public int compareTo(Object other)
	{
	    if (other instanceof FileNode)
	    {
		if (this.file!=null)
		    return this.file.compareTo(((FileNode)other).getFile());
		else return Integer.MIN_VALUE;
	    }
	    else throw new ClassCastException("FileNodes must be compared to other FileNodes");
	}
    }
}

/* $Log */
