/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/


package org.skunk.swing.text.syntax;

import java.io.IOException;
import java.io.StringReader;

%%
%public
%class HTMLFlexScanner
%extends FlexScanner

%unicode
%pack

%char

%type int

%{
  private boolean inScript=false;
  private boolean open=false;
  private StringBuffer scriptBuffer=new StringBuffer();
  private boolean inLanguage=false;
  private String scriptLanguage="javascript";

  public void scan() throws IOException
  {
     while ( !yy_atEOF )	
     {
       yylex();
     }
  }

  private int applyStyle(int type) 
  {
     return applyStyle(type, yychar, yylength());
  }
%}

CommentStart = \<\!--
CommentEnd = --\>
OpenTagStart = \<
CloseTagStart = \<\/
TagEnd = \>

CharacterEntity = \&[:jletterdigit:]+;
StringLiteral=(([\"][^\"]*[\"])|([\'][^\']*[\']))
DocTagStart=("<!")
Doctype=([Dd][Oo][Cc][Tt][Yy][Pp][Ee])
DoctypeText=(([^\>\"\']|{StringLiteral})*)
DoctypeTag=({DocTagStart}{Doctype}{DoctypeText}{TagEnd})

Attribute=(([^\ \r\n\f\t\<\>\=])*)

ScriptEnd=\<\/[Ss][Cc][Rr][Ii][Pp][Tt]

%state TAG_STATE, TAG_ATTRIBUTE_STATE, TAG_VALUE_STATE, COMMENT_STATE, SCRIPT_STATE

%%

<YYINITIAL> {
  {CharacterEntity}     {       
				return applyStyle(SyntaxStyle.TAG3);  
			}

  {OpenTagStart}/[^/]	{ 
			  	yybegin(TAG_STATE);
                          	spanStart=yychar;
				open=true;
                          	return applyStyle(SyntaxStyle.SEPARATOR2);
			}
  {CloseTagStart}	{ 
			  	yybegin(TAG_STATE);
                          	spanStart=yychar;
				open=false;
                          	return applyStyle(SyntaxStyle.SEPARATOR2);
			}
  {TagEnd}		{
				if (inScript) 
				{
				   scriptBuffer.setLength(0);
                                   spanStart=yychar+yylength();
				   yybegin(SCRIPT_STATE);
				}
				return applyStyle(SyntaxStyle.SEPARATOR2);
			}

  {CommentStart}        {
				yybegin(COMMENT_STATE);
				spanStart=yychar;
				return applyStyle(SyntaxStyle.COMMENT2);
			}

  {DoctypeTag}          {
				return applyStyle(SyntaxStyle.LITERAL2);
			}
}

<TAG_STATE> {
  {Attribute}          	{	
				yybegin(TAG_ATTRIBUTE_STATE);
				inScript=open && yytext().trim().equalsIgnoreCase("script");
				return applyStyle(SyntaxStyle.TAG1);
      			}
  {TagEnd}              {
				yybegin(YYINITIAL);
				open=false;
				return applyStyle(SyntaxStyle.ERROR);
			}
}

<TAG_ATTRIBUTE_STATE> {
  {TagEnd}              {
		  		//yybegin((inScript) ? SCRIPT_STATE : YYINITIAL);
				open=false;
                                //I should pushback and return to YYINITIAL, and apply this test there
				//return applyStyle(SyntaxStyle.SEPARATOR2);
				yypushback(yylength());
				yybegin(YYINITIAL);
			}
  "="                   {       
				yybegin(TAG_VALUE_STATE);
				return applyStyle(SyntaxStyle.OPERATOR);
			}
  {Attribute}           {
 			        inLanguage=inScript && yytext().equalsIgnoreCase("language");                                 
			        return applyStyle(SyntaxStyle.TAG2);  
                        }
}

<TAG_VALUE_STATE> {
  {Attribute}|{StringLiteral} {
                                if (inLanguage)
                                {
	                          scriptLanguage=yytext().toLowerCase();
				  if (scriptLanguage.startsWith("\"")
                                      || scriptLanguage.startsWith("'"))
                                      scriptLanguage=scriptLanguage.substring(1, Math.max(1, scriptLanguage.length()-1));
                                  
				  if (scriptLanguage.startsWith("javascript") 
                                      || scriptLanguage.equals("jscript")
                                      || scriptLanguage.equals("ecmascript"))
                                      scriptLanguage="javascript";
                                }
                                else scriptLanguage="javascript";

				yybegin(TAG_ATTRIBUTE_STATE);
				return applyStyle(SyntaxStyle.LITERAL1);
			}
  {TagEnd}		{
				yybegin(YYINITIAL);
				open=false;
				return applyStyle(SyntaxStyle.ERROR);
			}
}

<COMMENT_STATE> {
  {CommentEnd}		{
				yybegin(YYINITIAL);
				return applyStyle(SyntaxStyle.COMMENT3,
					spanStart,
					yylength()+yychar-spanStart);
			}

}
<SCRIPT_STATE> {
  
  {ScriptEnd}         	{
				inScript=false;
				String s=scriptBuffer.toString();
				FlexScanner scanner=Flexicizer.getScanner(FileMode.getMode(scriptLanguage), 
									  new StringReader(s));
                                if (scanner!=null) 
				{ 
				   scanner.setStyleBuffer(this.styleBuffer);
				   scanner.setOffset(offset+spanStart);
                                   scanner.scan();
                                }
				
				yybegin(YYINITIAL);			
				yypushback(yylength());
	
			}
  .			{ 
	                        scriptBuffer.append(yytext());
                        }
  \n                    {      scriptBuffer.append("\n"); }
  
}

/* error fallback */
.|\n                             { return applyStyle(SyntaxStyle.DEFAULT); }

<<EOF>>	                         { return -1;}

