/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/

/*
   This jflex file is based on the java.flex file that comes with the JFlex distribution.
*/

package org.skunk.swing.text.syntax;

//for main method
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

%%
%public
%class JavaFlexScanner
%extends FlexScanner

%unicode
%pack

%char

%type int

%{
  public void scan() throws IOException
  {
     while ( !yy_atEOF )	
     {
       yylex();
     }
  }

  private int applyStyle(int type) 
  {
     return applyStyle(type, yychar, yylength());
  }

  /**
   * Runs the scanner on input files.
   *
   * This is a standalone scanner, i.e. it will print any unmatched
   * text to System.out unchanged.
   *
   * @param argv   the command line, contains the filenames to run
   *               the scanner on.
   */
  public static void main(String argv[]) {
      boolean DEBUG=false;
      
    for (int i = 0; i < argv.length; i++) {
      JavaFlexScanner scanner = null;
      try {
        scanner = new JavaFlexScanner( new FileReader(argv[i]) );
      }
      catch (FileNotFoundException e) {
        System.out.println("File not found : \""+argv[i]+"\"");
        System.exit(1);
      }
      catch (IOException e) {
        System.out.println("Error opening file \""+argv[i]+"\"");
        System.exit(1);
      }
      catch (ArrayIndexOutOfBoundsException e) {
        System.out.println("Usage : java JavaFlexScanner <inputfile>");
        System.exit(1);
      }

      try {
	  long time=System.currentTimeMillis();
	  while ( !scanner.yy_atEOF )	
	  {
	      int t=scanner.yylex();
	      if (t==-1) 
	      {
		  long nextTime=System.currentTimeMillis();
		  System.out.println(new StringBuffer("SCAN COMPLETED IN ")
				     .append(nextTime-time)
				     .append(" milliseconds")
				     .toString());
		  System.exit(0);
	      }
	      else if (DEBUG)
	      {
		  System.out.print(SyntaxStyle.getStyle(StyleBufferUtilities.getStyle(t)).getName()); 
		  System.out.print(": \t");
		  System.out.println(scanner.yytext());
	      }
	  }
      }
      catch (IOException e) {
        System.out.println("An I/O error occured while scanning :");
        System.out.println(e);
        System.exit(1);
      }
      catch (Exception e) {
        e.printStackTrace();
        System.exit(1);
      }
    }
  }
%}

 /* main character classes */
LineTerminator = \r|\n|\r\n
InputCharacter = [^\r\n]

WhiteSpace = {LineTerminator} | [ \t\f]

/* comments */

//MultiLineComment = "/*" [^*] {CommentContent} \*+ "/"    
//SingleLineComment = "//" {InputCharacter}* {LineTerminator}

CommentContent = ( [^*] | \*+[^*/] )*

/* identifiers */
Identifier = [:jletter:][:jletterdigit:]*

/* integer literals */
DecIntegerLiteral = 0 | [1-9][0-9]*
DecLongLiteral    = {DecIntegerLiteral} [lL]

HexIntegerLiteral = 0 [xX] 0* {HexDigit} {1,8}
HexLongLiteral    = 0 [xX] 0* {HexDigit} {1,16} [lL]
HexDigit          = [0-9a-fA-F]

OctIntegerLiteral = 0+ [1-3]? {OctDigit} {1,15}
OctLongLiteral    = 0+ 1? {OctDigit} {1,21} [lL]
OctDigit          = [0-7]
    
/* floating point literals */        
FloatLiteral  = ({FLit1}|{FLit2}|{FLit3}|{FLit4}) [fF]
DoubleLiteral = {FLit1}|{FLit2}|{FLit3}|{FLit4}

FLit1 = [0-9]+ \. [0-9]* {Exponent}?
FLit2 = \. [0-9]+ {Exponent}?
FLit3 = [0-9]+ {Exponent}
FLit4 = [0-9]+ {Exponent}?

Exponent = [eE] [+\-]? [0-9]+

/* string and character literals */
SingleCharacter = [^\r\n\'\\]

Javadoc = \@author|\@version|\@since|\@param|\@return|\@see|\@link|\@exception|\@deprecated

%state STRING_STATE, CHAR_STATE, SINGLE_LINE_COMMENT_STATE, DOCCOMMENT_STATE, MULTI_LINE_COMMENT_STATE

%%

<YYINITIAL> {


  /* keywords */

  "abstract"                     |
  "class"                        |
  "const"                        |
  "extends"                      |
  "final"                        |
  "implements"                   |
  "interface"                    |
  "native"                       |
  "new"                          |
  "private"                      |
  "protected"                    |
  "public"                       |
  "super"                        |
  "static"                       |
  "strictfp"                     |
  "synchronized"                 |
  "this"                         |
  "throws"                       |
  "transient"                    |
  "volatile"                     { return applyStyle(SyntaxStyle.KEYWORD1); }

  "boolean"                      |
  "byte"                         |
  "char"                         |
  "double"                       |
  "float"                        |
  "int"                          |
  "long"                         |
  "null"                         |
  "short"                        |
  "void"                         { return applyStyle(SyntaxStyle.KEYWORD2); }

  "break"                        |
  "case"                         |
  "catch"                        |
  "continue"                     |
  "do"                           |
  "else"                         |
  "finally"                      |
  "for"                          |
  "default"                      |
  "goto"                         |
  "if"                           |
  "return"                       |
  "switch"                       |
  "throw"                        |
  "try"                          |
  "while"                        { return applyStyle(SyntaxStyle.KEYWORD3); }

  "import"                       |
  "package"                      { return applyStyle(SyntaxStyle.INCLUDE);  }


  
  /* boolean literals */
  "true"                         |
  "false"                        { return applyStyle(SyntaxStyle.LITERAL4); }
    
  /* separators */
  "("                            |
  ")"                            |
  "{"                            |
  "}"                            |
  "["                            |
  "]"                            |
  ";"                            |
  ","                            |
  "."                            { return applyStyle(SyntaxStyle.SEPARATOR1); }
  
  /* operators */
  "="                            |
  ">"                            |
  "<"                            |
  "!"                            |
  "~"                            |
  "?"                            |
  ":"                            |
  "=="                           |
  "<="                           |
  ">="                           |
  "!="                           |
  "&&"                           |
  "||"                           |
  "++"                           |
  "--"                           |
  "+"                            |
  "-"                            |
  "*"                            |
  "/"                            |
  "&"                            |
  "|"                            |
  "^"                            |
  "%"                            |
  "<<"                           |
  ">>"                           |
  ">>>"                          |
  "+="                           |
  "-="                           |
  "*="                           |
  "/="                           |
  "&="                           |
  "|="                           |
  "^="                           |
  "%="                           |
  "<<="                          |
  ">>="                          |
  ">>>="                         |
  "instanceof"                   { return applyStyle(SyntaxStyle.OPERATOR); }

  /* numeric literals */

  {DecIntegerLiteral}            |
  {DecLongLiteral}               |
  {HexIntegerLiteral}            |
  {HexLongLiteral}               |
  {OctIntegerLiteral}            |
  {OctLongLiteral}               |
  {FloatLiteral}                 |
  {DoubleLiteral}                |
  {DoubleLiteral}[dD]            { return applyStyle(SyntaxStyle.LITERAL2); }
  
  /* comments */
  "/**"                          { 
                                   yybegin(DOCCOMMENT_STATE); 
                                   spanStart=yychar;
                                   return applyStyle(SyntaxStyle.COMMENT3);
                                 }

  "/*"                           { 
                                   yybegin(MULTI_LINE_COMMENT_STATE); 
                                   spanStart=yychar;
                                   return applyStyle(SyntaxStyle.COMMENT2);
                                 }

  "//"                           {
		                   yybegin(SINGLE_LINE_COMMENT_STATE);
                                   spanStart=yychar;
                                   return applyStyle(SyntaxStyle.COMMENT1);
                                 }

  /* string literal */
  \"                             { yybegin(STRING_STATE); spanStart=yychar; }

  /* character literal */
  \'                             { yybegin(CHAR_STATE); spanStart=yychar;}

  /* whitespace */
  {WhiteSpace}                   { /* ignore */ return applyStyle(SyntaxStyle.DEFAULT);}

  /* identifiers */ 
  {Identifier}                   { return applyStyle(SyntaxStyle.IDENTIFIER1); }  

}

<MULTI_LINE_COMMENT_STATE> {

   {CommentContent}              { applyStyle(SyntaxStyle.COMMENT2); }

  "*/"                           {
				   yybegin(YYINITIAL); 
                                   return applyStyle(SyntaxStyle.COMMENT2, 
                                                     spanStart, 
                                                     yylength()+yychar-spanStart);
                                 }                                                                   
}

<DOCCOMMENT_STATE> {
  {Javadoc}                      { 	                          
                                   applyStyle(SyntaxStyle.COMMENT3, 
                                              spanStart, 
                                              yychar-spanStart); 
                                   spanStart=yychar+yylength(); 
                                   return applyStyle(SyntaxStyle.TAG4); 
                                 }
 
   "*/"                          { 
                                   yybegin(YYINITIAL); 
                                   return applyStyle(SyntaxStyle.COMMENT3, 
                                                     spanStart, 
                                                     yylength()+yychar-spanStart);
                                 }
   .                             { return applyStyle(SyntaxStyle.COMMENT3); }	                           
}

<SINGLE_LINE_COMMENT_STATE> {
  {InputCharacter}               { applyStyle(SyntaxStyle.COMMENT1); }
  {LineTerminator}               { 
	                            yybegin(YYINITIAL);
                                    return applyStyle(SyntaxStyle.COMMENT1,
                                                      spanStart,
                                                      yylength()+yychar-spanStart);
                                 }
}   

<STRING_STATE> {
  [^\\]\"                        { 
                                   yybegin(YYINITIAL); 
                                   return applyStyle(SyntaxStyle.LITERAL1, 
                                                     spanStart, 
                                                     yylength()+yychar-spanStart); 
                                 }
    
  /* error cases.  We never throw an exception. */
  {LineTerminator}               { 
                                   yybegin(YYINITIAL); 
                                   return applyStyle(SyntaxStyle.ERROR, 
                                                     spanStart, 
                                                     yylength()+yychar-spanStart); 
                                 }
}

<CHAR_STATE> {
  {SingleCharacter}\'            |
  
  /* escape sequences */
  "\\b"\'                        |
  "\\t"\'                        |
  "\\n"\'                        |
  "\\f"\'                        |
  "\\r"\'                        |
  "\\\""\'                       |
  "\\'"\'                        |
  "\\\\"\'                       |
  \\[0-3]?{OctDigit}?{OctDigit}\' { 
                                    yybegin(YYINITIAL); 
                                    return applyStyle(SyntaxStyle.LITERAL3, 
                                    spanStart, 
                                    yylength()+yychar-spanStart); 
                                  }
  
  /* error cases */
  \\.                            |
  {LineTerminator}               { 
                                   yybegin(YYINITIAL); 
                                   return applyStyle(SyntaxStyle.ERROR, 
                                                     spanStart, 
                                                     yychar-spanStart); }
}

/* error fallback */
.|\n                             { return applyStyle(SyntaxStyle.DEFAULT); }

<<EOF>>	                         { return -1;}
