/*
 *  Copyright (c) 2001, Jacob Smullyan.
 *
 *  This is part of SkunkDAV, a WebDAV client.  See http://skunkdav.sourceforge.net/ 
 *  for the latest version.
 * 
 *  SkunkDAV is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  SkunkDAV is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with SkunkDAV; see the file COPYING.  If not, write to the Free
 *  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
*/


package org.skunk.swing.text.syntax;

import java.io.IOException;
import java.io.StringReader;
import org.skunk.trace.Debug;

%%
%public
%class JavascriptFlexScanner
%extends FlexScanner

%unicode
%pack

%char

%type int

%{
  private String quote;

  public void scan() throws IOException
  {
     while ( !yy_atEOF )	
     {
       yylex();
     }
  }

  private int applyStyle(int type) 
  {
     return applyStyle(type, yychar, yylength());
  }
%}

 /* main character classes */
LineTerminator = \r|\n|\r\n
InputCharacter = [^\r\n]

WhiteSpace = {LineTerminator} | [ \t\f]

/* comments */

CommentContent = ( [^*] | \*+[^*/] )*

/* identifiers */
Identifier = [:jletter:][:jletterdigit:]*

/* integer literals */
DecIntegerLiteral = 0 | [1-9][0-9]*
DecLongLiteral    = {DecIntegerLiteral} [lL]

HexIntegerLiteral = 0 [xX] 0* {HexDigit} {1,8}
HexLongLiteral    = 0 [xX] 0* {HexDigit} {1,16} [lL]
HexDigit          = [0-9a-fA-F]

OctIntegerLiteral = 0+ [1-3]? {OctDigit} {1,15}
OctLongLiteral    = 0+ 1? {OctDigit} {1,21} [lL]
OctDigit          = [0-7]
    
/* floating point literals */        
FloatLiteral  = ({FLit1}|{FLit2}|{FLit3}|{FLit4}) [fF]
DoubleLiteral = {FLit1}|{FLit2}|{FLit3}|{FLit4}

FLit1 = [0-9]+ \. [0-9]* {Exponent}?
FLit2 = \. [0-9]+ {Exponent}?
FLit3 = [0-9]+ {Exponent}
FLit4 = [0-9]+ {Exponent}?

Exponent = [eE] [+\-]? [0-9]+

%state STRING_STATE,  SINGLE_LINE_COMMENT_STATE,  MULTI_LINE_COMMENT_STATE

%%

<YYINITIAL> {


  /* keywords */

  "class"                        |
  "const"                        |
  "debugger"                     |
  "function"                     |
  "this"                         |
  "var"                          { return applyStyle(SyntaxStyle.KEYWORD1); }


  "null"                         |
  "void"                         { return applyStyle(SyntaxStyle.KEYWORD2); }

  "break"                        |
  "case"                         |
  "catch"                        |
  "continue"                     |
  "do"                           |
  "else"                         |
  "finally"                      |
  "for"                          |
  "default"                      |
  "goto"                         |
  "if"                           |
  "in"                           |
  "return"                       |
  "switch"                       |
  "throw"                        |
  "try"                          |
  "while"                        { return applyStyle(SyntaxStyle.KEYWORD3); }

  "import"                       |
  "export"                       |
  "with"                         { return applyStyle(SyntaxStyle.INCLUDE);  }


  
  /* boolean literals */
  "true"                         |
  "false"                        { return applyStyle(SyntaxStyle.LITERAL4); }
    
  /* separators */
  "("                            |
  ")"                            |
  "{"                            |
  "}"                            |
  "["                            |
  "]"                            |
  ";"                            |
  ","                            |
  "."                            { return applyStyle(SyntaxStyle.SEPARATOR1); }
  
  /* operators */
  "="                            |
  ">"                            |
  "<"                            |
  "!"                            |
  "~"                            |
  "?"                            |
  ":"                            |
  "=="                           |
  "==="                          |
  "!=="                          |
  "<="                           |
  ">="                           |
  "!="                           |
  "&&"                           |
  "||"                           |
  "++"                           |
  "--"                           |
  "+"                            |
  "-"                            |
  "*"                            |
  "/"                            |
  "&"                            |
  "|"                            |
  "^"                            |
  "%"                            |
  "<<"                           |
  ">>"                           |
  ">>>"                          |
  "+="                           |
  "-="                           |
  "*="                           |
  "/="                           |
  "&="                           |
  "|="                           |
  "^="                           |
  "%="                           |
  "<<="                          |
  ">>="                          |
  ">>>="                         |
  "delete"                       |
  "new"                          |
  "typeof"                       |
  "void"                         { return applyStyle(SyntaxStyle.OPERATOR); }

  /* numeric literals */

  {DecIntegerLiteral}            |
  {DecLongLiteral}               |
  {HexIntegerLiteral}            |
  {HexLongLiteral}               |
  {OctIntegerLiteral}            |
  {OctLongLiteral}               |
  {FloatLiteral}                 |
  {DoubleLiteral}                |
  {DoubleLiteral}[dD]            { return applyStyle(SyntaxStyle.LITERAL2); }
  
  /* comments */
  "/*"                           { 
                                   yybegin(MULTI_LINE_COMMENT_STATE); 
                                   spanStart=yychar;
                                   return applyStyle(SyntaxStyle.COMMENT2);
                                 }

  "//"                           {
		                   yybegin(SINGLE_LINE_COMMENT_STATE);
                                   spanStart=yychar;
                                   return applyStyle(SyntaxStyle.COMMENT1);
                                 }

  /* string literal */
  \"|\'                          { 
                                    quote=yytext();
                                    yybegin(STRING_STATE); 
                                    spanStart=yychar; 
                                 }

  /* whitespace */
  {WhiteSpace}                   { /* ignore */ return applyStyle(SyntaxStyle.DEFAULT);}

  /* identifiers */ 
  {Identifier}                   { return applyStyle(SyntaxStyle.IDENTIFIER1); }  

}

<MULTI_LINE_COMMENT_STATE> {

   {CommentContent}              { applyStyle(SyntaxStyle.COMMENT2); }

  "*/"                           {
				   yybegin(YYINITIAL); 
                                   return applyStyle(SyntaxStyle.COMMENT2, 
                                                     spanStart, 
                                                     yylength()+yychar-spanStart);
                                 }                                                                   
}

<SINGLE_LINE_COMMENT_STATE> {
  {InputCharacter}               { applyStyle(SyntaxStyle.COMMENT1); }
  {LineTerminator}               { 
	                            yybegin(YYINITIAL);
                                    return applyStyle(SyntaxStyle.COMMENT1,
                                                      spanStart,
                                                      yylength()+yychar-spanStart);
                                 }
}   

<STRING_STATE> {
  [^\\](\"|\')                  { 
				   if (yytext().endsWith(quote)) 
                                   {
                                      yybegin(YYINITIAL); 
                                      return applyStyle(SyntaxStyle.LITERAL1, 
                                                        spanStart, 
                                                        yylength()+yychar-spanStart); 
                                   }
                                 }
    
  /* error cases.  We never throw an exception. */
  {LineTerminator}               { 
                                   yybegin(YYINITIAL); 
                                   return applyStyle(SyntaxStyle.ERROR, 
                                                     spanStart, 
                                                     yylength()+yychar-spanStart); 
                                 }
}



/* error fallback */
.|\n                             { return applyStyle(SyntaxStyle.DEFAULT); }

<<EOF>>	                         { return -1;}
